/* Functions for the OpenStep and MacOSX (Aqua) window system.

   Copyright (C) 1989, 1992, 1993, 1994 Free Software Foundation.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Originally by Carl Edman
Updated by Christian Limpach (chris@nice.ch)
OpenStep/Rhapsody port by Scott Bender (sbender@harmony-ds.com)
MacOSX/Aqua port by Christophe de Dinechin (descubes@earthlink.net)
GNUstep port by Adrian Robert (arobert@cogsci.ucsd.edu)

*/


#import <AppKit/AppKit.h>

#include <signal.h>
#include <math.h>
// #include <mach/host_info.h>
#include "config.h"
#include "lisp.h"
#include "dispextern.h"
#include "nsterm.h"
#include "frame.h"
#include "window.h"
#include "buffer.h"
#include "dispextern.h"
#include "keyboard.h"
#include "blockinput.h"
#include "regex.h"
#include "termhooks.h"
#include "paths.h"
#include "charset.h"
#include "fontset.h"
#include "epaths.h"

#if 0
int fns_trace_num = 0;
#define TRACE(x)        fprintf(stderr, "%s:%d: [%d] " #x "\n",        \
                                __FILE__, __LINE__, ++fns_trace_num)
#else
#define TRACE(x)
#endif

#ifdef HAVE_NS

extern void free_frame_menubar();
void x_set_frame_parameters (struct frame *f, Lisp_Object alist);

extern struct re_pattern_buffer *compile_pattern ();

Lisp_Object Qns_frame_parameter;
Lisp_Object Qforeground_color;
Lisp_Object Qbackground_color;
Lisp_Object Qcursor_color;
Lisp_Object Qinternal_border_width;
Lisp_Object Qvisibility;
Lisp_Object Qicon_type;
Lisp_Object Qcursor_type;
Lisp_Object Qicon_left;
Lisp_Object Qicon_top;
Lisp_Object Qicon_type;
Lisp_Object Qleft;
Lisp_Object Qright;
Lisp_Object Qtop;
Lisp_Object Qicon_name;
Lisp_Object Qdisplay;
Lisp_Object Qnone;
Lisp_Object Qvertical_scroll_bars;
Lisp_Object Qauto_raise;
Lisp_Object Qauto_lower;
Lisp_Object Qbox;
Lisp_Object Qscroll_bar_width;
Lisp_Object Vx_bitmap_file_path;
Lisp_Object Qx_resource_name;

Lisp_Object Qface_set_after_frame_default;
Lisp_Object Qfontsize;
Lisp_Object Qbuffered;


extern Lisp_Object Qunderline, Qundefined;
extern Lisp_Object Qheight, Qminibuffer, Qname, Qonly, Qwidth;
extern Lisp_Object Qunsplittable, Qmenu_bar_lines, Qbuffer_predicate, Qtitle;
extern NSArray *ns_send_types, *ns_return_types, *ns_drag_types;

/* hack for OS X file panels */
char panelOK = 0;

/* Alist of elements (REGEXP . IMAGE) for images of icons associated
   to frames.*/
Lisp_Object Vns_icon_type_alist;


void check_ns (void)
{
  if (NSApp == nil)
    error ("OpenStep is not in use or not initialized");
}

/* Nonzero if we can use mouse menus. */
int have_menus_p ()
{
  return NSApp != nil;
}

/* Extract a frame as a FRAME_PTR, defaulting to the selected frame
   and checking validity for NS.  */

FRAME_PTR
check_ns_frame (frame)
     Lisp_Object frame;
{
  FRAME_PTR f;

  if (NILP (frame))
      f = selected_frame;
  else
    {
      CHECK_LIVE_FRAME (frame, 0);
      f = XFRAME (frame);
    }
  if (! FRAME_NS_P (f))
    error ("non-NS frame used");
  return f;
}

/* Let the user specify an NS display with a frame.
   nil stands for the selected frame--or, if that is not an NS frame,
   the first NS display on the list.  */

static struct ns_display_info *
check_ns_display_info (frame)
     Lisp_Object frame;
{
  if (NILP (frame))
    {
      if (FRAME_NS_P (selected_frame))
        return FRAME_NS_DISPLAY_INFO (selected_frame);
      else if (ns_display_list != 0)
        return ns_display_list;
      else
        error ("NS windows are not in use or not initialized");
    }
  else if (STRINGP (frame))
    return ns_display_info_for_name (frame);
  else
    {
      FRAME_PTR f;
      
      CHECK_LIVE_FRAME (frame, 0);
      f = XFRAME (frame);
      if (! FRAME_NS_P (f))
        error ("non-NS frame used");
      return FRAME_NS_DISPLAY_INFO (f);
    }
  return NULL;  // shut compiler up
}

static Lisp_Object ns_get_arg (Lisp_Object alist, Lisp_Object param,
                               char *owner, char *name)
{
  Lisp_Object tem;
   
  tem = Fassq (param, alist);
  if (!NILP(tem)) return Fcdr(tem);
   
  tem = Fassq (param, Vdefault_frame_alist);
  if (!NILP(tem)) return Fcdr(tem);
   
  if (!name) return Qunbound;
   
  tem = Fns_get_resource(owner ? build_string(owner) : Qnil,
                         build_string (name));
   
  if (NILP(tem)) return Qunbound;

  if (XTYPE(tem)==Lisp_String)
    {
      if (!strcmp(XSTRING(tem)->data,"YES")||
          !strcmp(XSTRING(tem)->data,"ON")||
          !strcmp(XSTRING(tem)->data,"TRUE")) return Qt;
      if (!strcmp(XSTRING(tem)->data,"NO")||
          !strcmp(XSTRING(tem)->data,"OFF")||
          !strcmp(XSTRING(tem)->data,"FALSE")) return Qnil;
    }

  return tem;
}

static Lisp_Object ns_default_parameter (struct frame *f, Lisp_Object alist,
                                         Lisp_Object prop, Lisp_Object deflt,
                                         char *owner, char *name,
                                         Lisp_Object *accum)
{
  Lisp_Object tem;

  tem = ns_get_arg (alist, prop, owner, name);
  if (EQ (tem, Qunbound)) tem=deflt;

  // accumulate into accum and caller is responsible for calling set()
  *accum = Fcons( Fcons (prop, tem), *accum);
  return tem;
}

void ns_set_foreground_color (struct frame *f, Lisp_Object arg,
                              Lisp_Object oldval)
{
  NSColor * col;

  if (ns_lisp_to_color(arg,&col))
    {
      store_frame_param(f,Qforeground_color,oldval);
      error("Unknown color");
    }
   
  [f->output_data.ns->foreground_color release];
  f->output_data.ns->foreground_color = [col retain];

  if (FRAME_NS_VIEW(f))
    {
      ns_recompute_basic_faces (f);
      if (FRAME_VISIBLE_P (f))
        redraw_frame (f);
    }
}

void ns_set_background_color (struct frame *f, Lisp_Object arg,
                              Lisp_Object oldval)
{
  NSColor * col;
  NSView *view = FRAME_NS_VIEW(f);
  float alpha;
  
  if (ns_lisp_to_color(arg,&col))
    {
      store_frame_param(f,Qbackground_color,oldval);
      error("Unknown color");
    }
  
  [f->output_data.ns->background_color release];
  f->output_data.ns->background_color = [col retain];
  if (view != nil)
    {
      [[view window] setBackgroundColor:col];
      alpha = [col alphaComponent];

#ifdef MACOSX
      // the alpha code below only works on 10.4, so we need to do something
      // else (albeit less good) here
      // Take a look at NSApplication.h for useful NSAppKitVersionNumber values
      if (NSAppKitVersionNumber < 744.0)
          [[view window] setAlphaValue: alpha];
#endif

      if (alpha != 1.0)
          [[view window] setOpaque: NO];
      else
          [[view window] setOpaque: YES];
      /// experimental
      col = NS_FACE_BACKGROUND(FRAME_DEFAULT_FACE(f));
      FRAME_DEFAULT_FACE(f)->background =
          (EMACS_UINT) [[col colorWithAlphaComponent: alpha] retain];
      [col release];
      ///
      ns_recompute_basic_faces(f);
      if (FRAME_VISIBLE_P(f))
        redraw_frame(f);
    }
}

void ns_set_cursor_color (struct frame *f, Lisp_Object arg, Lisp_Object oldval)
{
  NSColor *col;
  NSView *view = FRAME_NS_VIEW(f);
  if (ns_lisp_to_color(arg,&col))
    {
      store_frame_param(f,Qcursor_color,oldval);
      error("Unknown color");
    }
  
  [f->output_data.ns->desired_cursor_color release];
  f->output_data.ns->desired_cursor_color = [col retain];
  
  if (view != nil)
    ns_dumpcursor(1, f,f->cursor_x,f->cursor_y);
}

void ns_set_internal_border_width (struct frame *f, Lisp_Object arg,
                                   Lisp_Object oldval)
{
  if (XTYPE(arg)==Lisp_String)
    {
      arg=Fstring_to_number(arg, Qnil);
      store_frame_param(f,Qinternal_border_width,arg);
    }
    
  if (!NUMBERP(arg) || (XINT(arg)<0))
    {
      store_frame_param(f,Qinternal_border_width,oldval);
      error("Internal border width out of range");
    }
    
  f->output_data.ns->internal_border_width=XINT(arg);
  x_set_window_size (f, 0, f->width, f->height);
}

void ns_set_visibility (struct frame *f, Lisp_Object arg, Lisp_Object oldval)
{
  Lisp_Object frame;
  XSETFRAME (frame, f);

  if (XTYPE(arg)==Lisp_String)
    {
      arg=Fintern(arg, Qnil);
      store_frame_param(f,Qvisibility,arg);
    }

  if (NILP (arg))
    Fmake_frame_invisible (frame, Qt);
  else if (EQ (arg, Qicon))
    Ficonify_frame (frame);
  else
    Fmake_frame_visible (frame);
}

void ns_set_icon_name (struct frame *f, Lisp_Object arg,
                       Lisp_Object oldval)
{
  NSView *view = FRAME_NS_VIEW(f);
  TRACE(ns_set_icon_name);

  // see if it's changed
  if (STRINGP (arg))
    {
      if (STRINGP (oldval) && EQ (Fstring_equal (oldval, arg), Qt))
        return;
    }
  else if (!STRINGP (oldval) && EQ (oldval, Qnil) == EQ (arg, Qnil))
    return;
  
  f->icon_name = arg;

  if (NILP (arg))
    {
      if (!NILP (f->title))
        arg = f->title;
      else
        // explicit name and no icon-name -> explicit_name
        if (f->explicit_name)
          arg = f->name;
        else
          {
            // no explicit name and no icon-name ->
            // name has to be rebuild from icon_title_format
            windows_or_buffers_changed++;
            return;
          }
    }

  /* Don't change the name if it's already NAME.  */
  if ([[view window] miniwindowTitle] &&
      ([[[view window] miniwindowTitle]
             isEqualToString:[NSString stringWithUTF8String:
                                           XSTRING (arg)->data]]))
    return;
  
  [[view window] setMiniwindowTitle:
        [NSString stringWithUTF8String:XSTRING (arg)->data]];
}

void x_set_name_iconic (struct frame *f, Lisp_Object name, int explicit)
{
  NSView *view = FRAME_NS_VIEW(f);
  TRACE(x_set_name_iconic);

  /* Make sure that requests from lisp code override requests from 
     Emacs redisplay code.  */
  if (explicit)
    {
      /* If we're switching from explicit to implicit, we had better
         update the mode lines and thereby update the title.  */
      if (f->explicit_name && NILP (name))
        update_mode_lines = 1;

      f->explicit_name = ! NILP (name);
    }
  else if (f->explicit_name)
    name = f->name;

  // title overrides explicit name
  if (! NILP (f->title))
    name = f->title;
  
  // icon_name overrides title and explicit name
  if (! NILP (f->icon_name))
    name = f->icon_name;
    
  if (NILP (name))
    name = build_string
        ([[[NSProcessInfo processInfo] processName] UTF8String]);
  else
    CHECK_STRING (name, 0);
  
  /* Don't change the name if it's already NAME.  */
  if ([[view window] miniwindowTitle] &&
      ([[[view window] miniwindowTitle]
             isEqualToString:[NSString stringWithUTF8String:
                                           XSTRING (name)->data]]))
    return;
  
  [[view window] setMiniwindowTitle:
        [NSString stringWithUTF8String:XSTRING (name)->data]];
}

void x_set_name (struct frame *f, Lisp_Object name, int explicit)
{
  NSView *view = FRAME_NS_VIEW(f);
  TRACE(x_set_name);

  /* Make sure that requests from lisp code override requests from 
     Emacs redisplay code.  */
  if (explicit)
    {
      /* If we're switching from explicit to implicit, we had better
         update the mode lines and thereby update the title.  */
      if (f->explicit_name && NILP (name))
        update_mode_lines = 1;
      
      f->explicit_name = ! NILP (name);
    }
  else if (f->explicit_name)
    return;
  
  if (NILP (name))
    name = build_string
        ([[[NSProcessInfo processInfo] processName] UTF8String]);
  
  f->name = name;

  // title overrides explicit name
  if (! NILP (f->title))
    name = f->title;

  CHECK_STRING (name, 0);

  /* Don't change the name if it's already NAME.  */
  if ([[[view window] title]
            isEqualToString:[NSString stringWithUTF8String:
                                          XSTRING (name)->data]])
    return;
  [[view window] setTitle:[NSString stringWithUTF8String:
                                        XSTRING (name)->data]];
}

/* This function should be called when the user's lisp code has
   specified a name for the frame; the name will override any set by the
   redisplay code.  */
void
x_explicitly_set_name (f, arg, oldval)
     FRAME_PTR f;
     Lisp_Object arg, oldval;
{
  TRACE(x_explicitly_set_name);
  x_set_name_iconic (f, arg, 1);
  x_set_name (f, arg, 1);
}

/* This function should be called by Emacs redisplay code to set the
   name; names set this way will never override names set by the user's
   lisp code.  */
void
x_implicitly_set_name (f, arg, oldval)
     FRAME_PTR f;
     Lisp_Object arg, oldval;
{
  TRACE(x_implicitly_set_name);
  if (FRAME_ICONIFIED_P (f))
    x_set_name_iconic (f, arg, 0);
  else
    x_set_name (f, arg, 0);
}

/* Change the title of frame F to NAME.
   If NAME is nil, use the frame name as the title.

   If EXPLICIT is non-zero, that indicates that lisp code is setting the
   name; if NAME is a string, set F's name to NAME and set
   F->explicit_name; if NAME is Qnil, then clear F->explicit_name.

   If EXPLICIT is zero, that indicates that Emacs redisplay code is
   suggesting a new name, which lisp code should override; if
   F->explicit_name is set, ignore the new name; otherwise, set it.  */

void
ns_set_title (f, name)
     struct frame *f;
     Lisp_Object name;
{
  TRACE(ns_set_title);
  /* Don't change the title if it's already NAME.  */
  if (EQ (name, f->title))
    return;

  update_mode_lines = 1;

  f->title = name;
}

void x_set_name_as_filename (struct frame *f)
{
  NSView *view = FRAME_NS_VIEW(f);
  Lisp_Object name;
  Lisp_Object buf = XWINDOW (f->selected_window)->buffer;
  const char *title;
  NSAutoreleasePool *pool;
  TRACE(x_set_name_as_filename);

  if (f->explicit_name || ! NILP (f->title))
    return;

  pool = [[NSAutoreleasePool alloc] init];
  name=XBUFFER(buf)->filename;
  if (NILP(name) || FRAME_ICONIFIED_P (f)) name=XBUFFER(buf)->name;

  if (FRAME_ICONIFIED_P (f) && !NILP (f->icon_name))
    name = f->icon_name;
    
  if (NILP (name))
    name = build_string
        ([[[NSProcessInfo processInfo] processName] UTF8String]);
  else
    CHECK_STRING (name, 0);

  title = FRAME_ICONIFIED_P (f) ? [[[view window] miniwindowTitle] UTF8String]
                                : [[[view window] title] cString];
  
  if (title && (! strcmp (title, XSTRING (name)->data)))
    {
      [pool release];
      return;
    }

  if (! FRAME_ICONIFIED_P (f))
    {
#ifdef MACOSX
      // work around a bug observed on 10.3 where
      // setTitleWithRepresentedFilename does not clear out previous state
      // if given filename does not exist
      NSString *str = [NSString stringWithUTF8String:XSTRING(name)->data];
      if (![[NSFileManager defaultManager] fileExistsAtPath: str])
        {
          [[view window] setTitleWithRepresentedFilename: @""];
          [[view window] setTitle: str];
        }
      else
        {
          [[view window] setTitleWithRepresentedFilename: str];
        }
#else
      [[view window] setTitleWithRepresentedFilename:
                         [NSString stringWithUTF8String:XSTRING (name)->data]];
#endif
      f->name = name;
    }
  else
    {
      [[view window] setMiniwindowTitle:
            [NSString stringWithUTF8String:XSTRING (name)->data]];
    }
  [pool release];
}

void ns_set_doc_edited (struct frame *f, Lisp_Object arg,
                        Lisp_Object oldval)
{
  NSView *view = FRAME_NS_VIEW(f);
  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
  [[view window] setDocumentEdited:!NILP(arg)];
  [pool release];
}

void x_set_menu_bar_lines (struct frame *f, Lisp_Object arg,
                           Lisp_Object oldval)
{
  FRAME_MENU_BAR_LINES (f) = 0;
  if (XTYPE(arg)==Lisp_String)
    {
      arg=Fstring_to_number(arg, Qnil);
    }

  if (FRAME_MINIBUF_ONLY_P(f))
    return;

  if (NILP(arg) || (XTYPE(arg)==Lisp_Int && XINT(arg)==0))
    {
      if (FRAME_EXTERNAL_MENU_BAR (f) == 1)
        free_frame_menubar (f);
      FRAME_EXTERNAL_MENU_BAR (f) = 0;
    }
  else
    {
      FRAME_EXTERNAL_MENU_BAR (f) = 1;
    }
}

void ns_implicitly_set_icon_type (struct frame *f)
{
  Lisp_Object tem;
  EmacsView *view = FRAME_NS_VIEW(f);
  id image=nil;
  Lisp_Object chain, elt;
  NSAutoreleasePool *pool;
  TRACE(ns_implicitly_set_icon_type);

  pool = [[NSAutoreleasePool alloc] init];  
  if (f->output_data.ns->miniimage
      && [[NSString stringWithUTF8String:XSTRING(f->name)->data] 
               isEqualToString:[(NSImage *)f->output_data.ns->miniimage name]])
    {
      [pool release];
      return;
    }

  tem = assq_no_quit (Qicon_type, f->param_alist);
  if (CONSP (tem) && ! NILP (XCONS (tem)->cdr))
    return;

  // Avoid generating a warning on MacOS X for internal names (*scratch*,
  // etc.).
  if (XSTRING(f->name)->data[0] == '/')
    {
      for (chain = Vns_icon_type_alist;
           (image==nil) && CONSP (chain);
           chain = XCONS (chain)->cdr)
        {
          elt = XCONS(chain)->car;
          if (SYMBOLP (elt) && elt == Qt)
            {
              NSString *str =
                [NSString stringWithUTF8String: XSTRING(f->name)->data];
              if ([[NSFileManager defaultManager] fileExistsAtPath: str])
                image=[[[NSWorkspace sharedWorkspace] iconForFile:str] retain];
            }
          else if (CONSP (elt) &&
                   STRINGP (XCONS(elt)->car) &&
                   STRINGP (XCONS(elt)->cdr) && 
                   fast_string_match (XCONS(elt)->car, f->name) >= 0)
            {
              image = [EmacsImage allocInitFromFile:XCONS (elt)->cdr];
              if (image == nil)
                image=[[NSImage imageNamed:
                                  [NSString stringWithUTF8String:
                                              XSTRING(XCONS (elt)->cdr)->data]]
                        retain];
            }
        }
    }

  if (image==nil) 
    image=[[[NSWorkspace sharedWorkspace] iconForFileType:@"text"] retain];
  
  [f->output_data.ns->miniimage release];
  f->output_data.ns->miniimage=image;
  [view setMiniwindowImage];
  [pool release];
}

void ns_set_icon_type (struct frame *f, Lisp_Object arg,
                       Lisp_Object oldval)
{
  EmacsView *view = FRAME_NS_VIEW(f);
  id image=nil;
  TRACE(ns_set_icon_type);

  if (!NILP (arg) && SYMBOLP (arg))
    {
      arg=build_string(XSYMBOL(arg)->name->data);
      store_frame_param(f,Qicon_type,arg);
    }
  
  // do it the implicit way
  if (NILP (arg))
    {
      ns_implicitly_set_icon_type (f);
      return;
    }
  
  CHECK_STRING (arg,0);
  
  image = [EmacsImage allocInitFromFile:arg];
  if (image == nil)
    image=[NSImage imageNamed:[NSString stringWithUTF8String:
                                            XSTRING(arg)->data]];
  
  if (image==nil) image=[NSImage imageNamed:@"text"];
  
  f->output_data.ns->miniimage=image;
  [view setMiniwindowImage];
}

int ns_lisp_to_cursor_type (Lisp_Object arg)
{
  char *str;
  if (XTYPE(arg) == Lisp_String)
    str=XSTRING(arg)->data;
  else if (XTYPE(arg) == Lisp_Symbol)
    str=XSYMBOL(arg)->name->data;
  else return -1;
  if (!strcmp(str,"bar"))    return bar;
  if (!strcmp(str,"box"))    return filled_box;
  if (!strcmp(str,"hollow")) return hollow_box;
  if (!strcmp(str,"line"))   return line;
  if (!strcmp(str,"no"))     return no_highlight;
  return -1;
}

Lisp_Object ns_cursor_type_to_lisp (int arg)
{
  switch (arg)
    {
    case bar: return intern("bar");
    case filled_box: return Qbox;
    case hollow_box: return intern("hollow");
    case line: return intern("line");
    case no_highlight:
    default: return intern("no");
    }
}

void ns_set_cursor_type (struct frame *f, Lisp_Object arg,
                         Lisp_Object oldval)
{
  int val;
  
  val=ns_lisp_to_cursor_type(arg);
  if (val>=0)
    {
      f->output_data.ns->desired_cursor=val;
    }
  else
    {
      store_frame_param(f,Qcursor_type,oldval);
      error ("the `cursor-type' frame parameter should be either `no', `bar', `box', `hollow' or `line'.");
    }
  
  update_mode_lines++;
}

void ns_set_autoraise (struct frame *f, Lisp_Object arg, Lisp_Object oldval)
{
  f->auto_raise = !NILP(arg);
}

void ns_set_autolower (struct frame *f, Lisp_Object arg, Lisp_Object oldval)
{
  f->auto_lower = !NILP(arg);
}

void ns_set_unsplittable (struct frame *f, Lisp_Object arg,
                          Lisp_Object oldval)
{
  f->no_split = !NILP (arg);
}

void ns_set_vertical_scroll_bars (struct frame *f, Lisp_Object arg,
                                  Lisp_Object oldval)
{
  if ((EQ (arg, Qleft) && FRAME_HAS_VERTICAL_SCROLL_BARS_ON_RIGHT (f))
      || (EQ (arg, Qright) && FRAME_HAS_VERTICAL_SCROLL_BARS_ON_LEFT (f))
      || (NILP (arg) && FRAME_HAS_VERTICAL_SCROLL_BARS (f))
      || (!NILP (arg) && ! FRAME_HAS_VERTICAL_SCROLL_BARS (f)))
    {
      FRAME_VERTICAL_SCROLL_BAR_TYPE (f)
#if defined(MACOSX)
        // Aqua guidelines mandate scroll bar on the right
        = (NILP (arg)
           ? vertical_scroll_bar_none
           : vertical_scroll_bar_right);
#else
      = (NILP (arg)
         ? vertical_scroll_bar_none
         : EQ (Qright, arg)
         ? vertical_scroll_bar_right 
         : vertical_scroll_bar_left);
#endif  // MACOSX

      /* We set this parameter before creating the X window for the
         frame, so we can get the geometry right from the start.
         However, if the window hasn't been created yet, we shouldn't
         call x_set_window_size.  */
      if (FRAME_NS_VIEW (f))
        x_set_window_size (f, 0, FRAME_WIDTH (f), FRAME_HEIGHT (f));
    }
}

void ns_set_scroll_bar_width (struct frame *f, Lisp_Object arg,
                              Lisp_Object oldval)
{
  int wid = NS_FONT_WIDTH(f);
  int sbwid = NS_SCROLL_BAR_WIDTH_DEFAULT
#ifdef GNUSTEP
    - 1 // right side of sb is white, so ignore it
#endif
    ;
  if (!NILP(arg) && INTEGERP(arg) && XINT(arg) > 0)
    sbwid = XFASTINT(arg);

  FRAME_SCROLL_BAR_PIXEL_WIDTH (f) = sbwid;
  FRAME_SCROLL_BAR_COLS (f) = (sbwid + wid-1) / wid;
  if (FRAME_NS_VIEW (f))
    x_set_window_size (f, 0, FRAME_WIDTH (f), FRAME_HEIGHT (f));
}

void ns_set_font (struct frame *f, Lisp_Object arg, Lisp_Object oldval)
{
  Lisp_Object frame;

  if (XTYPE(arg) == Lisp_Symbol)
    {
      arg=Fintern(arg, Qnil);
      store_frame_param (f, Qfont, arg);
    }
  else if (XTYPE(arg) != Lisp_String)
    {
      store_frame_param (f, Qfont, oldval);
      error ("Font not a string or symbol.");
    }
     
  if (FRAME_NS_VIEW(f))
    {
      if (ns_new_font(f, XSTRING(arg)->data) == Qnil)
        {
          store_frame_param (f, Qfont, oldval);
          error("Not an available fixed-width font.");
        }
      ns_recompute_basic_faces (f);
    }

  XSETFRAME (frame, f);
  call1 (Qface_set_after_frame_default, frame);
  SET_FRAME_GARBAGED (f);
}

void ns_set_fontsize (struct frame *f, Lisp_Object arg, Lisp_Object oldval)
{
  Lisp_Object frame;

  if (XTYPE(arg)==Lisp_String)
    arg=Fstring_to_number(arg, Qnil);
  if (!NUMBERP(arg) || (XFLOATINT(arg)<=0.0))
    error ("Fontsize not a positive number.");

  store_frame_param (f, Qfontsize, arg);

  if (FRAME_NS_VIEW(f))
      ns_recompute_basic_faces (f);

  XSETFRAME (frame, f);
  call1 (Qface_set_after_frame_default, frame);
  SET_FRAME_GARBAGED (f);
}

void ns_set_underline (struct frame *f, Lisp_Object arg, Lisp_Object oldval)
{
  store_frame_param (f, Qunderline, arg);
  if (FRAME_NS_VIEW(f))
      ns_recompute_basic_faces (f);

  SET_FRAME_GARBAGED (f);
}

struct ns_frame_parm_table
{
  char *name;
  void (*setter)(struct frame *f, Lisp_Object arg, Lisp_Object oldval);
};

static struct ns_frame_parm_table ns_frame_parms[] =
{
  //FIXME/cl
  { "auto-raise", ns_set_autoraise } ,
  { "auto-lower", ns_set_autolower } ,
  { "background-color", ns_set_background_color } ,
  //  { "border-color", x_set_border_color } ,
  //  { "border-width", x_set_border_width } ,
  { "cursor-color", ns_set_cursor_color } ,
  { "cursor-type", ns_set_cursor_type } ,
  { "doc-edited", ns_set_doc_edited } ,
  { "font", ns_set_font } ,
  { "fontsize", ns_set_fontsize } ,
  { "foreground-color", ns_set_foreground_color } ,
  { "icon-name", ns_set_icon_name } ,
  { "icon-type", ns_set_icon_type } ,
  { "internal-border-width", ns_set_internal_border_width } ,
  { "menu-bar-lines", x_set_menu_bar_lines } ,
  //  { "mouse-color", x_set_mouse_color } ,
  { "name", x_explicitly_set_name } ,
  { "scroll-bar-width", ns_set_scroll_bar_width } ,
  //  { "title", ns_set_title } ,
  { "underline", ns_set_underline } ,
  { "unsplittable", ns_set_unsplittable } ,
  { "vertical-scroll-bars", ns_set_vertical_scroll_bars } ,
  { "visibility", ns_set_visibility }
};

void x_set_frame_parameters (struct frame *f, Lisp_Object alist)
{
  Lisp_Object tail;

  /* If both of these parameters are present, it's more efficient to
     set them both at once.  So we wait until we've looked at the
     entire list before we set them.  */
  Lisp_Object width, height;

  /* Same here.  */
  Lisp_Object left, top;

  /* Same with these.  */
  Lisp_Object icon_left, icon_top;

  /* Record in these vectors all the parms specified.  */
  Lisp_Object *parms;
  Lisp_Object *values;
  int i;

  i = 0;
  for (tail = alist; CONSP (tail); tail = Fcdr (tail))
    i++;

  parms  = (Lisp_Object *) alloca (i * sizeof (Lisp_Object));
  values = (Lisp_Object *) alloca (i * sizeof (Lisp_Object));

  /* Extract parm names and values into those vectors.  */

  i = 0;
  for (tail = alist; CONSP (tail); tail = Fcdr (tail))
    {
      Lisp_Object elt;

      elt = Fcar (tail);
      parms[i] = Fcar (elt);
      values[i] = Fcdr (elt);
      //fprintf(stderr,"Setting '%s'\n", XSYMBOL(parms[i])->name->data);
      i++;
    }

  width = height = top = left = Qunbound;
  icon_left = icon_top = Qunbound;

  /* Now process them in reverse of specified order.  */
  for (i--; i >= 0; i--)
    {
      Lisp_Object prop, val;

      prop = parms[i];
      val = values[i];

      if (EQ (prop, Qwidth))
        width = (XTYPE(val)==Lisp_String) ? Fstring_to_number(val, Qnil) :val;
      else if (EQ (prop, Qheight))
        height = (XTYPE(val)==Lisp_String) ? Fstring_to_number(val, Qnil) : val;
      else if (EQ (prop, Qtop))
        top = (XTYPE(val)==Lisp_String) ? Fstring_to_number(val, Qnil) : val;
      else if (EQ (prop, Qleft))
        left = (XTYPE(val)==Lisp_String) ? Fstring_to_number(val, Qnil) : val;
      else if (EQ (prop, Qicon_top))
        icon_top = (XTYPE(val)==Lisp_String) ? Fstring_to_number(val, Qnil) : val;
      else if (EQ (prop, Qicon_left))
        icon_left = (XTYPE(val)==Lisp_String) ? Fstring_to_number(val, Qnil) : val;
      else
        {
          Lisp_Object param_index, old_value;

          param_index = Fget (prop, Qns_frame_parameter);
          old_value = get_frame_param (f, prop);

          store_frame_param (f, prop, val);
          if (XTYPE (param_index) == Lisp_Int
              && XINT (param_index) >= 0
              && (XINT (param_index)
                  < sizeof (ns_frame_parms)/sizeof (ns_frame_parms[0])))
            (*ns_frame_parms[XINT (param_index)].setter)(f, val, old_value);
        }
    }

  /* Don't die if just one of these was set.  */
  if (EQ (left, Qunbound))
    XSETINT (left, (int)f->output_data.ns->left);
  if (EQ (top, Qunbound))
    XSETINT (top, (int)f->output_data.ns->top);

  /* Don't die if just one of these was set.  */
  if (EQ (icon_left, Qunbound))
    icon_left = f->output_data.ns->icon_left;
  if (EQ (icon_top, Qunbound))
    icon_top = f->output_data.ns->icon_top;

  /* Don't die if just one of these was set.  */
  if (EQ (width, Qunbound))
    {
      if (FRAME_NEW_WIDTH (f))
        XSETINT (width, FRAME_NEW_WIDTH (f));
      else
        XSETINT (width, FRAME_WIDTH (f));
    }
  if (EQ (height, Qunbound))
    {
      if (FRAME_NEW_HEIGHT (f))
        XSETINT (height, FRAME_NEW_HEIGHT (f));
      else
        XSETINT (height, FRAME_HEIGHT (f));
    }

  /* Don't set these parameters these unless they've been explicitly
     specified.  The window might be mapped or resized while we're in
     this function, and we don't want to override that unless the lisp
     code has asked for it.

     Don't set these parameters unless they actually differ from the
     window's current parameters; the window may not actually exist
     yet.  */
  {
    Lisp_Object frame;

    XSETFRAME (frame, f);

    if (!INTEGERP (width) || !INTEGERP (height)
        || (INTEGERP (width) && XINT (width) != FRAME_WIDTH (f))
        || (INTEGERP (height) && XINT (height) != FRAME_HEIGHT (f))
        || FRAME_NEW_HEIGHT (f) || FRAME_NEW_WIDTH (f))
      Fset_frame_size (frame, width, height);
    if (!INTEGERP (left) || !INTEGERP (top)
        || (INTEGERP (left) && XINT (left) != f->output_data.ns->left)
        || (INTEGERP (top) && XINT (top) != f->output_data.ns->top))
      Fset_frame_position (frame, left, top);
    if ((! EQ (icon_left, f->output_data.ns->icon_left))
        || (! EQ (icon_top, f->output_data.ns->icon_top)))
      {
        NSWindow *window = [FRAME_NS_VIEW(f) window];
        NSScreen *screen = [window screen];
          
        f->output_data.ns->icon_top = icon_top;
        f->output_data.ns->icon_left = icon_left;
        if (NUMBERP (icon_top) && NUMBERP (icon_left))
          if ([window isMiniaturized])
            [window setFrameTopLeftPoint:
                      NSMakePoint(SCREENMAXBOUND(QXFLOAT (icon_left)),
                                  SCREENMAXBOUND([screen frame].size.height
                                  - QXFLOAT (icon_top)))];
      }
  }
}

void x_report_frame_params (struct frame *f, Lisp_Object *alistptr)
{
  store_in_alist (alistptr, Qicon_left, f->output_data.ns->icon_left);
  store_in_alist (alistptr, Qicon_top, f->output_data.ns->icon_top);
  store_in_alist (alistptr, Qleft, make_number ((int) f->output_data.ns->left));
  store_in_alist (alistptr, Qtop, make_number ((int) f->output_data.ns->top));
  store_in_alist (alistptr, Qinternal_border_width,
                  make_number ((int) f->output_data.ns->internal_border_width));
  store_in_alist (alistptr, Qicon_name, f->icon_name);
  FRAME_SAMPLE_VISIBILITY (f);
  store_in_alist (alistptr, Qvisibility,
                  (FRAME_VISIBLE_P (f) ? Qt
                   : FRAME_ICONIFIED_P (f) ? Qicon : Qnil));
  store_in_alist (alistptr, Qdisplay,
                  XCONS (FRAME_NS_DISPLAY_INFO (f)->name_list_element)->car);
  store_in_alist (alistptr, Qmenu_bar_lines, FRAME_EXTERNAL_MENU_BAR(f));
}

DEFUN ("ns-popup-prefs-panel", Fns_popup_prefs_panel, Sns_popup_prefs_panel, 0,0,"",
       "Pop up the preferences panel.")
     ()
{
  check_ns();
  [NSApp showPreferencesWindow:NSApp];
  return Qnil;
}

DEFUN ("ns-popup-font-panel", Fns_popup_font_panel, Sns_popup_font_panel,
       0,1,"", "Pop up the font panel.")
     (frame)
     Lisp_Object frame;
{
  id fm;
  struct frame *f;

  check_ns();
  fm = [NSFontManager new];
  if (NILP(frame))
    f = selected_frame;
  else
    {
      CHECK_FRAME(frame, 0);
      f = XFRAME (frame);
    }

  [fm setSelectedFont:f->output_data.ns->font->nsfont isMultiple:NO];
  [fm orderFrontFontPanel:NSApp];
  return Qnil;
}

DEFUN ("ns-popup-color-panel", Fns_popup_color_panel, Sns_popup_color_panel, 0,1,"",
       "Pop up the color panel.")
     (frame)
     Lisp_Object frame;
{
  struct frame *f;

  check_ns();
  if (NILP(frame))
    f=selected_frame;
  else
    {
      CHECK_FRAME(frame, 0);
      f=XFRAME (frame);
    }

  [NSApp orderFrontColorPanel:NSApp];
  return Qnil;
}

DEFUN ("ns-read-file-name", Fns_read_file_name, Sns_read_file_name, 1, 4, 0,
       "As read-file-name except that NS panels are used for querying, and args slightly different.  Nil returned if no selection made.  Set isLoad non-nil if file being read for a save.")
     (prompt, dir, isLoad, init)
     Lisp_Object prompt, dir, isLoad, init;
{
  static id fileDelegate = nil;
  int ret;
  id panel;
  NSString *fname;

  NSString *promptS = NILP(prompt) || !STRINGP(prompt) ? nil :
    [NSString stringWithUTF8String: XSTRING(prompt)->data];
  NSString *dirS = NILP(dir) || !STRINGP(dir) ?
    [NSString stringWithUTF8String: XSTRING(current_buffer->directory)->data] :
    [NSString stringWithUTF8String: XSTRING(dir)->data];
  NSString *initS = NILP(init) || !STRINGP(init) ? nil :
    [NSString stringWithUTF8String: XSTRING(init)->data];

  check_ns();

  if (fileDelegate == nil)
    fileDelegate = [EmacsFileDelegate new];

  [NSCursor setHiddenUntilMouseMoves: NO];

  if ([dirS characterAtIndex: 0] == '~')
    dirS = [dirS stringByExpandingTildeInPath];

  panel = NILP(isLoad) ?
    [EmacsSavePanel savePanel] : [EmacsOpenPanel openPanel];

  [panel setTitle: promptS];

  // Puma (10.1) does not have
  if ([panel respondsToSelector: @selector(setAllowsOtherFileTypes:)])
    [panel setAllowsOtherFileTypes: YES];

  [panel setTreatsFilePackagesAsDirectories:YES];
  [panel setDelegate: fileDelegate];

  panelOK = 0;
  if (NILP(isLoad))
    {
      ret = [panel runModalForDirectory:dirS file:initS];
    }
  else
    {
      [panel setCanChooseDirectories: YES];
      ret = [panel runModalForDirectory:dirS file:initS types:nil];
    }

  ret = (ret==NSOKButton) || panelOK;

  fname = [panel filename];

  [[FRAME_NS_VIEW(selected_frame) window] makeKeyWindow];

  return ret ? build_string([fname UTF8String]) : Qnil;
}

DEFUN ("ns-yes-or-no-p", Fns_yes_or_no_p, Sns_yes_or_no_p, 1, 1, 0,
       "As yes-or-no-p except that NS panels are used for querying.")
     (prompt)
{
  int ret;
  CHECK_STRING(prompt, 0);
  ret=NSRunAlertPanel([[NSProcessInfo processInfo] processName],
                      [NSString stringWithUTF8String:XSTRING(prompt)->data],
                      @"Yes", @"No", @"Cancel");
  [[FRAME_NS_VIEW(selected_frame) window] makeKeyWindow];
  if (ret==NSAlertDefaultReturn)   return Qt;
  if (ret==NSAlertAlternateReturn) return Qnil;
  if (ret==NSAlertOtherReturn)     Fsignal(Qquit,Qnil);
  error ("NXRunAlertPanel returns invalid.");
  return Qnil;
}

DEFUN ("ns-get-resource", Fns_get_resource, Sns_get_resource, 2, 2, 0,
       "Return the value of the property NAME of OWNER from the defaults database.\n\
If OWNER is nil, Emacs is assumed.")
     (owner, name)
     Lisp_Object owner, name;
{
  const char *value;

  check_ns();
  if (NILP(owner))
    owner = build_string
        ([[[NSProcessInfo processInfo] processName] UTF8String]);
  CHECK_STRING(owner, 0);
  CHECK_STRING(name, 0);
  ///PENDING: owner not used?

  value=[[[NSUserDefaults standardUserDefaults]
             objectForKey:[NSString stringWithUTF8String:XSTRING(name)->data]]
            UTF8String];

  if (value)
    return build_string(value);
//fprintf(stderr,"Nothing found for NS resource '%s'.\n",XSTRING(name)->data);
  return Qnil;
}

DEFUN ("x-get-resource", Fx_get_resource, Sx_get_resource, 2, 4, 0,
       "Return the value of ATTRIBUTE, of class CLASS, from the X defaults database.\n\
This uses `INSTANCE.ATTRIBUTE' as the key and `Emacs.CLASS' as the\n\
class, where INSTANCE is the name under which Emacs was invoked, or\n\
the name specified by the `-name' or `-rn' command-line arguments.\n\
\n\
The optional arguments COMPONENT and SUBCLASS add to the key and the\n\
class, respectively.  You must specify both of them or neither.\n\
If you specify them, the key is `INSTANCE.COMPONENT.ATTRIBUTE'\n\
and the class is `Emacs.CLASS.SUBCLASS'.")
     (attribute, class, component, subclass)
     Lisp_Object attribute, class, component, subclass;
{
  register const char *value;

  check_ns ();

  CHECK_STRING (attribute, 0);
  CHECK_STRING (class, 0);

  if (!NILP (component))
    CHECK_STRING (component, 1);
  if (!NILP (subclass))
    CHECK_STRING (subclass, 2);
  if (NILP (component) != NILP (subclass))
    error ("x-get-resource: must specify both COMPONENT and SUBCLASS or neither");

  value=[[[NSUserDefaults standardUserDefaults] 
               objectForKey:[NSString stringWithUTF8String:
                                       XSTRING (attribute)->data]] UTF8String];


  if (value != (char *) 0)
    return build_string (value);
//fprintf(stderr,"Nothing found for X resource '%s'.\n",XSTRING(attribute)->data);
  return Qnil;
}

DEFUN ("ns-set-resource", Fns_set_resource, Sns_set_resource, 3, 3, 0,
       "Set property NAME of OWNER to VALUE, from the defaults database.\n\
If OWNER is nil, Emacs is assumed.\n\
If VALUE is nil, the default is removed.")
     (owner, name, value)
     Lisp_Object owner, name, value;
{
  check_ns();
  if (NILP(owner))
    owner =
        build_string([[[NSProcessInfo processInfo] processName] UTF8String]);
  CHECK_STRING(owner, 0);
  CHECK_STRING(name, 0);
  if (NILP(value))
    {
      [[NSUserDefaults standardUserDefaults] removeObjectForKey:
                          [NSString stringWithUTF8String:XSTRING(name)->data]];
    }
  else
    {
      CHECK_STRING(value, 0);
      [[NSUserDefaults standardUserDefaults] setObject:
                [NSString stringWithUTF8String:XSTRING(value)->data]
                                        forKey:[NSString stringWithUTF8String:
                                                         XSTRING(name)->data]];
    }

  return Qnil;
}

static void
ns_icon (f, parms)
     struct frame *f;
     Lisp_Object parms;
{
  Lisp_Object icon_x, icon_y;

  f->output_data.ns->icon_top = Qnil;
  f->output_data.ns->icon_left = Qnil;

  /* Set the position of the icon.  */
  icon_x = ns_get_arg (parms, Qicon_left, 0, 0);
  icon_y = ns_get_arg (parms, Qicon_top, 0, 0);
  if (!EQ (icon_x, Qunbound) && !EQ (icon_y, Qunbound))
    {
      CHECK_NUMBER (icon_x, 0);
      CHECK_NUMBER (icon_y, 0);
      f->output_data.ns->icon_top = icon_y;
      f->output_data.ns->icon_left = icon_x;
    }
  else if (!EQ (icon_x, Qunbound) || !EQ (icon_y, Qunbound))
    error ("Both left and top icon corners of icon must be specified");
}

DEFUN ("x-create-frame", Fns_create_frame, Sns_create_frame,
       1, 1, 0,
       "Make a new NS window, which is called a \"frame\" in Emacs terms.\n\
Return an Emacs frame object representing the X window.\n\
ALIST is an alist of frame parameters.\n\
If the parameters specify that the frame should not have a minibuffer,\n\
and do not specify a specific minibuffer window to use,\n\
then `default-minibuffer-frame' must be a frame whose minibuffer can\n\
be shared by the new frame.")
     (parms)
     Lisp_Object parms;
{
  struct frame *f;
  Lisp_Object frame, tem;
  Lisp_Object name;
  int minibuffer_only = 0;
  int count = specpdl_ptr - specpdl;
  Lisp_Object display;
  struct ns_display_info *dpyinfo;
  struct kboard *kb;
  Lisp_Object tfont, tfontsize;
  Lisp_Object accum = Qnil;

  check_ns();

  display = ns_get_arg (parms, Qdisplay, 0, 0);
  if (EQ (display, Qunbound))
    display = Qnil;
  dpyinfo = check_ns_display_info (display);
#ifdef MULTI_KBOARD
  kb = dpyinfo->kboard;
#else
  kb = &the_only_kboard;
#endif

  name = ns_get_arg (parms, Qname, 0, 0);
  if (!STRINGP (name)
      && ! EQ (name, Qunbound)
      && ! NILP (name))
    error ("Invalid frame name--not a string or nil");

  tem = ns_get_arg(parms, Qminibuffer, 0, 0);
  if (XTYPE(tem)==Lisp_String) tem=Fintern(tem, Qnil);

  if (EQ (tem,Qnone) || NILP(tem))
    {
      f= make_frame_without_minibuffer(Qnil, current_kboard, display);
    }
  else if (EQ (tem, Qonly))
    {
      f= make_minibuffer_frame ();
      minibuffer_only =1;
    }
  else if (WINDOWP (tem))
    {
      f= make_frame_without_minibuffer(tem, current_kboard, display);
    }
  else
    {
      f= make_frame (1);
    }

  FRAME_CAN_HAVE_SCROLL_BARS (f) = 1;

  XSETFRAME (frame, f);
  f->output_method = output_ns;
  f->output_data.ns = (struct ns_output *) xmalloc (sizeof *(f->output_data.ns));
  bzero (f->output_data.ns, sizeof (*(f->output_data.ns)));

  f->icon_name = ns_get_arg (parms, Qicon_name, 0, 0);
  if (EQ (f->icon_name, Qunbound) || (XTYPE (f->icon_name) != Lisp_String))
    f->icon_name = Qnil;
   
  FRAME_NS_DISPLAY_INFO (f) = dpyinfo;
#ifdef MULTI_KBOARD
  FRAME_KBOARD (f) = kb;
#endif

  /* Note that the frame has no physical cursor right now.  */
  f->phys_cursor_x = -1;
  f->phys_cursor_y = -1;

  /* Set the name; the functions to which we pass f expect the name to
     be set.  */
  if (EQ (name, Qunbound) || NILP (name) || (XTYPE (name) != Lisp_String))
    {
      f->name =
          build_string([[[NSProcessInfo processInfo] processName] UTF8String]);
      f->explicit_name=0;
    }
  else
    {
      f->name = name;
      f->explicit_name=1;
      /* use the frame's title when getting resources for this frame.  */
      specbind (Qx_resource_name, name);
    }

  ns_default_parameter(f, parms, Qinternal_border_width, make_number(2),
                       0, "InternalBorderWidth", &accum);
  ns_default_parameter(f, parms, Qvertical_scroll_bars, Qt,
                       0, "VerticalScrollBars", &accum);
  ns_default_parameter(f, parms, Qicon_type, Qnil, 0, "BitmapIcon", &accum);
  ns_default_parameter(f, parms, Qauto_raise, Qnil,0, "AutoRaise", &accum);
  ns_default_parameter(f, parms, Qauto_lower, Qnil,0, "AutoLower", &accum);
  ns_default_parameter(f, parms, Qbuffered, Qt, 0, "Buffered", &accum);
  {
    // use for default font name
    id font=[NSFont userFixedPitchFontOfSize:-1.0];
    tfont = ns_default_parameter(f, parms, Qfont,
                                 build_string([[font fontName] UTF8String]),
                                 0, "Font", &accum);
    ns_default_parameter(f, parms, Qforeground_color, build_string("Black"),
                         0, "Foreground", &accum);
    ns_default_parameter(f, parms, Qbackground_color, build_string("White"),
                         0, "Background", &accum);
    tfontsize = ns_default_parameter(f, parms, Qfontsize, make_number((int)[font pointSize]),
                                     0, "FontSize", &accum);
    ns_default_parameter(f, parms, Qunderline, Qnil,
                         0, "Underline", &accum);
    x_set_frame_parameters(f, accum);
    accum = Qnil;

    ns_new_font(f, XSTRING(tfont)->data);
    // (font size is picked up in ns_load_font)
  }
  ns_default_parameter(f, parms, Qheight, make_number(48), 0, "Height",
                       &accum);
  ns_default_parameter(f, parms, Qwidth, make_number(80), 0, "Width", &accum);
  ns_default_parameter(f, parms, Qcursor_color, build_string("grey"),
                       0, "CursorColor", &accum);
  ns_default_parameter(f, parms, Qcursor_type, Qbox, 0, "CursorType", &accum);
  ns_default_parameter(f, parms, Qtop, make_number(20), 0, "Top", &accum);
  ns_default_parameter(f, parms, Qleft, make_number(20), 0, "Left", &accum);
  ns_default_parameter(f, parms, Qmenu_bar_lines, make_number(1), 0, "Menus",
                       &accum);
  ns_default_parameter(f, parms, Qscroll_bar_width, Qnil, 0,
                       "ScrollBarWidth", &accum);
  ns_default_parameter (f, parms, Qbuffer_predicate, Qnil, 0,
                        "BufferPredicate", &accum);
  ns_default_parameter (f, parms, Qtitle, Qnil, 0, "Title", &accum);

  x_set_frame_parameters(f, accum);

  tem= ns_get_arg (parms, Qunsplittable, 0, 0);
  f->no_split = minibuffer_only || (!EQ (tem, Qunbound) && !EQ (tem, Qnil));

  ns_icon (f, parms);
  ns_init_frame_faces(f);

  [[EmacsView alloc] initFrameFromEmacs:f];

  /* It is now ok to make the frame official even if we get an error below.
     The frame needs to be on Vframe_list or making it visible won't work. */
  Vframe_list = Fcons (frame, Vframe_list);

  tem=ns_get_arg(parms, Qvisibility,0,0);
  if (EQ (tem, Qunbound))
    tem = Qnil;

  ns_set_visibility(f,tem,Qnil);
  if (EQ (tem, Qt))
    [[FRAME_NS_VIEW(f) window] makeKeyWindow];

  return unbind_to (count, frame);
}

Lisp_Object x_get_focus_frame (frame)
     struct frame *frame;
{
  Lisp_Object nsfocus;

  if (!ns_focus_frame)
    return Qnil;

  XSETFRAME (nsfocus, ns_focus_frame);
  return nsfocus;
}

DEFUN ("ns-make-key-frame", Fns_make_key_frame, Sns_make_key_frame, 0, 1, 0,
       "Make FRAME the OpenStep key window.\n\
If FRAME is omitted, make the currently selected frame the OpenStep\n\
key window.")
     (frame)
     Lisp_Object frame;
{
  FRAME_PTR f = check_ns_frame (frame);
  NSView *view = FRAME_NS_VIEW(f);
  [[view window] makeKeyWindow];
  return Qnil;
}


DEFUN ("ns-list-fonts", Fns_list_fonts, Sns_list_fonts, 1, 4, 0,
       "Return a list of the names of available fonts matching PATTERN.\n\
If optional arguments FACE and FRAME are specified, return only fonts\n\
the same size as FACE on FRAME.\n\
If optional argument MAX is specified, return at most MAX matches.\n\
\n\
PATTERN is a regular expression; FACE is a face name - a symbol.\n\
\n\
The return value is a list of strings, suitable as arguments to\n\
set-face-font.\n\
\n\
Fonts Emacs can't use (i.e. proportional fonts) may or may not be excluded\n\
even if they match PATTERN and FACE.")
     (pattern, face, frame, max)
     Lisp_Object pattern, face, frame, max;
{
  int maxnames;
  FRAME_PTR f;
  Lisp_Object list = Qnil,
    patterns,
    rpattern,
    key,
    tem,
    args[3];
  struct re_pattern_buffer *bufp;
  id fm = [NSFontManager new];

  TRACE(ns_list_fonts);
//fprintf(stderr,"ns-list_fonts: '%s'\n",XSTRING(pattern)->data);  

  if (NILP (max))
    maxnames = 4;
  else
    {
      CHECK_NATNUM (max, 0);
      maxnames = XINT (max);
    }

  /* We can't simply call check_x_frame because this function may be
     called before any frame is created.  */
  if (NILP (frame))
    f = selected_frame;
  else
    {
      CHECK_LIVE_FRAME (frame, 0);
      f = XFRAME (frame);
    }
  if (! FRAME_WINDOW_P (f))
    {
      /* Perhaps we have not yet created any frame.  */
      f = NULL;
    }

  patterns = Fcons (pattern, Qnil); // support multiple patterns for future
  for (; CONSP (patterns); patterns = XCONS (patterns)->cdr)
    {
      NSEnumerator *fenum;
      NSString *fontname;
      BOOL synthItal = NO;
      
      pattern = XCONS (patterns)->car;
      key = f ? Fcons (pattern, make_number (maxnames)) : Qnil;
      tem = f ? XCONS (FRAME_NS_DISPLAY_INFO (f)->name_list_element)->cdr
        : Qnil;

      /* See if we cached the result for this particular query.
         The cache is an alist of the form:
         ((((PATTERN . MAXNAMES) FONTNAME) ...) ...)
      */
      if (f && !NILP (list = Fassoc (key, tem)))
        {
          list = Fcdr_safe (list);
          /* We have a cached list.  Don't have to get the list again.  */
          if (NILP(list))
            continue;
          else
            return list;
        }

      CHECK_STRING(pattern,0);
      args[0]=build_string("^");
      args[1]=pattern;
      args[2]=build_string("$");
      rpattern = Fconcat(3, args);
      bufp = compile_pattern (rpattern, 0, 0, 0);

      if (strstr(XSTRING(pattern)->data, "-synthItal")) synthItal = YES;

      list = Qnil;
      fenum = [[fm availableFonts] objectEnumerator];
      while ( (fontname = [fenum nextObject]) )
        {
          const char *fname = synthItal
            ? [[fontname stringByAppendingString: @"-synthItal"] UTF8String]
            : [fontname UTF8String];
          int len = strlen(fname);
          if (re_search(bufp, fname, len, 0, len, 0)>=0)
            list = Fcons(build_string (fname),list);
        }
      
      // store result in cache
      if (f != NULL)
        XCONS (FRAME_NS_DISPLAY_INFO (f)->name_list_element)->cdr
          = Fcons (Fcons (key, list),
                   XCONS (FRAME_NS_DISPLAY_INFO (f)->name_list_element)->cdr);
    }

    return list;
}

/* This is a version of the above for use by the fontset system.
   In other GUIs it is defined separately in *term.c, but we're just
   going to reuse the implementation above, particularly as we don't
   implement fontsets fully yet.  Things are simpler anyway since NS
   handles font size stuff internally -- we can request any font in
   any size.

   Note: The size parameter is documented (for X, etc.) as: If SIZE is not
         0, it is the size (maximum bound width) of fonts to be listed.
 */
Lisp_Object
ns_list_fonts (FRAME_PTR f, Lisp_Object pattern, int size, int maxnames)
{
    Lisp_Object frame;
    VOID_TO_LISP(frame, f);
    return Fns_list_fonts(pattern, Qnil, frame, make_number(maxnames));
}


DEFUN ("ns-list-colors", Fns_list_colors, Sns_list_colors, 0, 1, 0,
       "Return a list of all available colors.\n\
The optional argument FRAME is currently ignored.")
     (frame)
     Lisp_Object frame;
{
  Lisp_Object list = Qnil;
  NSArray *colorlistlist = [NSColorList availableColorLists];
  int l, n = [colorlistlist count];
  int i;
  if (!NILP(frame))
    {
      CHECK_FRAME(frame,0);
      if (! FRAME_NS_P (XFRAME(frame)))
        error ("non-NS frame used in `ns-list-fonts'");
    }

  for (l = 0; l < n; l++)
    {
      NSColorList *clist = [colorlistlist objectAtIndex:l];
      if ([[clist name] length] < 7 ||
          [[clist name] rangeOfString: @"PANTONE"].location == 0)
        {
          for (i = [[clist allKeys] count] - 1; i >= 0; i--)
            list = Fcons(build_string([[[clist allKeys] objectAtIndex:i]
                                          UTF8String]), list);
        }
    }
  return list;
}

DEFUN ("x-color-defined-p", Fns_color_defined_p, Sns_color_defined_p, 1, 2, 0,
       "Return t if the current NS display supports the color named COLOR.\n\
The optional argument FRAME is currently ignored.")
     (color, frame)
     Lisp_Object color, frame;
{
  NSColor * col;
  check_ns();
  return ns_lisp_to_color(color, &col) ? Qnil : Qt;
}

DEFUN ("x-color-values", Fns_color_values, Sns_color_values, 1, 2, 0,
       "Return a description of the color named COLOR.\n\
The value is a list of integer RGBA values--(RED GREEN BLUE ALPHA).\n\
These values appear to range from 0 to 65280; white is (65280 65280 65280 0).\n\
The optional argument FRAME is currently ignored.")
     (color, frame)
     Lisp_Object color, frame;
{
  NSColor * col;
  float red,green,blue,alpha;
  Lisp_Object rgba[4];

  check_ns ();
  CHECK_STRING (color, 0);
   
  if (ns_lisp_to_color(color, &col))
    return Qnil;

  [[col colorUsingColorSpaceName:NSCalibratedRGBColorSpace] getRed:&red green:&green blue:&blue alpha:&alpha];
  rgba[0] = make_number ((int) rint(red*65280));
  rgba[1] = make_number ((int) rint(green*65280));
  rgba[2] = make_number ((int) rint(blue*65280));
  rgba[3] = make_number ((int) rint(alpha*65280));

  return Flist (4, rgba);
}

static id ns_get_window(Lisp_Object frame)
{// FIXME/cl for multi-display
  id view=nil,window=nil;
  if (!FRAMEP(frame) || !FRAME_NS_P(XFRAME(frame)))
    frame = wrong_type_argument (Qframep, frame);
  view = FRAME_NS_VIEW(XFRAME(frame));
  if (view) window=[view window];
  return window;
}

static NSScreen *ns_get_screen(Lisp_Object display)
{
  id window=nil;
  NSScreen *screen = 0;

  if (STRINGP(display)) // FIXME/cl for multi-display
    display = Qnil;
  if (FRAMEP(display) && FRAME_NS_P(XFRAME(display)))
    window = ns_get_window(display);
  if (window != nil) screen =[window screen];
  if (!screen) screen = [NSScreen mainScreen];
  return screen;
}

DEFUN ("x-display-color-p", Fns_display_color_p, Sns_display_color_p, 0, 1, 0,
       "Return t if the NS display supports color.\n\
The optional argument DISPLAY specifies which display to ask about.\n\
DISPLAY should be either a frame or a display name (a string).\n\
If omitted or nil, that stands for the selected frame's display.")
     (display)
     Lisp_Object display;
{
  NSWindowDepth depth;
  NSString *colorSpace;
  check_ns();
  depth = [ns_get_screen(display) depth];
  colorSpace = NSColorSpaceFromDepth(depth);

  return    [colorSpace isEqualToString: NSDeviceWhiteColorSpace]
         || [colorSpace isEqualToString: NSDeviceWhiteColorSpace]
      ? Qnil : Qt;
}

DEFUN ("x-display-grayscale-p", Fns_display_grayscale_p,
       Sns_display_grayscale_p, 0, 1, 0,
       "Return t if the NS display supports shades of gray.\n\
Note that color displays do support shades of gray.\n\
The optional argument DISPLAY specifies which display to ask about.\n\
DISPLAY should be either a frame or a display name (a string).\n\
If omitted or nil, that stands for the selected frame's display.")
     (display)
     Lisp_Object display;
{
  NSWindowDepth depth;
  check_ns();
  depth = [ns_get_screen(display) depth];

  return NSBitsPerPixelFromDepth(depth) > 1 ? Qt : Qnil;
}

DEFUN ("x-display-pixel-width", Fns_display_pixel_width, Sns_display_pixel_width,
       0, 1, 0,
       "Returns the width in pixels of the NS display DISPLAY.\n\
The optional argument DISPLAY specifies which display to ask about.\n\
DISPLAY should be either a frame or a display name (a string).\n\
If omitted or nil, that stands for the selected frame's display.")
     (display)
     Lisp_Object display;
{
  check_ns();
  return make_number ((int) [ns_get_screen(display) frame].size.width);
}

DEFUN ("x-display-pixel-height", Fns_display_pixel_height,
       Sns_display_pixel_height, 0, 1, 0,
       "Returns the height in pixels of the NS display DISPLAY.\n\
The optional argument DISPLAY specifies which display to ask about.\n\
DISPLAY should be either a frame or a display name (a string).\n\
If omitted or nil, that stands for the selected frame's display.")
     (display)
     Lisp_Object display;
{
  check_ns();
  return make_number ((int) [ns_get_screen(display) frame].size.height);
}

DEFUN ("ns-display-planes", Fns_display_planes, Sns_display_planes,
       0, 1, 0,
       "Returns the number of bitplanes of the NS display DISPLAY.\n\
The optional argument DISPLAY specifies which display to ask about.\n\
DISPLAY should be either a frame or a display name (a string).\n\
If omitted or nil, that stands for the selected frame's display.")
     (display)
     Lisp_Object display;
{
  check_ns();
  return make_number (NSBitsPerSampleFromDepth ([ns_get_screen(display) depth]));
}

DEFUN ("ns-display-color-cells", Fns_display_color_cells, Sns_display_color_cells,
       0, 1, 0,
       "Returns the number of color cells of the NS display DISPLAY.\n\
The optional argument DISPLAY specifies which display to ask about.\n\
DISPLAY should be either a frame or a display name (a string).\n\
If omitted or nil, that stands for the selected frame's display.")
     (display)
     Lisp_Object display;
{
  check_ns();
  return make_number (1<<NSBitsPerSampleFromDepth ([ns_get_screen(display) depth]));
}

DEFUN ("ns-server-max-request-size", Fns_server_max_request_size,
       Sns_server_max_request_size,
       0, 1, 0,
       "This function is only present for completeness.  It does not return\n\
a usable result for NS windows.")
     (display)
     Lisp_Object display;
{
  check_ns();
  /* This function has no real equivalent under NeXTstep.  Return nil to
     indicate this. */
  return Qnil;
}

DEFUN ("ns-server-vendor", Fns_server_vendor, Sns_server_vendor, 0, 1, 0,
       "Returns the vendor ID string of the NS server of display DISPLAY.\n\
The optional argument DISPLAY specifies which display to ask about.\n\
DISPLAY should be either a frame or a display name (a string).\n\
If omitted or nil, that stands for the selected frame's display.")
     (display)
     Lisp_Object display;
{
  check_ns();
#ifdef GNUSTEP
  return build_string ("GNU");
#else
  return build_string ("Apple");
#endif
}

DEFUN ("ns-server-version", Fns_server_version, Sns_server_version, 0, 1, 0,
       "Returns the version number of the NS release of display DISPLAY.\n\
See also the function `ns-server-vendor'.\n\n\
The optional argument DISPLAY specifies which display to ask about.\n\
DISPLAY should be either a frame or a display name (a string).\n\
If omitted or nil, that stands for the selected frame's display.")
     (display)
     Lisp_Object display;
{
  //PENDING: return GUI version on GNUSTEP, ?? on OS X
  return build_string("1.0");
}

DEFUN ("ns-display-screens", Fns_display_screens, Sns_display_screens, 0, 1, 0,
       "Returns the number of screens on the NS server of display DISPLAY.\n\
The optional argument DISPLAY specifies which display to ask about.\n\
DISPLAY should be either a frame or a display name (a string).\n\
If omitted or nil, that stands for the selected frame's display.")
     (display)
     Lisp_Object display;
{
  int num;

  check_ns();
  num = [[NSScreen screens] count];

  return (num!=0) ? make_number (num) : Qnil;
}

DEFUN ("ns-display-mm-height", Fns_display_mm_height, Sns_display_mm_height, 0, 1, 0,
       "Returns the height in millimeters of the NS display DISPLAY.\n\
The optional argument DISPLAY specifies which display to ask about.\n\
DISPLAY should be either a frame or a display name (a string).\n\
If omitted or nil, that stands for the selected frame's display.")
     (display)
     Lisp_Object display;
{
  check_ns();
  return make_number ((int) ([ns_get_screen(display) frame].size.height/(92.0/25.4)));
}

DEFUN ("ns-display-mm-width", Fns_display_mm_width, Sns_display_mm_width, 0, 1, 0,
       "Returns the width in millimeters of the NS display DISPLAY.\n\
The optional argument DISPLAY specifies which display to ask about.\n\
DISPLAY should be either a frame or a display name (a string).\n\
If omitted or nil, that stands for the selected frame's display.")
     (display)
     Lisp_Object display;
{
  check_ns();
  return make_number ((int) ([ns_get_screen(display) frame].size.width/(92.0/25.4)));
}

DEFUN ("ns-display-backing-store", Fns_display_backing_store,
       Sns_display_backing_store, 0, 1, 0,
       "Returns an indication of whether NS display DISPLAY does backing store.\n\
The value may be `buffered', `retained', or `non-retained'.\n\
The optional argument DISPLAY specifies which display to ask about.\n\
DISPLAY should be either a frame or a display name (a string).\n\
If omitted or nil, that stands for the selected frame's display.\n\
Under NS, this may differ for each frame.")
     (display)
     Lisp_Object display;
{
  check_ns();
  switch ([ns_get_window(display) backingType])
    {
    case NSBackingStoreBuffered:
      return intern ("buffered");
    case NSBackingStoreRetained:
      return intern ("retained");
    case NSBackingStoreNonretained:
      return intern ("non-retained");
    default:
      error ("Strange value for backingType parameter of frame");
    }
  return Qnil;  // not reached, shut compiler up
}

DEFUN ("ns-display-visual-class", Fns_display_visual_class,
       Sns_display_visual_class, 0, 1, 0,
       "Returns the visual class of the NS display DISPLAY.\n\
The value is one of the symbols `static-gray', `gray-scale',\n\
`static-color', `pseudo-color', `true-color', or `direct-color'.\n\n\
The optional argument DISPLAY specifies which display to ask about.\n\
DISPLAY should be either a frame or a display name (a string).\n\
If omitted or nil, that stands for the selected frame's display.")
     (display)
     Lisp_Object display;
{
  NSWindowDepth depth;
  check_ns();
  depth = [ns_get_screen(display) depth];

  if ( depth == NSBestDepth(NSCalibratedWhiteColorSpace, 2, 2, YES, NULL))
    return intern ("static-gray");
  else if (depth == NSBestDepth(NSCalibratedWhiteColorSpace, 8, 8, YES, NULL))
    return intern ("gray-scale");
  else if ( depth == NSBestDepth(NSCalibratedRGBColorSpace, 8, 8, YES, NULL))
    return intern ("pseudo-color");
  else if ( depth == NSBestDepth(NSCalibratedRGBColorSpace, 4, 12, NO, NULL))
    return intern ("true-color");
  else if ( depth == NSBestDepth(NSCalibratedRGBColorSpace, 8, 24, NO, NULL))
    return intern ("direct-color");
  else
    // color mgmt as far as we do it is really handled by NS itself anyway
    return intern ("direct-color");
}

DEFUN ("ns-display-save-under", Fns_display_save_under,
       Sns_display_save_under, 0, 1, 0,
       "Returns t if the NS display DISPLAY supports the save-under feature.\n\
The optional argument DISPLAY specifies which display to ask about.\n\
DISPLAY should be either a frame or a display name (a string).\n\
If omitted or nil, that stands for the selected frame's display.\n\
Under NS, this may differ for each frame.")
     (display)
     Lisp_Object display;
{
  check_ns();
  switch ([ns_get_window(display) backingType])
    {
    case NSBackingStoreBuffered:
      return Qt;

    case NSBackingStoreRetained:
    case NSBackingStoreNonretained:
      return Qnil;

    default:
      error ("Strange value for backingType parameter of frame");
    }
  return Qnil;  // not reached, shut compiler up
}

int x_pixel_width (struct frame *f)
{
  NSView *view = FRAME_NS_VIEW(f);
  NSRect r;
  r = [view bounds];
  return r.size.width+2*f->output_data.ns->internal_border_width;
}

int x_pixel_height (struct frame *f)
{
  NSView *view = FRAME_NS_VIEW(f);
  NSRect r;
  r = [view bounds];
  return r.size.height+2*f->output_data.ns->internal_border_width;
}

int x_char_width (struct frame *f)
{
  return NS_FONT_WIDTH(f);
}

int x_char_height (struct frame *f)
{
  return f->output_data.ns->line_height;
}

int x_screen_planes (struct frame *f)
{
  Lisp_Object frame;
  XSETFRAME (frame, f);

  return NSBitsPerSampleFromDepth  ([ns_get_screen (frame) depth]);
}

/* Return the X display structure for the display named NAME.
   Open a new connection if necessary.  */

struct ns_display_info *
ns_display_info_for_name (name)
     Lisp_Object name;
{
  Lisp_Object names;
  struct ns_display_info *dpyinfo;

  CHECK_STRING (name, 0);

  for (dpyinfo = ns_display_list, names = ns_display_name_list;
       dpyinfo;
       dpyinfo = dpyinfo->next, names = XCONS (names)->cdr)
    {
      Lisp_Object tem;
      tem = Fstring_equal (XCONS (XCONS (names)->car)->car, name);
      if (!NILP (tem))
        return dpyinfo;
    }

  error ("Emacs for OpenStep does not yet support multi-display.");
  
  dpyinfo = ns_term_init (name);

  if (dpyinfo == 0)
    error ("OpenStep on %s not responding.\n", XSTRING (name)->data);

  return dpyinfo;
}

DEFUN ("ns-open-connection", Fns_open_connection, Sns_open_connection,
       1, 3, 0, "Open a connection to a NS server.\n\
DISPLAY is the name of the display to connect to.\n\
Optional arguments XRM-STRING and MUST-SUCCEED are currently ignored.")
     (display, resource_string, must_succeed)
     Lisp_Object display, resource_string, must_succeed;
{
  struct ns_display_info *dpyinfo;

  CHECK_STRING (display, 0);

  nxatoms_of_nsselect();
  dpyinfo = ns_term_init(display);
  if (dpyinfo == 0)
    {
      if (!NILP (must_succeed))
        fatal ("OpenStep on %s not responding.\n",
               XSTRING (display)->data);
      else
        error ("OpenStep on %s not responding.\n",
               XSTRING (display)->data);
    }

  // Register our external input/output types, used for determining
  // applicable services and also drag/drop eligibility.
  ns_send_types = [[NSArray arrayWithObject:NSStringPboardType] retain];
  ns_return_types = [[NSArray arrayWithObject:NSStringPboardType] retain];
  ns_drag_types = [[NSArray arrayWithObjects:
                            NSStringPboardType,
                            NSTabularTextPboardType,
                            NSFilenamesPboardType,
                            NSURLPboardType,
                            NSColorPboardType,
                            NSFontPboardType, nil] retain];
  
  return Qnil;
}

DEFUN ("ns-close-connection", Fns_close_connection, Sns_close_connection,
       1, 1, 0, "Close the connection to the current NS server.\n\
The second argument DISPLAY is currently ignored, but nil would stand for\n\
the selected frame's display.")
     (display)
     Lisp_Object display;
{
  check_ns();
#ifndef GNUSTEP
  PSFlush();
#endif
  [NSApp terminate:NSApp];
  return Qnil;
}

DEFUN ("ns-display-list", Fns_display_list, Sns_display_list, 0, 0, 0,
       "Return the list of display names that Emacs has connections to.")
     ()
{
  Lisp_Object tail, result;

  result = Qnil;
  for (tail = ns_display_name_list; ! NILP (tail); tail = XCONS (tail)->cdr)
    result = Fcons (XCONS (XCONS (tail)->car)->car, result);

  return result;
}

extern void debug_log_dps_call(const char *);

DEFUN ("hide-emacs", Fns_hide_emacs, Sns_hide_emacs,
       1, 1, 0, "If ON is non-nil, the entire emacs application is hidden.\n\
Otherwise if emacs is hidden, it is unhidden.\n\
If ON is equal to 'activate, emacs is unhidden and becomes the active application.")
     (on)
     Lisp_Object on;
{
  check_ns();
  if (EQ(on,intern("activate")))
    {
      [NSApp unhide:NSApp];
      [NSApp activateIgnoringOtherApps:YES];
    }
  else if (NILP(on))
    [NSApp unhide:NSApp];
  else
    [NSApp hide:NSApp];
  return Qnil;
}


DEFUN ("emacs-info-panel", Fns_emacs_info_panel, Sns_emacs_info_panel,
       0, 0, 0, "Shows the 'Info' or 'About' panel for Emacs.")
     ()
{
  check_ns();
  [NSApp orderFrontStandardAboutPanel: nil];
  return Qnil;
}


static Lisp_Object
interpret_command_keys(id menu,
                       Lisp_Object prefix,Lisp_Object old)
{
  int i;
  id cell;
  unsigned short key;
  NSString *keys;
  NSArray *items = [menu itemArray];

  TRACE(interpret_command_keys);
  
  for (i=0; i < [items count]; i++)
    {
      cell=[items objectAtIndex:i];
      if (0 /*[cell hasSubmenu]*/)
        {
          old=interpret_command_keys([cell target],
                                     Fcons([(EmacsMenuObject *)[cell representedObject] value],
                                           prefix),
                                     old);
        }
      else if ((keys=[cell userKeyEquivalent]) && [keys length])
        {
          key = [keys characterAtIndex:0];
          old=Fcons(Fcons(make_number(key|super_modifier),
                          Freverse(Fcons([(EmacsMenuObject *)[cell representedObject] value],
                                         prefix))),old);
        }
    }
  return old;
}

DEFUN ("ns-list-command-keys", Fns_list_command_keys, Sns_list_command_keys,
       0, 0, 0, "List NS command equivalents for menus.")
     ()
{
  Lisp_Object ret=Qnil;
  check_ns();
  ret=interpret_command_keys([NSApp mainMenu],Qnil,Qnil);
  return ret;
}

static Lisp_Object
interpret_services_menu(NSMenu *menu, Lisp_Object prefix, Lisp_Object old)
// ----------------------------------------------------------------------------
//   Turn the input menu (an NSMenu) into a lisp list for tracking on lisp side
// ----------------------------------------------------------------------------
{
  int i;
  id<NSMenuItem> cell;
  const char *name;
  char *muleStr;
  Lisp_Object nameStr;
  unsigned short key;
  int count, len;
  NSString *keys;
  Lisp_Object res;

  count = [menu numberOfItems];
  for (i=0; i<count; i++)
    {
      cell = [menu itemAtIndex: i];
      name = [[cell title] UTF8String];
      if (!name) continue;

      // convert to Mule since that's what other menu items will be
      len = strlen(name);
      muleStr = (char *) xmalloc(len*2); //PENDING: get better estimate
      UTF8sToMulem(name, len, muleStr);
      nameStr = build_string(muleStr);
      xfree(muleStr);

      if ([cell hasSubmenu])
        {
          old = interpret_services_menu([cell submenu],
                                        Fcons(nameStr, prefix), old);
        }
      else
        {
          keys = [cell keyEquivalent];
          if (keys && [keys length] )
            {
              key = [keys characterAtIndex:0];
              res = make_number(key|super_modifier);
            }
          else
            {
              res=Qundefined;
            }
          old = Fcons(Fcons(res,
                            Freverse(Fcons(nameStr,
                                           prefix))),
                    old);
        }
    }
  return old;
}

DEFUN ("ns-list-services", Fns_list_services, Sns_list_services, 0, 0, 0,
       "List NS services by querying NSApp.\n\
WARNING:  This function crashes NS version older than NS 3.2.")
     ()
{
  Lisp_Object ret = Qnil;
  NSMenu *svcs;
  id delegate;

  check_ns();
  svcs = [[NSMenu alloc] initWithTitle: @"Services"];
  [NSApp setServicesMenu: svcs];  // this and next rebuild on <10.4
  [NSApp registerServicesMenuSendTypes: ns_send_types
                           returnTypes: ns_return_types];

// On Tiger, services menu updating was made lazier (waits for user to
// actually click on the menu), so we have to force things along:
#ifdef MACOSX
  if (NSAppKitVersionNumber >= 744.0)
    {
      delegate = [svcs delegate];
      if (delegate != nil)
        {
          if ([delegate respondsToSelector: @selector(menuNeedsUpdate:)])
            [delegate menuNeedsUpdate: svcs];
          if ([delegate respondsToSelector:
                            @selector(menu:updateItem:atIndex:shouldCancel:)])
            {
              int i, len = [delegate numberOfItemsInMenu: svcs];
              for (i=0; i<len; i++)
                [svcs addItemWithTitle: @"" action: NULL keyEquivalent: @""];
              for (i=0; i<len; i++)
                if (![delegate menu: svcs
                         updateItem: (NSMenuItem *)[svcs itemAtIndex: i]
                            atIndex: i shouldCancel: NO])
                  break;
            }
        }
    }
#endif

  [svcs setAutoenablesItems: NO];
#ifdef MACOSX
  [svcs update]; // on OS X, converts from '/' structure
#endif

  ret = interpret_services_menu(svcs, Qnil, ret);
  return ret;
}

DEFUN ("ns-perform-service", Fns_perform_service, Sns_perform_service,
       2, 2, 0, "Perform NS SERVICE on SEND which is either a string or nil.\n\
Returns result of service as string or nil if no result.")
     (service, send)
     Lisp_Object service, send;
{
  id pb;
  NSString *svcName;
  char *muleStr, *utfStr;
  int len;

  CHECK_STRING (service, 0);
  check_ns();

  muleStr = XSTRING(service)->data;
  len = strlen(muleStr);
  utfStr = (char *)xmalloc(len*2); // better estimate
  mulemToUTF8s(muleStr, len, utfStr);
  svcName = [NSString stringWithUTF8String: utfStr];
  xfree(utfStr);

  pb=[NSPasteboard pasteboardWithUniqueName];
  ns_string_to_pasteboard(pb, send);

  if (NSPerformService(svcName, pb) == NO)
    Fsignal (Qerror, Fcons (build_string ("service not available"), Qnil));

  return ns_string_from_pasteboard(pb);
}

DEFUN ("ns-utf-to-mule", Fns_utf_to_mule, Sns_utf_to_mule,
       1, 1, 0, "Convert given UTF-8 string to Mule string.")
     (str)
     Lisp_Object str;
{
  char *utfStr, *muleStr;
  int len;
  Lisp_Object retStr;

  CHECK_STRING (str, 0);
  check_ns();

  utfStr = XSTRING(str)->data;
  len = strlen(utfStr);
  muleStr = (char *)xmalloc(len*3); // better estimate
  UTF8sToMulem(utfStr, len, muleStr);
  retStr = build_string(muleStr);
  xfree(muleStr);

  return retStr;
}

DEFUN ("ns-mule-to-utf", Fns_mule_to_utf, Sns_mule_to_utf,
       1, 1, 0, "Convert given Mule string to UTF-8 string.")
     (str)
     Lisp_Object str;
{
  char *muleStr, *utfStr;
  int len;
  Lisp_Object retStr;

  CHECK_STRING (str, 0);
  check_ns();

  muleStr = XSTRING(str)->data;
  len = strlen(muleStr);
  utfStr = (char *)xmalloc(len*2); // better estimate
  mulemToUTF8s(muleStr, len, utfStr);
  retStr = build_string(utfStr);
  xfree(utfStr);

  return retStr;
}


DEFUN ("ns-show-ps", Fns_show_ps, Sns_show_ps,
       1, 1, 0, "Turn postscript logging on or off depending of the value of flag.")
     (flag)
     Lisp_Object flag;
{
  check_ns();
#if !defined(MACOSX) && !defined(GNUSTEP)
  debug_log_dps_call("ns-show-ps: [NSDPSContext setAllContextsOutputTraced]");
  [NSDPSContext setAllContextsOutputTraced:!NILP(flag)];
#endif
  return Qnil;
}

DEFUN ("ns-show-events", Fns_show_events, Sns_show_events,
       1, 1, 0, "Turn event logging on or off depending of the value of flag.")
     (flag)
     Lisp_Object flag;
{
  check_ns();
#if !defined(MACOSX) && !defined(GNUSTEP)
  debug_log_dps_call("ns-show-event: [NSDPSServerContext setEventsTraced]");
  NSDPSServerContext setEventsTraced:!NILP(flag)];
#endif
return Qnil;
}

void x_sync(Lisp_Object frame)
{
/* XXX Not implemented XXX */
return;
}


void syms_of_nsfns ()
{
  int i;

  Qns_frame_parameter = intern ("ns-frame-parameter");
  staticpro (&Qns_frame_parameter);
  Qforeground_color = intern ("foreground-color");
  staticpro (&Qforeground_color);
  Qbackground_color = intern ("background-color");
  staticpro (&Qbackground_color);
  Qcursor_color = intern ("cursor-color");
  staticpro (&Qcursor_color);
  Qinternal_border_width = intern ("internal-border-width");
  staticpro (&Qinternal_border_width);
  Qvisibility = intern ("visibility");
  staticpro (&Qvisibility);
  Qicon_type = intern ("icon-type");
  staticpro (&Qicon_type);
  Qcursor_type = intern ("cursor-type");
  staticpro (&Qcursor_type);
  Qicon_left = intern ("icon-left");
  staticpro (&Qicon_left);
  Qicon_top = intern ("icon-top");
  staticpro (&Qicon_top);
  Qleft = intern ("left");
  staticpro (&Qleft);
  Qright = intern ("right");
  staticpro (&Qright);
  Qtop = intern ("top");
  staticpro (&Qtop);
  Qicon_name = intern ("icon-name");
  staticpro (&Qicon_name);
  Qdisplay = intern ("display");
  staticpro (&Qdisplay);
  Qnone = intern ("none");
  staticpro (&Qnone);
  Qx_resource_name = intern ("x-resource-name");
  staticpro (&Qx_resource_name);
  Qvertical_scroll_bars = intern ("vertical-scroll-bars");
  staticpro (&Qvertical_scroll_bars);
  Qauto_raise = intern ("auto-raise");
  staticpro (&Qauto_raise);
  Qauto_lower = intern ("auto-lower");
  staticpro (&Qauto_lower);
  Qbox = intern ("box");
  staticpro (&Qbox);
  Qscroll_bar_width = intern ("scroll-bar-width");
  staticpro (&Qscroll_bar_width);
  Qfontsize = intern ("fontsize");
  staticpro (&Qfontsize);
  Qbuffered = intern ("bufferd");
  staticpro (&Qbuffered);

  Qface_set_after_frame_default = intern ("face-set-after-frame-default");
  staticpro (&Qface_set_after_frame_default);

  for (i = 0; i< sizeof (ns_frame_parms)/ sizeof (ns_frame_parms[0]); i++)
    Fput (intern (ns_frame_parms[i].name), Qns_frame_parameter,
          make_number (i));

  DEFVAR_LISP ("ns-icon-type-alist", &Vns_icon_type_alist,
               "Alist of elements (REGEXP . IMAGE) for images of icons associated to\n\
frames.  If the title of a frame matches REGEXP, then IMAGE.tiff is\n\
selected as the image of the icon representing the frame when it's\n\
miniaturized.  If an element is t, then Emacs tries to select an icon\n\
based on the filetype of the visited file.\n\
\n\
The images have to be installed in a folder called English.lproj in the\n\
Emacs.app folder.  You have to restart Emacs after installing new icons.\n\
\n\
Example: Install an icon Gnus.tiff and execute the following code\n\
\n\
  (setq ns-icon-type-alist\n\
        (append ns-icon-type-alist\n\
                '((\"^\\\\*\\\\(Group\\\\*$\\\\|Summary \\\\|Article\\\\*$\\\\)\"\n\
                   . \"Gnus\"))))\n\
\n\
When you miniaturize a Group, Summary or Article frame, Gnus.tiff will\n\
be used as the image of the icon representing the frame.");
  Vns_icon_type_alist = Fcons (Qt, Qnil);

  DEFVAR_LISP ("x-bitmap-file-path", &Vx_bitmap_file_path,
               "List of directories to search for bitmap files for NS.");
  Vx_bitmap_file_path = decode_env_path ((char *) 0, PATH_BITMAPS);
  
  defsubr (&Sns_read_file_name);
  defsubr (&Sx_get_resource);
  defsubr (&Sns_get_resource);
  defsubr (&Sns_set_resource);
  defsubr (&Sns_display_color_p);
  defsubr (&Sns_display_grayscale_p);
  defsubr (&Sns_list_fonts);
  defsubr (&Sns_list_colors);
  defsubr (&Sns_color_defined_p);
  defsubr (&Sns_color_values);
  defsubr (&Sns_server_max_request_size);
  defsubr (&Sns_server_vendor);
  defsubr (&Sns_server_version);
  defsubr (&Sns_display_pixel_width);
  defsubr (&Sns_display_pixel_height);
  defsubr (&Sns_display_mm_width);
  defsubr (&Sns_display_mm_height);
  defsubr (&Sns_display_screens);
  defsubr (&Sns_display_planes);
  defsubr (&Sns_display_color_cells);
  defsubr (&Sns_display_visual_class);
  defsubr (&Sns_display_backing_store);
  defsubr (&Sns_display_save_under);
  defsubr (&Sns_create_frame);
  defsubr (&Sns_make_key_frame);
  defsubr (&Sns_open_connection);
  defsubr (&Sns_close_connection);
  defsubr (&Sns_display_list);

  defsubr (&Sns_hide_emacs);
  defsubr (&Sns_emacs_info_panel);
  defsubr (&Sns_list_command_keys);
  defsubr (&Sns_list_services);
  defsubr (&Sns_perform_service);
  defsubr (&Sns_utf_to_mule);
  defsubr (&Sns_mule_to_utf);
  defsubr (&Sns_show_ps);
  defsubr (&Sns_yes_or_no_p);
  defsubr (&Sns_show_events);
  defsubr (&Sns_popup_prefs_panel);
  defsubr (&Sns_popup_font_panel);
  defsubr (&Sns_popup_color_panel);

  // this stuff is mainly for fontsets, which we don't fully implement yet
  get_font_info_func = ns_get_font_info;
  list_fonts_func = ns_list_fonts;
  load_font_func = ns_load_font;
  query_font_func = ns_query_font;
  set_frame_fontset_func = ns_set_font;
  check_window_system_func = check_ns;

}

//PENDING: move to nsfaces.m
@implementation EmacsImage

static EmacsImage *ImageList = nil;

+ allocInitFromFile:(Lisp_Object)file
{
  EmacsImage *image = ImageList;
  Lisp_Object found;
  int fd;

  // look for an existing image of the same name
  while (image != nil &&
         [[image name] compare: [NSString stringWithUTF8String:
                                              XSTRING (file)->data]])
    image = [image imageListNext];
  
  if (image != nil)
    {
      [image reference];
      return image;
    }
  
  /* Search bitmap-file-path for the file, if appropriate.  */
  fd = openp (Vx_bitmap_file_path, file, "", &found, 0);
  if (fd < 0)
    return nil;

  close (fd);

  image = [[EmacsImage alloc] initByReferencingFile:
                     [NSString stringWithUTF8String:XSTRING (found)->data]];

  if ([image bestRepresentationForDevice:nil] == nil)
    {
      [image release];
      return nil;
    }

  [image setName:[NSString stringWithUTF8String:XSTRING (file)->data]];
  [image reference];
  ImageList = [image imageListSetNext:ImageList];

  return image;
}

- reference
{
  refCount++;
  return self;
}

- imageListSetNext:(id)arg
{
  imageListNext = arg;
  return self;
}
  
- imageListNext
{
  return imageListNext;
}
  
- (void)dealloc
{
  id list = ImageList;
  
  if (refCount > 1)
    {
      refCount--;
      return;
    }

  if (list == self)
    ImageList = imageListNext;
  else
    {
      while (list != nil && [list imageListNext] != self)
        list = [list imageListNext];
      [list imageListSetNext:imageListNext];
    }
  if (stippleRep && stippleRep != self)
    [stippleRep release];
  { [super dealloc]; return; };
}

- initFromXBM:(unsigned char *)bits width:(int)w height:(int)h
{
  return [self initFromSkipXBM:bits width:w height:h length:0];
}

- initFromSkipXBM:(unsigned char *)bits width:(int)w height:(int)h
           length:(int)length;
{
  NSSize s = {w, h};
  id rep;
  unsigned char *planes[5];
  int bpr = (w + 7) / 8;
  
  [self initWithSize:s];

  rep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:NULL pixelsWide:w pixelsHigh:h bitsPerSample:1 samplesPerPixel:2 hasAlpha:YES isPlanar:YES colorSpaceName:NSCalibratedBlackColorSpace bytesPerRow:bpr bitsPerPixel:0];

  [rep getBitmapDataPlanes:planes];
  {
    // turn the bytes around and invert for alpha on plane 2
    int i, j;
    unsigned char *s = bits;
    unsigned char *d = planes[0];
    unsigned char *d2 = planes[1];
    unsigned char swt[16] = {0, 8, 4, 12, 2, 10, 6, 14, 1, 9, 5, 13,
                             3, 11, 7, 15};
    unsigned char c;
    
    for (j = 0; j < h; j++)
      for (i = 0; i < bpr; i++)
        {
          if (length)
            {
              unsigned char s1, s2;
              while (*s++ != 'x' && s < bits + length);
              if (s >= bits + length)
                {
                  [rep release];
                  return nil;
                }
#define hexchar(x) (isdigit (x) ? x - '0' : x - 'a' + 10)
              s1 = *s++;
              s2 = *s++;
              c = hexchar (s1) * 0x10 + hexchar (s2);
            }
          else
            c = *s++;
          
          *d = 0;
          *d = swt[c >> 4] | (swt[c & 0xf] << 4);
          *d2 = *d ^ 255;
          d++;
          d2++;
        }
  }
  
  [self addRepresentation:rep];

  return self;
}

// since 48 is a popular pixmap size, 96 is a good choice
#define STIPPLE_MIN_WIDTH 96
#define STIPPLE_MIN_HEIGHT 96
- prepareForStippling
{
  NSSize r;

  if (stippleRep && stippleRep != self)
    [stippleRep release];
  
  r = [self size];
  if ((r.width < STIPPLE_MIN_WIDTH || r.height < STIPPLE_MIN_HEIGHT)
      && r.width != 0 && r.height != 0)
    {
      int i;
      NSScreen *screen = [NSScreen mainScreen];// FIXME/cl for multi-display
      NSSize new;

      for (i = r.width; i < STIPPLE_MIN_WIDTH; i *= 2, new.width = i);
      for (i = r.height; i < STIPPLE_MIN_HEIGHT; i *= 2, new.height = i);

      stippleRep = [[NSImage alloc] initWithSize:new];
      [stippleRep addRepresentation:
                    [[[NSCachedImageRep alloc]
                             initWithSize:[stippleRep size]
                                    depth:[screen depth]
                                 separate:[stippleRep isCachedSeparately]
                                    alpha:YES] autorelease]];
      [stippleRep lockFocus];
      {
        NSPoint p = {0, 0};
        [self compositeToPoint:p operation:NSCompositeCopy];
        if (new.width != r.width)
          {
            NSRect s = { {0, 0}, {0, r.height} };
            NSPoint d = {0, 0};
            for (i = r.width; i != new.width; i *= 2)
              {
                s.size.width = i;
                d.x = i;
                [stippleRep compositeToPoint:d fromRect:s operation:NSCompositeCopy];
              }
          }
        if (new.height != r.height)
          {
            NSRect s = { {0, 0}, {new.width, 0} };
            NSPoint d = {0, 0};
            for (i = r.height; i != new.height; i *= 2)
              {
                s.size.height = i;
                d.y = i;
                [stippleRep compositeToPoint:d fromRect:s operation:NSCompositeCopy];
              }
          }
        [stippleRep unlockFocus];
      }
    }
  else
    stippleRep = self;

  return self;
}

- (NSImage *)stippleRep
{
  if (! stippleRep)
    [self prepareForStippling];
  return stippleRep;
}

@end


@implementation EmacsSavePanel
#ifdef MACOSX
// These are overridden to intercept on OS X: ending panel restarts NSApp
// event loop if it is stopped.  Not sure if this is correct behavior,
// perhaps should check if running and if so send an appdefined.
- (void) ok:(id)sender
{
  [super ok: sender];
  panelOK = 1;
  [NSApp stop: self];
}
- (void) cancel:(id)sender
{
  [super cancel: sender];
  [NSApp stop: self];
}
#endif
@end

@implementation EmacsOpenPanel
#ifdef MACOSX
// These are overridden to intercept on OS X: ending panel restarts NSApp
// event loop if it is stopped.  Not sure if this is correct behavior,
// perhaps should check if running and if so send an appdefined.
- (void) ok:(id)sender
{
  [super ok: sender];
  panelOK = 1;
  [NSApp stop: self];
}
- (void) cancel:(id)sender
{
  [super cancel: sender];
  [NSApp stop: self];
}
#endif
@end

@implementation EmacsFileDelegate
// these are the delegate methods
- (BOOL)panel:(id)sender isValidFilename:(NSString *)filename
{
  return YES;
}
- (BOOL)panel:(id)sender shouldShowFilename:(NSString *)filename
{
  return YES;
}
- (NSString *)panel:(id)sender userEnteredFilename:(NSString *)filename
          confirmed:(BOOL)okFlag
{
  return filename;
}
@end

#endif
