/* Definitions and headers for communication with NeXTstep / MacOSX
   Copyright (C) 1989, 1993 Free Software Foundation, Inc.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "config.h"

#ifdef HAVE_NS

// this gets included from a couple of the plain (non-NS) .c files
#ifdef __OBJC__
#import <AppKit/AppKit.h>
#import <Foundation/NSDistantObject.h>

#ifdef MACOSX
#include <AvailabilityMacros.h>
#endif


// ============================================================================
// 
//   The Emacs application
// 
// ============================================================================

// We override sendEvent: as a means to stop/start the event loop
@interface EmacsApp : NSApplication
{
}
- (void)sendEvent:(NSEvent *)theEvent;
- (void)showPreferencesWindow:(id)sender;
@end


// ============================================================================
// 
//   The main Emacs view
// 
// ============================================================================

@interface EmacsView : NSView <NSTextInput>
   {
   struct frame *emacsframe;
   char *old_title;
   BOOL windowClosing;
   NSTrackingRectTag windowTrackNum;
   NSTrackingRectTag frameTrackNum;
   NSTrackingRectTag faceTrackNum;
   NSRect trackrec;  // for debugging..
@public
   int rows, cols;
   int scrollbarsNeedingUpdate;
   NSRectArray sbRects;
   int nsbRects;
   }

// AppKit-side interface
- (void)resetCursorRects;
- menuDown:sender;
- (void)keyDown:(NSEvent *)theEvent;
- (void)mouseDown:(NSEvent *)theEvent;
- (void)mouseUp:(NSEvent *)theEvent;
- setMiniwindowImage;

// Emacs-side interface
- initFrameFromEmacs: (struct frame *) f;
- (void) setRows: (int) r andColumns: (int) c;
- (void) setWindowClosing: (BOOL)closing;
- (void) adjustTrackingRects;
- (void) setMouseTracking:(NSRect)rect;
- (void) removeMouseTracking;
- (void) addSBRect:(NSRect)r;
- (void) dropSBRect:(NSRect)r;
@end

@interface EmacsMenu : NSMenu
   {
   int minmarked;
   id supercell;
   }

- addItem:(Lisp_Object)val string:(Lisp_Object)aString bind:(Lisp_Object)bind;

- (id <NSMenuItem>)addMinItemWithTitle:(NSString *)aString 
				action:(SEL)aSelector 
			 keyEquivalent:(NSString *)charCode;

- adjustMap:(Lisp_Object)keymap;
- (void)adjustKey:(Lisp_Object)key binding:(Lisp_Object)bind;
- (void)adjustKey:(Lisp_Object)key binding:(Lisp_Object)bind string:(Lisp_Object)string;
- findCell:(Lisp_Object)value;
- (int)findCellNumber:(Lisp_Object)value;

- removeCell:cell;

- supercell;
- setSupercell:(id) nsupercell;

- mark;
- freeMarked;

- dump:(NSString *)indent;

@end

@interface EmacsMenuObject : NSObject
{
  Lisp_Object value;
  id supercell;
}

- (Lisp_Object)value;
- setValue:(Lisp_Object)nvalue;
- supercell;
- setSupercell:(id) nsupercell;
@end

@interface EmacsMenuCard : NSPopUpButton
-init;
-addTitle: (char *)name;
-addItem:  (char *)name value:(Lisp_Object)val enabled:(BOOL)enabled;
@end

@interface EmacsMenuPanel : NSPanel
   {
   NSMutableArray *list;
   int maxwidth;
   }
- initFromMenu:(Lisp_Object)menu;
- addItem:(char *)str value:(Lisp_Object)val enabled:(BOOL)enabled at:(int)pane;
- addTitle:(char *)str to:(int)pane;
- (int)cards;
- (Lisp_Object)runMenuAt:(float)x:(float)y;
@end

@interface EmacsDialogPanel : NSPanel
   {
   NSTextField *command;
   NSTextField *title;
   NSMatrix *matrix;
   int rows,cols;
   }
- initFromContents:(Lisp_Object)menu;
- addButton:(char *)str value:(Lisp_Object)val row:(int)row;
- addString:(char *)str row:(int)row;
- addSplit;
- (Lisp_Object)runDialogAt:(float)x:(float)y;
@end

//TODO: replace EmacsMenuPanel / EmacsMenuCard with this
//  @interface EmacsPopupMenu : NSMenu
//  {
//  }
//  - initFromMenu:(Lisp_Object)menu;
//  - addItem:(char *)str value:(Lisp_Object)val enabled:(BOOL)enabled at:(int)pane;
//  @end


@interface EmacsFileDelegate : NSObject
@end

// this and next are needed to work around OS X behavior of restarting
// application loop when user dismisses panel
@interface EmacsSavePanel : NSSavePanel
{
}
@end
@interface EmacsOpenPanel : NSOpenPanel
{
}
@end

@interface EmacsImage : NSImage
{
  id imageListNext;
  int refCount;
  NSImage *stippleRep;
}
+ allocInitFromFile:(Lisp_Object)file;
- reference;
- imageListSetNext:(id)arg;
- imageListNext;
- (void)dealloc;
- initFromXBM:(unsigned char *)bits width:(int)w height:(int)h;
- initFromSkipXBM:(unsigned char *)bits width:(int)w height:(int)h
           length:(int)length;
- prepareForStippling;
- (NSImage *)stippleRep;
@end

@interface EmacsScroller : NSScroller
  {
   Lisp_Object win;
   struct frame *frame;
   NSResponder *prevResponder; 

   /* The offset to the bottom of the knob of the last mouse
      down normalized to the knob height.  */
   float last_mouse_offset;

   BOOL condemned;

   /* Number of event we should fake in `getNextEvent:waitFor:threshold:'. */
   BOOL fake_event_p;
   NSEvent *fake_event;
   BOOL repeat;
   float lastInc;

   // optimize against excessive positioning calls generated by emacs
   int em_position;
   int em_portion;
   int em_whole;
   }
- initFrame:(NSRect )r window:(Lisp_Object)win;
- (void)setFrame:(NSRect)r;
- (void)dealloc;

- scrollerMoved:sender;
- setPosition:(int) position portion:(int) portion whole:(int) whole;
- (int) checkSamePosition:(int) position portion:(int) portion whole:(int) whole;

- condemn;
- reprieve;
- judge;
@end

#if defined(MACOSX) && (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_3)
// rendering util
@interface EmacsGlyphStorage : NSObject <NSGlyphStorage>
{
@public
  NSAttributedString *attrStr;
  NSMutableDictionary *dict;
  CGGlyph cglyphs[1024];
  unsigned int maxChar, maxGlyph;
  int i, len;
}
- (void) setString: (NSString *)str font: (NSFont *)font;
@end
#endif

// running the preferences window
@interface EmacsPrefsController : NSObject
{
    struct frame *frame;
    IBOutlet NSWindow *prefsWindow;
    IBOutlet NSMatrix *altIsMetaMatrix;
    IBOutlet NSMatrix *cursorTypeMatrix;
    IBOutlet NSSlider *cursorBlinkSlider;
    IBOutlet NSSlider *shrinkSpaceSlider;
#ifdef MACOSX
    IBOutlet NSButton *smoothFontsCheck;
    IBOutlet NSButton *useQuickdrawCheck;
    IBOutlet NSButton *useSysHiliteCheck;
    BOOL prevUseHighlightColor;
#endif
    float prevShrinkSpace;
    float prevBlinkRate;
}
- (IBAction)cancel:(id)sender;
- (IBAction)ok:(id)sender;
- (IBAction)resetToDefaults:(id)sender;
- (IBAction)runHelp:(id)sender;
- (IBAction)setColors:(id)sender;
- (IBAction)setDefaultFont:(id)sender;

- (void) showForFrame: (struct frame *)f;
- (void) setPanelFromValues;
- (void) setValuesFromPanel;
@end

extern id ns_emacscolors;
extern Lisp_Object ns_alternate_is_meta,ns_cursor_blink_rate;
extern Lisp_Object ns_shrink_space;
extern Lisp_Object ns_menu_path;
extern NSArray *ns_send_types,*ns_return_types;
extern struct frame *ns_highlight_frame;
extern struct frame *ns_focus_frame;
extern struct frame *ns_focus_event_frame;
extern EmacsMenu *mainMenu, *svcsMenu;
#endif  /* __OBJC__ */

enum ns_cursor_types
   {
   no_highlight=0,
   filled_box,
   hollow_box,
   bar,
   line
   };

struct ns_display_info
{
  /* Chain of all ns_display_info structures.  */
  struct ns_display_info *next;

  /* This is a cons cell of the form (NAME . FONT-LIST-CACHE).
     The same cons cell also appears in ns_display_name_list.  */
  Lisp_Object name_list_element;


  /* A table of all the fonts we have already loaded.  */
  struct font_info *font_table;

  /* The current capacity of x_font_table.  */
  int font_table_size;

  /* The number of fonts actually stored in x_font_table.
     font_table[n] is used and valid iff 0 <= n < n_fonts.
     0 <= n_fonts <= font_table_size.  */
  int n_fonts;

#ifdef MULTI_KBOARD
  struct kboard *kboard;
#endif
};

/* This is a chain of structures for all the NS displays currently in use.  */
extern struct ns_display_info *ns_display_list;

extern Lisp_Object ns_display_name_list;

extern struct ns_display_info *ns_display_info_for_display ();
extern struct ns_display_info *ns_display_info_for_name ();

extern struct ns_display_info *ns_term_init ();
extern void ns_term_shutdown();

struct ns_output
   {
#ifdef __OBJC__
   id view;
   id miniimage;
   NSColor * current_cursor_color;
   NSColor * desired_cursor_color;
   NSColor *foreground_color;
   NSColor *background_color;
#else
   void *view;
   void *miniimage;
   void *current_cursor_color;
   void *desired_cursor_color;
   void *foreground_color;
   void *background_color;
#endif
//   struct face *face;
   XFontStruct *font;
   float top;
   float left;
   Lisp_Object icon_top;
   Lisp_Object icon_left;
   int line_height;
   float internal_border_width;
   float border_width, border_height;
   enum ns_cursor_types current_cursor, desired_cursor;

   /* The size of the extra width currently allotted for vertical
      scroll bars, in pixels.  */
   int vertical_scroll_bar_extra;
   
  /* Table of parameter faces for this frame.  Any NS resources (pixel
      values, fonts) referred to here have been allocated explicitly
      for this face, and should be freed if we change the face.  */
   struct face **param_faces;
   int n_param_faces;

   /* Table of computed faces for this frame.  These are the faces
      whose indexes go into the upper bits of a glyph, computed by
      combining the parameter faces specified by overlays, text
      properties, and what have you.  The NS resources mentioned here
      are all shared with parameter faces.  */
   struct face **computed_faces;
   int n_computed_faces;                /* How many are valid */
   int size_computed_faces;     /* How many are allocated */

  /* This is the Emacs structure for the NS display this frame is on.  */
  struct ns_display_info *display_info;

};

#define FRAME_FOREGROUND_COLOR(f) ((f)->output_data.ns->foreground_color)
#define FRAME_BACKGROUND_COLOR(f) ((f)->output_data.ns->background_color)
#define FRAME_CURSOR_COLOR(f) ((f)->output_data.ns->current_cursor_color)
#define FRAME_NEW_CURSOR_COLOR(f) ((f)->output_data.ns->desired_cursor_color)
#define NS_FACE_FOREGROUND(f) ((NSColor *)f->foreground)
#define NS_FACE_BACKGROUND(f) ((NSColor *)f->background)
#define NS_FONT_WIDTH(f)	((f)->output_data.ns->font->width)
#ifndef COMPILING_MULTI_TERM
#define FONT_WIDTH(f)	((f)->width)
#define FONT_HEIGHT(f)	((f)->height)

/* Get at the computed faces of an NS window frame.  */
#define FRAME_PARAM_FACES(f) ((f)->output_data.ns->param_faces)
#define FRAME_N_PARAM_FACES(f) ((f)->output_data.ns->n_param_faces)
#define FRAME_DEFAULT_PARAM_FACE(f) (FRAME_PARAM_FACES (f)[0])
#define FRAME_MODE_LINE_PARAM_FACE(f) (FRAME_PARAM_FACES (f)[1])

#define FRAME_COMPUTED_FACES(f) ((f)->output_data.ns->computed_faces)
#define FRAME_COMPUTED_FACE(f,n) ((f)->output_data.ns->computed_faces[n])
#define FRAME_N_COMPUTED_FACES(f) ((f)->output_data.ns->n_computed_faces)
#define FRAME_SIZE_COMPUTED_FACES(f) ((f)->output_data.ns->size_computed_faces)
#define FRAME_DEFAULT_FACE(f) ((f)->output_data.ns->computed_faces[0])
#define FRAME_MODE_LINE_FACE(f) ((f)->output_data.ns->computed_faces[1])
#endif /* ! COMPILING_MULTI_TERM */

#define FRAME_NS_VIEW(f) ((f)->output_data.ns->view)
#define FRAME_CURSOR(f)  ((f)->output_data.ns->current_cursor)
#define FRAME_NEW_CURSOR(f)  ((f)->output_data.ns->desired_cursor)

#ifndef COMPILING_MULTI_TERM
#define FRAME_FONT(f) ((f)->output_data.ns->font)
#endif /* ! COMPILING_MULTI_TERM */
#define NS_FRAME_FONT(f) ((f)->output_data.ns->face)
#define NS_FRAME_INTERNAL_BORDER_WIDTH(f) ((f)->output_data.ns->internal_border_width)
#define FRAME_INTERNAL_BORDER_WIDTH(f) NS_FRAME_INTERNAL_BORDER_WIDTH(f)
#define NS_FRAME_LINE_HEIGHT(f) ((f)->output_data.ns->line_height)
#define FRAME_LINE_HEIGHT(f) NS_FRAME_LINE_HEIGHT(f)

/* This gives the ns_display_info structure for the display F is on.  */
#define FRAME_NS_DISPLAY_INFO(f) ((f)->output_data.ns->display_info)

/* This is the `Display *' which frame F is on.  */
#define FRAME_NS_DISPLAY(f) (0)

/* These two really ought to be called FRAME_PIXEL_{WIDTH,HEIGHT}.  */
#define NS_PIXEL_WIDTH(f) (x_pixel_width (f))
#define NS_PIXEL_HEIGHT(f) (x_pixel_height (f))
#define PIXEL_WIDTH(f) NS_PIXEL_WIDTH(f)
#define PIXEL_HEIGHT(f) NS_PIXEL_HEIGHT(f)


/* Manipulating pixel sizes and character sizes.
   Knowledge of which factors affect the overall size of the window should
   be hidden in these macros, if that's possible.

   Return the upper/left pixel position of the character cell on frame F
   at ROW/COL.  */
#define CHAR_TO_PIXEL_ROW(f, row) \
  ((f)->output_data.ns->internal_border_width \
   + (row) * (f)->output_data.ns->line_height)
#define CHAR_TO_PIXEL_COL(f, col) \
  ((f)->output_data.ns->internal_border_width \
   + (col) * NS_FONT_WIDTH (f))

/* Return the pixel width/height of frame F if it has
   WIDTH columns/HEIGHT rows.  */
#define CHAR_TO_PIXEL_WIDTH(f, width) \
  (CHAR_TO_PIXEL_COL (f, width) \
   + (f)->output_data.ns->vertical_scroll_bar_extra \
   + (f)->output_data.ns->internal_border_width)
#define CHAR_TO_PIXEL_HEIGHT(f, height) \
  (CHAR_TO_PIXEL_ROW (f, height) \
   + (f)->output_data.ns->internal_border_width)

/* Compute pixel size for vertical scroll bars */
#define NS_SCROLL_BAR_WIDTH(f)                              \
(FRAME_HAS_VERTICAL_SCROLL_BARS(f)                          \
 ? rint(FRAME_SCROLL_BAR_PIXEL_WIDTH (f) > 0                \
        ? FRAME_SCROLL_BAR_PIXEL_WIDTH (f)                  \
        : (FRAME_SCROLL_BAR_COLS (f) * NS_FONT_WIDTH(f)))   \
 : 0)
 
#define FRAME_NS_FONT_TABLE(f) (FRAME_NS_DISPLAY_INFO (f)->font_table)

/* First position where characters can be shown (instead of scrollbar, if
   it is on left. */
#define FIRST_CHAR_POSITION(f) (!(FRAME_HAS_VERTICAL_SCROLL_BARS_ON_LEFT(f)) ? \
         0 : FRAME_SCROLL_BAR_COLS(f))


#ifndef COMPILING_MULTI_TERM
#define FRAME_DESIRED_CURSOR(f) ((f)->output_data.ns->desired_cursor)
#endif /* ! COMPILING_MULTI_TERM */


/* Dump n glyphs starting at gp on frame f at position (x,y). fl indicates
   various modifiers on how to draw the glyphs. */
extern void ns_dumpcursor(int where, struct frame *f,int x,int y);
extern void ns_dumpglyphs(struct frame *f,int x,int y,GLYPH *gp,int n, int fl);
#define NS_DUMPGLYPH_NORMAL             0
#define NS_DUMPGLYPH_HIGHLIGHT          1
#define NS_DUMPGLYPH_CURSOR             2
#define NS_DUMPGLYPH_MOUSEFACE          3
#define NS_DUMPGLYPH_BACKGROUND         4
#define NS_DUMPGLYPH_FOREGROUND         5

extern Lisp_Object ns_new_font();
extern struct font_info *ns_load_font (struct frame *f,char *name,int sz);
extern XFontStruct *ns_load_face_font(struct face *face, struct frame *f);
extern Lisp_Object ns_list_fonts ();
extern struct font_info *ns_get_font_info ();
extern struct font_info *ns_query_font (struct frame *f, char *fontname);

extern void check_ns(void);
extern void ns_set_frame_parameters();
extern Lisp_Object ns_map_event_to_object ();
extern Lisp_Object ns_string_from_pasteboard();
extern void ns_string_to_pasteboard();
extern void nxatoms_of_nsselect();
extern int ns_lisp_to_cursor_type();
extern Lisp_Object ns_cursor_type_to_lisp(int arg);
extern struct frame *ns_focus_frame;

void set_frame_menubar (struct frame *f, int first_time, int deep_p);

#ifdef __OBJC__
extern int ns_lisp_to_color (Lisp_Object color, NSColor **col);
#endif
extern void x_set_window_size(struct frame *f, int change_grav,
                              int cols, int rows);

/* Create the first two computed faces for a frame -- the ones that
   have GC's.  */
extern void ns_init_frame_faces (/* FRAME_PTR */);

/* Free the resources for the faces associated with a frame.  */
extern void ns_free_frame_faces (/* FRAME_PTR */);

/* Given a frame and a face name, return the face's ID number, or
   zero if it isn't a recognized face name.  */
extern int ns_face_name_id_number (/* FRAME_PTR, Lisp_Object */);

/* Return non-zero if FONT1 and FONT2 have the same size bounding box.
   We assume that they're both character-cell fonts.  */
extern int same_size_fonts (/* XFontStruct *, XFontStruct * */);

/* Recompute the GC's for the default and modeline faces.
   We call this after changing frame parameters on which those GC's
   depend.  */
extern void ns_recompute_basic_faces (/* FRAME_PTR */);

/* Return the face ID associated with a buffer position POS.  Store
   into *ENDPTR the next position at which a different face is
   needed.  This does not take account of glyphs that specify their
   own face codes.  F is the frame in use for display, and W is a
   window displaying the current buffer.
   REGION_BEG, REGION_END delimit the region, so it can be highlighted.  */
extern int compute_char_face (struct frame *f, struct window *w, int pos,
                              int region_beg, int region_end, int *endptr,
                              int limit, int mouse);

/* Update height and voffset for all param face fonts based on shrinkspace. */
void ns_frame_update_shrink_space(struct frame *f);

/* Update the line_height of frame F according to the biggest font in
   any face.  Return nonzero if if line_height changes.  */
extern int ns_frame_update_line_height (struct frame *f);

/* Return the face ID to use to display a special glyph which selects
   FACE_CODE as the face ID, assuming that ordinarily the face would
   be BASIC_FACE.  F is the frame.  */
extern int ns_compute_glyph_face (/* FRAME_PTR, int */);

#ifdef GNUSTEP
extern char gnustep_base_version[];  /* version tracking */
#endif

#define MINWIDTH 10
#define MINHEIGHT 10

#ifdef __OBJC__
#define XNS_SCROLL_BAR(vec) ((id) LISP_TO_VOID (vec))
#else
#define XNS_SCROLL_BAR(vec) LISP_TO_VOID (vec)
#endif

/* Screen max coordinate
 Using larger coordinates causes movewindow/placewindow to abort */
#define SCREENMAX 16000

#define NS_SCROLL_BAR_WIDTH_DEFAULT     [EmacsScroller scrollerWidth]

/* Little utility macros */
#define IN_BOUND(min,x,max) (((x)<(min)) ? (min) : (((x)>(max)) ? (max) : (x)))
#define SCREENMAXBOUND(x) (IN_BOUND(-SCREENMAX,x,SCREENMAX))

#define QXFLOAT(lispObj) (XFLOAT(lispObj)->data)

#endif
