/* NeXTstep / MacOSX communication module
   Copyright (C) 1989, 1993, 1994 Free Software Foundation, Inc.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Originally by Carl Edman
Updated by Christian Limpach (chris@nice.ch)
OpenStep/Rhapsody  port by Scott Bender (sbender@harmony-ds.com)
MacOSX/Aqua port by Christophe de Dinechin (descubes@earthlink.net)
GNUstep port by Adrian Robert (arobert@cogsci.ucsd.edu)

*/

#import <AppKit/AppKit.h>

#include <math.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "config.h"
#include "lisp.h"
#include "blockinput.h"
#include "sysselect.h"
#include "dispextern.h"
#include "nsterm.h"
#include "systime.h"
#include "charset.h"
#include "fontset.h"

#include "termhooks.h"
#include "termopts.h"
#include "termchar.h"

#include "frame.h"
#include "disptab.h"
#include "buffer.h"
#include "window.h"
#include "keyboard.h"
#include "paths.h"
#include "nscharset.h"

/*
  Originally drawing code was in postscript direct DPS calls.  Might be
  faster than replacement nswrap_moveshow() function, but not usable on
  OS X and not sure on GNUstep).
 */
// #include "nswraps.h"
void nswrap_moveshow(float x, float y, const int txt[], int size,
                     struct ns_font *font, NSColor *col, NSColor *bgCol,
                     int ul, float voffset, float w);

// call tracing
#if 0
int term_trace_num = 0;
#define TRACE(x)        fprintf(stderr, "%s:%d: [%d] " #x "\n",         \
                                __FILE__, __LINE__, ++term_trace_num)
#else
#define TRACE(x)
#endif

// make the tracking rects used for mouse-face highlighting visible
#if 0
#define TRACKREC(r, v) {NSRect s=r; [v lockFocus]; NSHighlightRect(s);  \
  s.size.width-=2; s.size.height-=2; s.origin.x+=1; s.origin.y+=1;      \
  NSHighlightRect(s); trackrec=r; [v unlockFocus];                      \
  [[v window] flushWindow];}
#else
#define TRACKREC(r, v)
#endif


// ============================================================================
//
//    Local declarations
//
// ============================================================================

#define min(a,b) ((a)<(b) ? (a) : (b))
#define max(a,b) ((a)>(b) ? (a) : (b))

// Special keycodes that we pass down the event chain
#define KEY_NS_POWER_OFF      ((1<<28)|(0<<16)|1)
#define KEY_NS_OPEN_FILE      ((1<<28)|(0<<16)|2)
#define KEY_NS_OPEN_TEMP_FILE ((1<<28)|(0<<16)|3)
#define KEY_NS_DRAG_FILE      ((1<<28)|(0<<16)|4)
#define KEY_NS_DRAG_COLOR     ((1<<28)|(0<<16)|5)
#define KEY_NS_DRAG_ASCII     ((1<<28)|(0<<16)|6)
#define KEY_NS_CHANGE_FONT    ((1<<28)|(0<<16)|7)
#define KEY_NS_OPEN_FILE_LINE ((1<<28)|(0<<16)|8)
#define KEY_NS_CHANGE_GDB     ((1<<28)|(0<<16)|9)

// Convert a symbol indexed with an NSxxx value to a value as defined
// in keyboard.c (lispy_function_key). I hope this is a correct way
// of doing things...
static unsigned convert_ns_to_X_keysym[] =
{
  NSHomeFunctionKey,            0xFF50,
  NSLeftArrowFunctionKey,       0xFF51,
  NSUpArrowFunctionKey,         0xFF52,
  NSRightArrowFunctionKey,      0xFF53,
  NSDownArrowFunctionKey,       0xFF54,
  NSPageUpFunctionKey,          0xFF55,
  NSPageDownFunctionKey,        0xFF56,
  NSEndFunctionKey,             0xFF57,
  NSBeginFunctionKey,           0xFF58,
  NSSelectFunctionKey,          0xFF60,
  NSPrintFunctionKey,           0xFF61,
  NSExecuteFunctionKey,         0xFF62,
  NSInsertFunctionKey,          0xFF63,
  NSUndoFunctionKey,            0xFF65,
  NSRedoFunctionKey,            0xFF66,
  NSMenuFunctionKey,            0xFF67,
  NSFindFunctionKey,            0xFF68,
  NSHelpFunctionKey,            0xFF6A,
  NSBreakFunctionKey,           0xFF6B,

  NSF1FunctionKey,              0xFFBE,
  NSF2FunctionKey,              0xFFBF,
  NSF3FunctionKey,              0xFFC0,
  NSF4FunctionKey,              0xFFC1,
  NSF5FunctionKey,              0xFFC2,
  NSF6FunctionKey,              0xFFC3,
  NSF7FunctionKey,              0xFFC4,
  NSF8FunctionKey,              0xFFC5,
  NSF9FunctionKey,              0xFFC6,
  NSF10FunctionKey,             0xFFC7,
  NSF11FunctionKey,             0xFFC8,
  NSF12FunctionKey,             0xFFC9,
  NSF13FunctionKey,             0xFFCA,
  NSF14FunctionKey,             0xFFCB,
  NSF15FunctionKey,             0xFFCC,

  NSDeleteFunctionKey,          0xFFFF
};


Lisp_Object ns_input_file,ns_input_font,ns_input_fontsize,ns_input_line;
Lisp_Object ns_input_color,ns_input_ascii;
extern Lisp_Object Qmouse_face;
extern Lisp_Object Qleft, Qright, Qcursor_color, Qcursor_type;
extern Lisp_Object Qbuffered;

EmacsPrefsController *prefsController;

/* Defaults managed through the OpenStep defaults system.  These pertain to
   the NS interface specifically.  Although a customization group could be
   created, it's more natural to manage them via defaults. */

// Use the alt key for meta (instead of command).  This is most sensible
// except on some European keyboards where it interferes with things.
Lisp_Object ns_alternate_is_meta;

// A floating point value specifying the rate at which to blink the cursor.
// YES indicates 0.5, NO indicates no blinking.
Lisp_Object ns_cursor_blink_rate;

// A floating point value specifying compensation for fonts that request
// excessive line heights.  See nsfaces.m ns_load_font.  YES indicates 0.75,
// NO indicates 1.0.
Lisp_Object ns_shrink_space;

// Control via default 'GSFontAntiAlias' on OS X and GNUstep.
int ns_antialias_text;

// On OS X picks up the default NSGlobalDomain AppleAntiAliasingThreshold,
// the maximum font size to NOT antialias.  On GNUstep there is currently
// no way to control this behavior.
float ns_antialias_threshold;

// Controls use of an undocumented CG function to do Quickdraw-style font
//  smoothing (less heavy) instead of regular Quartz smoothing.
int ns_use_qd_smoothing;

// Used to pick up AppleHighlightColor on OS X
int ns_use_system_highlight_color;
NSString *ns_selection_color;

NSArray *ns_send_types=0,*ns_return_types=0,*ns_drag_types=0;

/* This is a chain of structures for all the NS displays currently in use.  */
struct ns_display_info *ns_display_list;
Lisp_Object ns_display_name_list;

// display update
extern struct frame *updating_frame;
struct frame *ns_highlight_frame = 0;
struct frame *ns_focus_frame = 0;
struct frame *ns_focus_event_frame = 0;
static NSView *focus_view = NULL;
static int ns_window_num=0;
static NSRect uRect;
static int curs_x,curs_y,flexlines,highlight;
static void ns_frame_rehighlight(struct frame *f);
static void ns_adjust_size(struct frame *f);

// event loop
static BOOL send_appdefined = YES;
static NSEvent *last_appdefined_event = 0;
static NSTimer *timed_entry = 0;
static NSTimer *fd_entry = 0;
static NSTimer *cursor_blink_entry = 0;
static fd_set select_readfds, t_readfds;
static struct timeval select_timeout;
static int select_nfds;
static NSAutoreleasePool *outerpool;
static BOOL ns_shutdown_properly = NO;
static struct input_event *events=0;
static int eventsleft=0;

// BEGIN mouse face related machinery
static void note_mouse_movement (struct frame *frame, int x, int y);
static void note_mouse_highlight (struct frame *frame, int x, int y);
static int fast_find_position (Lisp_Object window,
                               int pos, int end, int *columnp, int *rowp,
                               int leftOffset);
static void show_mouse_face (int hl);
static void clear_mouse_face ();

static NSRect last_mouse_glyph;
static unsigned long last_mouse_movement_time;
static struct frame *last_mouse_frame = 0;
static int mouse_face_beg_row, mouse_face_beg_col;
static int mouse_face_end_row, mouse_face_end_col;
static int mouse_face_past_end;
static Lisp_Object mouse_face_window;
static int mouse_face_face_id;
static FRAME_PTR mouse_face_mouse_frame;
static int mouse_face_mouse_x, mouse_face_mouse_y;
static int mouse_face_defer;
static int mouse_face_deferred_gc;
static NSRect mouse_face_mouse_rect;
static int mouse_face_mouse_row;
static id mouse_face_tracked_view;
// END mouse face related machinery


/* Convert modifiers in a NeXTSTEP event to emacs style modifiers.  */
#define EV_MODIFIERS(e)                                             \
((([e modifierFlags] & NSHelpKeyMask) ? hyper_modifier : 0)         \
 |((!NILP(ns_alternate_is_meta) &&                                  \
    ([e modifierFlags] & NSAlternateKeyMask)) ? meta_modifier:0)    \
 |(([e modifierFlags] & NSShiftKeyMask) ? shift_modifier:0)         \
 |(([e modifierFlags] & NSControlKeyMask) ? ctrl_modifier:0)        \
 |(([e modifierFlags] & NSCommandKeyMask) ? super_modifier:0))

#define EV_UDMODIFIERS(e)                                   \
((([e type] == NSLeftMouseDown) ? down_modifier : 0)        \
 |(([e type] == NSRightMouseDown) ? down_modifier : 0)      \
 |(([e type] == NSLeftMouseDragged) ? down_modifier : 0)    \
 |(([e type] == NSRightMouseDragged) ? down_modifier : 0)   \
 |(([e type] == NSLeftMouseUp)   ? up_modifier   : 0)       \
 |(([e type] == NSRightMouseUp)   ? up_modifier   : 0))

// note, emacs expects a mouse 4 or mouse 5 button event for scroll wheel
// movement, but NS sends a button 0 with deltaY = +/- 1
#define EV_BUTTON(e)                                                        \
((([e type] == NSLeftMouseDown) || ([e type] == NSLeftMouseUp)) ? 0 :       \
 (([e type] == NSRightMouseDown) || ([e type] == NSRightMouseUp)) ? 1 : \
 (([e type] != NSScrollWheel) ? 2 : (([e deltaY] > 0) ? 3 : 4)) )

/* Convert the time field to a timestamp in milliseconds. */
#ifdef GNUSTEP
// Apple says timestamp is in seconds, but gnustep seems to be returning msec
#define EV_TIMESTAMP(e) ([e timestamp])
#else
#define EV_TIMESTAMP(e) ([e timestamp] * 1000)
#endif // not gnustep

/* This is a piece of code which is common to all the event handling
   methods.  Maybe it should even be a function.  */
#define EV_TRAILER(e)                               \
  {                                                 \
  XSETFRAME (events->frame_or_window, emacsframe);  \
  events->timestamp = EV_TIMESTAMP (e);             \
  events++;                                         \
  eventsleft--;                                     \
  ns_send_appdefined (-1);                          \
  }


// ============================================================================
//
//    Utilities
//
// ============================================================================

char *ns_dynamic_path(char *startingPath)
{
  NSMutableString *initPath =
      [NSMutableString stringWithUTF8String: startingPath];
  NSString *prefixDir;
  NSString *invDir =
      [NSString stringWithUTF8String: XSTRING(Vinvocation_directory)->data];
  NSRange range, searchRange;
  int pos;
  char *ibuf, *buffer;

  // get base
  range = [initPath rangeOfString: @"/Resources"];
  pos = range.location;
  if (pos == NSNotFound)
      return startingPath;
  prefixDir = [initPath substringToIndex: pos + range.length];

  // go back until either contents or .app
  range = [invDir rangeOfString: @"Contents"];
  if (range.location == NSNotFound)
      range = [invDir rangeOfString: @".app"];
  pos = range.location;
  if (pos == NSNotFound)
      return startingPath;
  pos += range.length;
  invDir = [invDir substringToIndex: pos + 1];
  invDir = [invDir stringByAppendingString: @"Resources"];

//NSLog(@"Replacing '%@' with '%@'", prefixDir, invDir);
  for (searchRange = NSMakeRange(0, [initPath length]); searchRange.length>0; )
    {
      range= [initPath rangeOfString: prefixDir options: 0 range: searchRange];
      if (range.location == NSNotFound)
        break;
      [initPath replaceCharactersInRange: range withString: invDir];
      searchRange.location = range.location + [invDir length];
      searchRange.length = [initPath length] - searchRange.location;
    }
  // method not on OS X 10.1
//    [initPath replaceOccurrencesOfString: prefixDir withString: invDir
//        options: 0 range: NSMakeRange(0, [initPath length])];

  ibuf = [initPath UTF8String];
  buffer = (char *)xmalloc((strlen(ibuf)+1)*sizeof(char));
  bcopy(ibuf, buffer, strlen(ibuf)+1);
//  fprintf(stderr, "dynPath start : '%s'\n",startingPath);
//  fprintf(stderr, "dynPath result: '%s'\n",buffer);
  return buffer;
}

/* Subtract the `struct timeval' values X and Y,
   storing the result in RESULT.
   Return 1 if the difference is negative, otherwise 0.  */
static int
timeval_subtract (result, x, y)
     struct timeval *result, x, y;
{
  /* Perform the carry for the later subtraction by updating y.
     This is safer because on some systems
     the tv_sec member is unsigned.  */
  if (x.tv_usec < y.tv_usec)
    {
      int nsec = (y.tv_usec - x.tv_usec) / 1000000 + 1;
      y.tv_usec -= 1000000 * nsec;
      y.tv_sec += nsec;
    }
  if (x.tv_usec - y.tv_usec > 1000000)
    {
      int nsec = (y.tv_usec - x.tv_usec) / 1000000;
      y.tv_usec += 1000000 * nsec;
      y.tv_sec -= nsec;
    }

  /* Compute the time remaining to wait.  tv_usec is certainly positive.  */
  result->tv_sec = x.tv_sec - y.tv_sec;
  result->tv_usec = x.tv_usec - y.tv_usec;

  /* Return indication of whether the result should be considered negative.  */
  return x.tv_sec < y.tv_sec;
}

static void ns_timeout(int usecs)
{
  struct timeval wakeup;

  EMACS_GET_TIME (wakeup);

  /* Compute time to wait until, propagating carry from usecs.  */
  wakeup.tv_usec += usecs;
  wakeup.tv_sec += (wakeup.tv_usec / 1000000);
  wakeup.tv_usec %= 1000000;

  /* Keep waiting until past the time wakeup.  */
  while (1)
    {
      struct timeval timeout;

      EMACS_GET_TIME (timeout);

      /* In effect, timeout = wakeup - timeout.
	 Break if result would be negative.  */
      if (timeval_subtract (&timeout, wakeup, timeout))
	break;

      /* Try to wait that long--but we might wake up sooner.  */
      select (0, NULL, NULL, NULL, &timeout);
    }
}

void *ns_alloc_autorelease_pool()
// ----------------------------------------------------------------------------
//   Allocate a pool for temporary objects (callable from C)
// ----------------------------------------------------------------------------
{
  return [[NSAutoreleasePool alloc] init];
}


void ns_release_autorelease_pool(void *pool)
// ----------------------------------------------------------------------------
//   Free a pool and temporary objects it refers to (callable from C)
// ----------------------------------------------------------------------------
{
  [(NSAutoreleasePool *)pool release];
}


static void ns_focus(struct frame *f, NSRect *r)
// ----------------------------------------------------------------------------
//   Focus on given frame.  During small local updates this is used to draw,
//   however during large updates, ns_update_begin and ns_update_end are
//   called to wrap the whole thing, in which case these calls are stubbed out.
//   Except, on GNUstep, we accumulate the rectangle being drawn into,
//   because the back end won't do this automatically, and will just end
//   up flushing the entire window.
// ----------------------------------------------------------------------------
{
  TRACE(ns_focus);
// static int c=0;
// fprintf(stderr,"focus: %d",c++);
// if (r) fprintf(stderr," (%.0f, %.0f : %.0f x %.0f)", r->origin.x, r->origin.y, r->size.width, r->size.height);
// fprintf(stderr,"\n");
  if (f != updating_frame)
    {
      NSView *view = FRAME_NS_VIEW(f);
      if (view != focus_view)
        {
          if (focus_view != NULL)
            {
              [focus_view unlockFocus];
              [[focus_view window] flushWindow];
            }

          if (view)
#ifdef GNUSTEP
            r ? [view lockFocusInRect: *r] : [view lockFocus];
#else
          [view lockFocus];
#endif
          focus_view = view;
        }
#ifdef GNUSTEP
      else
        {
          // more than one rect being drawn into
          if (view && r)
            {
              [view unlockFocus]; // add prev rect to redraw list
              [view lockFocusInRect: *r]; // focus for draw in new rect
            }
        }
#endif
    }
#ifdef GNUSTEP
  else
    {
      // in batch mode, but in GNUstep must still track rectangles explicitly
      uRect = (r ? NSUnionRect(uRect,*r) : [FRAME_NS_VIEW(f) visibleRect]);
    }
#endif
}


static void ns_unfocus(struct frame *f)
// ----------------------------------------------------------------------------
//   Remove focus on given frame
// ----------------------------------------------------------------------------
{
  TRACE(ns_unfocus);
  if (f != updating_frame)
    {
      if (focus_view != NULL)
        {
          [focus_view unlockFocus];
          [[focus_view window] flushWindow];
          focus_view = NULL;
        }
    }
}


void ns_chars_to_rect(struct frame *f,int x1,int y1,int x2,int y2, NSRect *r)
// ----------------------------------------------------------------------------
//   Convert from char rectangle to AppKit rectangle
// ----------------------------------------------------------------------------
{
  int fw = NS_FONT_WIDTH(f);
  int lh = NS_FRAME_LINE_HEIGHT(f);

  TRACE(ns_chars_to_rect);
  r->origin.x = x1 * fw;
  r->origin.y = y1 * lh;
  r->size.width = (x2 - x1 + 1) * fw;
  r->size.height = (y2 - y1 + 1) * lh;
}


void ns_rect_to_chars(struct frame *f, NSRect *r,
                      int *top, int *left, int *bottom, int *right)
// ----------------------------------------------------------------------------
//   Convert from AppKit rectangle to char rectangle
// ----------------------------------------------------------------------------
{
  int fw = NS_FONT_WIDTH(f);
  int lh = NS_FRAME_LINE_HEIGHT(f);
  *left = r->origin.x / fw;
  *top  = r->origin.y / lh;
  *right = (r->origin.x + r->size.width + fw - 1)  / fw;
  *bottom = (r->origin.y + r->size.height + lh - 1)  / lh;
}


void pixel_to_glyph_coords(struct frame *f,
                           int px, int py,
                           int *x, int *y,
                           void *bounds /* ignored */, int noclip)
// ----------------------------------------------------------------------------
//   Convert from pixel coordinates to chars
// ----------------------------------------------------------------------------
{
  int fw = NS_FONT_WIDTH(f);
  int lh = NS_FRAME_LINE_HEIGHT(f);

  *x= (fw <= 0) ? 0 : floor(((float) px) / fw);
  *y= (lh <= 0) ? 0 : floor(((float) py) / lh);

  if (bounds)
    fprintf(stderr,"NS pixel_to_glyph_coord called with nonzero bounds\n");

  if (!noclip)
    {
      *x = IN_BOUND(0, *x, f->width);
      *y = IN_BOUND(0, *y, f->height);
    }
}

void glyph_to_pixel_coords(struct frame *f,int x,int y,int *pix_x,int *pix_y)
// ----------------------------------------------------------------------------
//   Convert from char coordinates to pixel
// ----------------------------------------------------------------------------
{
  *pix_x=(int) rint(x * NS_FONT_WIDTH(f));
  *pix_y=(int) rint(y * NS_FRAME_LINE_HEIGHT(f));
}


void ns_raise_frame (struct frame *f)
// ----------------------------------------------------------------------------
//   Bring window to foreground and make it active
// ----------------------------------------------------------------------------
{
  NSView *view = FRAME_NS_VIEW(f);
  check_ns();
  [[view window] makeKeyAndOrderFront:NSApp];
}


void ns_lower_frame (struct frame *f)
// ----------------------------------------------------------------------------
//   Send window to back
// ----------------------------------------------------------------------------
{
  NSView *view = FRAME_NS_VIEW(f);
  check_ns();
  [[view window] orderBack:NSApp];
}


void x_make_frame_visible (struct frame *f)
// ----------------------------------------------------------------------------
//   Show the window (X11 semantics)
// ----------------------------------------------------------------------------
{
  ns_raise_frame(f);
}


void x_make_frame_invisible (struct frame *f)
// ----------------------------------------------------------------------------
//   Hide the window (X11 semantics)
// ----------------------------------------------------------------------------
{
  NSView * view = FRAME_NS_VIEW(f);
  check_ns();
  [[view window] orderOut:NSApp];
}


void x_iconify_frame (struct frame *f)
// ----------------------------------------------------------------------------
//   Iconify window
// ----------------------------------------------------------------------------
{
  NSView * view = FRAME_NS_VIEW(f);
  check_ns();
  if ([[view window] windowNumber] <= 0)
    {
      // the window is still deferred.  Make it very small, bring it
      // on screen and order it out.
      NSRect s = { { 100, 100}, {0, 0} };
      NSRect t;
      t = [[view window] frame];
      [[view window] setFrame:s display:NO];
      [[view window] orderBack:NSApp];
      [[view window] orderOut:NSApp];
      [[view window] setFrame:t display:NO];
    }
  [[view window] miniaturize:NSApp];
}

void x_destroy_window (struct frame *f)
// ----------------------------------------------------------------------------
//   Delete the window
// ----------------------------------------------------------------------------
{
  NSView *view = FRAME_NS_VIEW(f);
  check_ns();
  [(EmacsView *)view setWindowClosing: YES]; // may not have been informed

  ns_free_frame_faces(f);

  if (f == ns_focus_frame)
    ns_focus_frame = 0;
  if (f == ns_highlight_frame)
    ns_highlight_frame = 0;
  if (f == mouse_face_mouse_frame)
    {
      mouse_face_beg_row = mouse_face_beg_col = -1;
      mouse_face_end_row = mouse_face_end_col = -1;
      mouse_face_window = Qnil;
      mouse_face_mouse_frame = 0;
    }
  if (view == mouse_face_tracked_view)
    mouse_face_tracked_view = nil;
  xfree (f->output_data.ns);

//HACK to work around fact that on linux, 'close' is #defined to sys_close
// in s/gnu-linux.h; in emacs-21+ this won't be necessary
#if defined(__linux__) && defined(close)
#undef close
  [[view window] close];
#define close sys_close
#else
  [[view window] close];
#endif  // linux && close

  ns_window_num--;
}


int ns_get_color (const char *name, NSColor **col)
// ----------------------------------------------------------------------------
//   Parse a color name
// ----------------------------------------------------------------------------
/* On *Step, we recognize several color formats, in addition to a catalog
   of colors found in the file Emacs.clr. Color formats include:
   - #rrggbb or RGBrrggbb where rr, gg, bb specify red, green and blue in hex
   - ARGBaarrggbb is similar, with aa being the alpha channel (FF= opaque)
   - HSVhhssvv and AHSVaahhssvv are similar for hue, saturation, value
   - CMYKccmmyykk is similar for cyan, magenta, yellow, black. */
{
  NSColor * new = nil;
  const char *hex = NULL;
  enum { rgb, argb, hsv, ahsv, cmyk, gray } color_space;
  NSString *nsname = [NSString stringWithUTF8String:name];

  if ([nsname isEqualToString: @"ns_selection_color"])
    {
      nsname = ns_selection_color;
      name = [ns_selection_color UTF8String];
    }

  if (name[0] == '0' || name[0] == '1' || name[0] == '.')
    {
      // RGB decimal
      NSScanner *scanner = [NSScanner scannerWithString: nsname];
      float r,g,b;
      [scanner scanFloat: &r];
      [scanner scanFloat: &g];
      [scanner scanFloat: &b];
      *col = [NSColor colorWithCalibratedRed: r green: g blue: b alpha: 1.0];
      return 0;
    }

  if (name[0]=='#')             // X11 format
    {
      hex = name + 1;
      color_space = rgb;
    }
  else if (!memcmp(name, "RGB", 3))
    {
      hex = name + 3;
      color_space = rgb;
    }
  else if (!memcmp(name, "ARGB", 4))
    {
      hex = name + 4;
      color_space = argb;
    }
  else if (!memcmp(name, "HSV", 3))
    {
      hex = name + 3;
      color_space = hsv;
    }
  else if (!memcmp(name, "AHSV", 4))
    {
      hex = name + 4;
      color_space = ahsv;
    }
  else if (!memcmp(name, "CMYK", 4))
    {
      hex = name + 4;
      color_space = cmyk;
    }
  else if (!memcmp(name, "GRAY", 4))
    {
      hex = name + 4;
      color_space = gray;
    }

  // Direct colors (hex values)
  if (hex)
    {
      unsigned int color = 0;
      if (sscanf(hex, "%x", &color))
        {
          float f1 = ((color >> 24) & 0xff) / 255.0;
          float f2 = ((color >> 16) & 0xff) / 255.0;
          float f3 = ((color >>  8) & 0xff) / 255.0;
          float f4 = ((color      ) & 0xff) / 255.0;

          switch (color_space)
            {
            case rgb:
              *col=[NSColor colorWithCalibratedRed: f2
                                             green: f3
                                              blue: f4
                                             alpha: 1.0];
              return 0;
            case argb:
              *col=[NSColor colorWithCalibratedRed: f2
                                             green: f3
                                              blue: f4
                                             alpha: f1];
              return 0;
            case hsv:
              *col=[NSColor colorWithCalibratedHue: f2
                                        saturation: f3
                                        brightness: f4
                                             alpha: 1.0];
              return 0;
            case ahsv:
              *col=[NSColor colorWithCalibratedHue: f2
                                        saturation: f3
                                        brightness: f4
                                             alpha: f1];
              return 0;
            case gray:
              *col=[NSColor colorWithCalibratedWhite: f4 alpha: 0.0];
              return 0;
            case cmyk:
              *col=[NSColor colorWithDeviceCyan: f1
                                        magenta: f2
                                         yellow: f3
                                          black: f4
                                          alpha: 1.0];
              return 0;
            }
        }
    }

  // Otherwise, color is expected to be from a list
  {
    NSEnumerator *lenum, *cenum;
    NSString *name;
    NSColorList *clist;
#ifdef GNUSTEP
    // PENDING: who is wrong, the requestor or the implementation?
    if ([nsname compare: @"Highlight" options: NSCaseInsensitiveSearch]
        == NSOrderedSame)
      nsname = @"highlightColor";
#endif

    lenum = [[NSColorList availableColorLists] objectEnumerator];
    while ( (clist = [lenum nextObject]) && new == nil)
      {
        cenum = [[clist allKeys] objectEnumerator];
        while ( (name = [cenum nextObject]) && new == nil )
          {
            if ([name compare: nsname
                      options: NSCaseInsensitiveSearch] == NSOrderedSame )
              new = [clist colorWithKey:name];
          }
      }
  }

  if ( new )
    *col=new;
  else
    NSLog(@"Failed to find color '%@'", nsname);
  return new ? 0 : 1;
}


NSColor * ns_get_color_default (const char *name, NSColor *dflt)
// ----------------------------------------------------------------------------
//   Parse a color or use a default value
// ----------------------------------------------------------------------------
{
  NSColor * col;

  if (ns_get_color(name, &col))
    return dflt;
  else
    return col;
}


int ns_lisp_to_color (Lisp_Object color, NSColor **col)
// ----------------------------------------------------------------------------
//   Convert a Lisp string object to a NS color
// ----------------------------------------------------------------------------
{
  if (XTYPE(color) == Lisp_String)
    return ns_get_color (XSTRING(color)->data, col);
  else if (XTYPE(color) == Lisp_Symbol)
    return ns_get_color (XSYMBOL(color)->name->data, col);
  return 1;
}


Lisp_Object ns_color_to_lisp (NSColor * col)
// ----------------------------------------------------------------------------
//   Convert a color to a lisp string with the RGB equivalent
// ----------------------------------------------------------------------------
{
  float red,green,blue,alpha,gray;
  char buf[1024];
  const char *str;

  if ([[col colorSpaceName] isEqualToString: NSNamedColorSpace])

//  //    NS_DURING
//  //      {
      if ((str=[[col colorNameComponent] UTF8String]))
          return build_string((char *)str);
//      }
//    NS_HANDLER
//      {
//        // We should check here that we get an NSInvalidArgumentException.
//      }
//    NS_ENDHANDLER

    [[col colorUsingColorSpaceName:NSCalibratedRGBColorSpace]
        getRed:&red green:&green blue:&blue alpha:&alpha];
  if (red==green && red==blue)
    {
      [[col colorUsingColorSpaceName:NSCalibratedWhiteColorSpace]
            getWhite:&gray alpha:&alpha];
      sprintf(buf,"GRAY%02.2x%02.2x",
              (int)rint(gray*0xff),
              (int)rint(alpha*0xff));
      return build_string(buf);
    }

  sprintf(buf,"ARGB%02.2x%02.2x%02.2x%02.2x",(int)rint(alpha*0xff),
          (int)rint(red*0xff),(int)rint(green*0xff),(int)rint(blue*0xff));

  return build_string(buf);
}


void ns_dumpglyphs(FRAME_PTR nf,
                   int x, int y,
                   GLYPH *glyphs, int len,
                   int flags)
// ----------------------------------------------------------------------------
//   Write 'len' glyphs to target frame 'nf' at position 'x,y'.
// ----------------------------------------------------------------------------
/* Display a sequence of N glyphs found at GP.
   WINDOW is the x-window to output to.  LEFT and TOP are starting coords.
   Flag bits include:
   1: Text is highlighted
   2: Cursor is on it
   3: Appears in mouse face
   4: Draw only the foreground

   Since the display generation code is responsible for calling
   compute_char_face and compute_glyph_face on everything it puts in
   the display structure, we can assume that the face code on each
   glyph is a valid index into FRAME_COMPUTED_FACES (f), and the one
   to which we can actually apply intern_face.

   Call this function with input blocked. */
{
  static int *t[256];   // each element points to run of chars
  static unsigned int l[256];     // length of run
  static int buf[4096]; // holds chars in multiple runs
  static NSRect r[256];           // bounding rects of char run
  static int faces[256];
  static NSRect cRect;            // bounding rect for whole run
  static NSRect rectList[24];

  int *p=buf;
  int no = 0;
  int faceid, g;
  BOOL break_on_face_change = NO;
  BOOL left = FRAME_HAS_VERTICAL_SCROLL_BARS_ON_LEFT(nf);
  int firstCharPos = FIRST_CHAR_POSITION(nf);

  TRACE(dumpglyph);

  if (flags == NS_DUMPGLYPH_FOREGROUND
        && [NS_FACE_BACKGROUND(FRAME_DEFAULT_FACE(nf)) alphaComponent] != 1.0)
      return; // overdrawing foreground doesn't work well w/transparency

  cRect = NSMakeRect(0,0,0,0);
//  if (flags == NS_DUMPGLYPH_NORMAL) fprintf(stderr, "%d chars starting with %d (%d), %d (%d)\n",len,glyphs[0],FAST_GLYPH_CHAR(glyphs[0]),glyphs[1],FAST_GLYPH_CHAR(glyphs[1]));
//  if (flags == NS_DUMPGLYPH_NORMAL) fprintf(stderr, "    pad = %d, %d\trevDir = %d, %d\n",glyphs[0]&GLYPH_MASK_PADDING,glyphs[1]&GLYPH_MASK_PADDING,glyphs[0]&GLYPH_MASK_REV_DIR,glyphs[1]&GLYPH_MASK_REV_DIR);

  while (len > 0)
    {
      // Select face based on input flags
      switch(flags)
        {
        case NS_DUMPGLYPH_HIGHLIGHT:
          faceid = 1;
          break;
        case NS_DUMPGLYPH_CURSOR:
        case NS_DUMPGLYPH_BACKGROUND:
          faceid = 0;
          break;
        case NS_DUMPGLYPH_MOUSEFACE:
          faceid = mouse_face_face_id;
          break;
        default:
          g = *glyphs;
          GLYPH_FOLLOW_ALIASES (GLYPH_TABLE_BASE, GLYPH_TABLE_LENGTH, g);
          faceid = FAST_GLYPH_FACE(g);
          break_on_face_change = YES;
          break;
        }

      faces[no] = faceid;

        // Add glyphs to current run of characters
      if (flags != NS_DUMPGLYPH_BACKGROUND)
        {
          int spaceRun = 0;
          for (t[no] = p; len > 0; len--)
            {
              g = *glyphs;
              GLYPH_FOLLOW_ALIASES (GLYPH_TABLE_BASE, GLYPH_TABLE_LENGTH, g);
              if (break_on_face_change && FAST_GLYPH_FACE(g) != faceid
                      && !(FAST_GLYPH_FACE(g) == 0 && FAST_GLYPH_CHAR(g) == 0))
                  break;          // End of run if face changes
              ///HACK to keep left scrollbar from being overwritten
              if (!left || (x + p-t[no]) >= firstCharPos || faceid == 1)
                {
                  *p++ = FAST_GLYPH_CHAR(g);
                  // HACK to add overhang when rendering engine clears text
                  // by rendering spaces where words were (aargh)
                  if (isspace(FAST_GLYPH_CHAR(g)))
                    spaceRun++;
                  else
                    spaceRun = 0;
                }
              else
                {
                  x++;
                }
              glyphs++;
            }
          if (p - t[no] > 0)
            {
              ns_chars_to_rect(nf, x, y, x + (p - t[no]) - 1, y, &r[no]);
              if (spaceRun > 1)
                r[no].size.width += 3;  // HACK: see above
              cRect = NSUnionRect(cRect, r[no]);
              l[no] = p - t[no];
              x += l[no];
              *p = '\0';
              no++;
            }
        }
      else // NS_DUMPGLYPH_BACKGROUND
        {
          ///HACK to keep scrollbar from being overwritten until the real
          //  cause of the problem can be found
          // (added for GNUstep but doesn't hurt OS X)
          if (left && (x < firstCharPos) && faceid != 1)
            x = firstCharPos;
          ns_chars_to_rect(nf, x, y, x + len - 1, y, &r[no]);
          cRect = NSUnionRect(cRect, r[no]);
          t[no] = p;
          l[no] = 0;
          len = 0;
          no++;
        }
      if (no >= 256)
        fprintf(stderr, "ASSERTION FAILURE! no > 256!\n");
    } // while glyphs remaining


  // Display assembled glyph runs
  if (no > 0)
    {
      int i;

      ns_focus(nf, &cRect);

      // Draw the text itself
      for (i = 0; i < no; i++)
        {

          // sort out face
          int face_idx = ((faces[i] < -1 ||
                           faces[i] >= FRAME_N_COMPUTED_FACES(nf) ||
                           FRAME_COMPUTED_FACE(nf, faces[i]) == 0)
                          ? 0
                          : faces[i]);
          struct face *face = FRAME_COMPUTED_FACE(nf, face_idx);
          XFontStruct *font = face->font == ~0 ? FRAME_FONT(nf) : face->font;

          // Fill background color unless drawing over cursor or fg-only
          if (flags!=NS_DUMPGLYPH_CURSOR && flags!=NS_DUMPGLYPH_FOREGROUND)
            {
              EmacsView *view = FRAME_NS_VIEW(nf);
              NSRectArray sbRects = view->sbRects;
              int nsbRects = view->nsbRects;

              if (!face->stipple)
                [(NS_FACE_BACKGROUND(face)!=nil?
                  NS_FACE_BACKGROUND(face):FRAME_BACKGROUND_COLOR(nf)) set];
              else
                [[NSColor colorWithPatternImage:
                            [face->stipple stippleRep]] set];

              if (nsbRects > 1 && face_idx != 1)
                {
                  int j;
                  int n=0;
                  NSRect tRect = r[i];
                  NSRect iRect;
                  for (j=0; j<nsbRects; j++)
                    {
                      NSRect sRect = sbRects[j];
                      //HACK
#ifdef MACOSX
                      sRect.origin.x -= 2; sRect.size.height += 34; sRect.origin.y -= 34;
#else
                      sRect.origin.x -= 1; sRect.size.height += 32; sRect.origin.y -= 32;
#endif
                      iRect = NSIntersectionRect(sRect, tRect);
                      if (!NSIsEmptyRect(iRect))
                        {
                          NSRect before = tRect, after = tRect;
                          before.size.width = iRect.origin.x - tRect.origin.x;
                          after.origin.x = iRect.origin.x + iRect.size.width;
                          after.size.width -= 
                            iRect.origin.x+iRect.size.width - tRect.origin.x;
                          if (!NSIsEmptyRect(before))
                            {
                              rectList[n++] = before;
//if (!NSIsEmptyRect(NSIntersectionRect(before,sbRects[j]))) fprintf(stderr,"NONEMPTY (1)!\n");
                            }
                          tRect = after;
                        }
                    }
                  if (!NSIsEmptyRect(tRect))
                    {
                      rectList[n++] = tRect;
//if (!NSIsEmptyRect(NSIntersectionRect(tRect,sbRects[0]))) fprintf(stderr,"NONEMPTY (2)!\n");
                    }
//fprintf(stderr, "filling %d rects...\n",n);
//                      { int k,l;
//                      for (k=0; k<n; k++)
//                          for (l=0; l<nsbRects; l++)
//                              if (!NSIsEmptyRect(NSIntersectionRect(rectList[k],sbRects[l]))) fprintf(stderr,"NONEMPTY: %d, %d!\n",k,l);
//                      }
                  NSRectFillList(rectList, n);
                }
              else
                {
                  NSRectFill(r[i]);
                }
            }

          if (l[i] > 0)
            {
              nswrap_moveshow(r[i].origin.x, r[i].origin.y,
                              t[i], l[i],
                              font,
                              (NS_FACE_FOREGROUND(face)!=nil ?
                               NS_FACE_FOREGROUND(face)
                               : FRAME_FOREGROUND_COLOR(nf)),
                              (flags!=NS_DUMPGLYPH_FOREGROUND ? nil
                               : (NS_FACE_BACKGROUND(face)!=nil ?
                                  NS_FACE_BACKGROUND(face)
                                  : FRAME_BACKGROUND_COLOR(nf))),
                              face->underline,
                              //PENDING: do in ns_frame_update_line_height
                              font->voffset + (NS_FRAME_LINE_HEIGHT(nf) -
                                               font->height)/2,
                              NS_FONT_WIDTH(nf));
            }
        }
    }

}


// mask is essentially a developer tool, allowing to turn off cursor
// dumps from selected parts of the program; the calls to ns_dumpcursor
// in this file each use a unique integer
int dumpcursor_mask = (1<<3) /*| (1<<17) | (1<<16)*/;
void ns_dumpcursor(int where, FRAME_PTR f,int nx,int ny)
// ----------------------------------------------------------------------------
//   Move the cursor to new indicated coordinates
// ----------------------------------------------------------------------------
{
  NSRect r,s;
  int pad;
  // preserve prev pos across jumps, unlike px,py
  static int lastX = -1, lastY = -1;

  TRACE(dumpcursor);
  if (dumpcursor_mask & (1<<where))
    return;

  // Fast exit if everything is drawn OK
  if (f->phys_cursor_x == nx && f->phys_cursor_y == ny &&
      FRAME_CURSOR(f) == FRAME_NEW_CURSOR(f) &&
      [FRAME_CURSOR_COLOR(f) isEqual: FRAME_NEW_CURSOR_COLOR(f)])
    return;

  //PENDING: clean up this code
  {
    struct frame_glyphs *current_glyphs = FRAME_CURRENT_GLYPHS (f);
    int px = f->phys_cursor_x;
    int py = f->phys_cursor_y;
    BOOL enabled = nx >= 0 && ny >= 0 && current_glyphs->enable[ny];
    int max_x = ny >= 0 ? current_glyphs->used[ny] : 0;

    ///
    int faceid = FAST_GLYPH_FACE(lastX >= FIRST_CHAR_POSITION(f) ?
                            current_glyphs->glyphs[lastY][lastX] : SPACEGLYPH);
    XFontStruct *font = faceid > -1 && faceid < FRAME_N_COMPUTED_FACES(f)
                     && FRAME_COMPUTED_FACE(f, faceid) != 0 ?
                 FRAME_COMPUTED_FACE(f, faceid)->font
                 : 0;
    BOOL ital = font ? font->synthItal ||
                 [[NSFontManager sharedFontManager] traitsOfFont: font->nsfont]
                     & NSItalicFontMask
                 : NO;

    // draw previous position(s) back to normal
    if (px >= FIRST_CHAR_POSITION(f) && py >= 0)
      {
        pad = px+1 < current_glyphs->used[py] &&
          current_glyphs->glyphs[py][px+1] & GLYPH_MASK_PADDING ? 1 : 0;
        ns_dumpglyphs(f, px, py, &(f->phys_cursor_glyph), 1+pad,
                      NS_DUMPGLYPH_NORMAL);
        // correct for synthetic italic clipping, case 1
        if (ital && px-1 >= FIRST_CHAR_POSITION(f) && px == nx-1 && py==ny)
          {
            int cp = (current_glyphs->glyphs[py][px-1] & GLYPH_MASK_PADDING)
              ? px-2 : px-1;
            ns_dumpglyphs(f, cp, py, &(current_glyphs->glyphs[py][cp]), 1,
                          NS_DUMPGLYPH_FOREGROUND);
          }
      }

    // correct for synthetic italic clipping, case 2
    if (ital && lastX-1 >= FIRST_CHAR_POSITION(f) && lastX == nx-1 && lastY==ny)
      {
        int cp = (current_glyphs->glyphs[lastY][lastX-1]
                  & GLYPH_MASK_PADDING) ? lastX-2 : lastX-1;
        ns_dumpglyphs(f, cp, lastY, &(current_glyphs->glyphs[lastY][cp]),
                      1, NS_DUMPGLYPH_FOREGROUND);
      }

    // update to new position
    if (nx != -1)
      {
        f->phys_cursor_glyph = ((enabled && (nx < max_x))
                                ? current_glyphs->glyphs[ny][nx]
                                : SPACEGLYPH);
        pad = (nx+1<max_x)
          && current_glyphs->glyphs[ny][nx+1] & GLYPH_MASK_PADDING ? 1 : 0;
      }
  }

  if (nx != -1)
    {
      lastX = nx;
      lastY = ny;
    }
  f->phys_cursor_x = nx;
  f->phys_cursor_y = ny;

  FRAME_CURSOR(f) = FRAME_NEW_CURSOR(f);
  FRAME_CURSOR_COLOR(f) = FRAME_NEW_CURSOR_COLOR(f);

  if (nx >= FIRST_CHAR_POSITION(f) && ny >= 0)
    {
      // draw the cursor
      ns_chars_to_rect(f, nx,ny, nx+pad,ny, &r);
      ns_focus(f, &r);
      [FRAME_BACKGROUND_COLOR(f) set];
      NSRectFill(r);
      [FRAME_CURSOR_COLOR(f) set];
      switch(FRAME_CURSOR(f))
        {
        case no_highlight:
          break;
        case filled_box:
          NSRectFill(r);
          break;
        case hollow_box:
          NSRectFill(r);
          [FRAME_BACKGROUND_COLOR(f) set];
          NSRectFill(NSInsetRect(r, 1, 1));
          [FRAME_CURSOR_COLOR(f) set];
          break;
        case bar:
          s=r;
          s.origin.y += 0.75 * s.size.height;
          s.size.height *= 0.25;
          NSRectFill(s);
          break;
        case line:
          s=r;
          s.size.width = 2;
          NSRectFill(s);
          break;
        }
      // draw the character under the cursor
      ns_dumpglyphs(f, nx, ny, &(f->phys_cursor_glyph), 1,
                    NS_DUMPGLYPH_CURSOR);
    }

  ns_unfocus(f);
}


static void ns_cursor_to(int row, int col)
{
  struct frame *f = (updating_frame ? updating_frame : selected_frame);

  TRACE(ns_cursor_to);

  curs_x = col;
  curs_y = row;

  if (f != updating_frame)
    ns_dumpcursor(2, f,curs_x,curs_y);
}

void x_set_offset (struct frame *f, int xoff, int yoff, int change_grav)
{
  NSScreen *screen;
  NSView *view = FRAME_NS_VIEW(f);

  TRACE(x_set_offset);
  f->output_data.ns->left = xoff;
  f->output_data.ns->top = yoff;
#ifdef GNUSTEP
  if (xoff < 100)
    f->output_data.ns->left = 100;  // don't overlap menu
#endif
  if (view != nil && (screen = [[view window] screen]))
    [[view window]
            setFrameTopLeftPoint:
        NSMakePoint(SCREENMAXBOUND(f->output_data.ns->left),
                    SCREENMAXBOUND([screen frame].size.height -
                                   f->output_data.ns->top))];
}


static unsigned ns_convert_key(unsigned code)
{
  const unsigned last_keysym = (sizeof(convert_ns_to_X_keysym)
                                / sizeof(convert_ns_to_X_keysym[0]));
  unsigned keysym;
  for (keysym = 0; keysym < last_keysym; keysym += 2)
    if (code == convert_ns_to_X_keysym[keysym])
      return convert_ns_to_X_keysym[keysym+1];
  return 0;
}


static void ns_adjust_size(struct frame *f)
{
  int rows, cols, vbextra, fw, lh;
  NSView *view = FRAME_NS_VIEW(f);
  NSWindow *win = [view window];
  NSRect r;

  TRACE(ns_adjust_size);

  [win invalidateCursorRectsForView:view];

  // don't set up tracking rects as long as the window is deferred
  if ([win windowNumber] > 0)
    [(EmacsView *)view adjustTrackingRects];

  r = [view bounds];

  fw = NS_FONT_WIDTH(f);
  if (fw <= 0)
    cols = f->width;
  else
    {
      vbextra = NS_SCROLL_BAR_WIDTH(f);
      f->output_data.ns->vertical_scroll_bar_extra = vbextra;
      cols = floor((r.size.width - vbextra) / fw + 0.5);
    }

  lh = NS_FRAME_LINE_HEIGHT(f);
  if (lh <= 0)
    rows = f->height;
  else
    rows = floor(r.size.height / lh + 0.5);

  change_frame_size (f, rows, cols, 0, 1);
  SET_FRAME_GARBAGED (f);
}

// sets frame pixel size based on character dimensions
void x_set_window_size (struct frame *f, int change_grav, int cols, int rows)
{
  NSView *view = FRAME_NS_VIEW(f);
  NSWindow *window;
  NSScreen *screen;
  NSRect r,wr;
  int vbextra, fw, lh;
  static BOOL reentrant = NO;
//     static int lastFW = 0, lastLH = 0;

  if (reentrant)
    return;
  reentrant = YES;
  /// this optimization fails in at least GNUstep on user-directed resize;
  //  perhaps a better check can be done..

//BOOL isOld = (view != nil && ((EmacsView *)view)->rows == rows && ((EmacsView *)view)->cols == cols && NS_FONT_WIDTH(f)==lastFW && NS_FRAME_LINE_HEIGHT(f)==lastLH);
//if (isOld) return;
//if (view != nil)
// {
//   lastFW = NS_FONT_WIDTH(f);
//   lastLH = NS_FRAME_LINE_HEIGHT(f);
//}

  TRACE(x_set_window_size);
//fprintf(stderr, "setWindowSize: %d x %d - %s\n",cols,rows, (isOld ?"old":"new"));

  if (view!=nil /*&& [[view window] windowNumber] > 0*/)
    {
      window=[view window];
      screen=[window screen];
      r = [view frame];
      wr = [window frame];

      ns_dumpcursor(3, f,-1,-1);

      fw = NS_FONT_WIDTH(f);
      lh = NS_FRAME_LINE_HEIGHT(f);
//fprintf(stderr, "setWindowSize: %d x %d, font = %d x %d\n",cols,rows,fw,lh);

      vbextra = NS_SCROLL_BAR_WIDTH(f);
      f->output_data.ns->vertical_scroll_bar_extra = vbextra;

      r.origin.x    = f->output_data.ns->internal_border_width;
      r.origin.y    = f->output_data.ns->internal_border_width;
      r.size.width  = (fw * cols) + vbextra;
      r.size.height = lh * rows;

      wr.size.width = r.size.width +
        f->output_data.ns->internal_border_width*2 +
        f->output_data.ns->border_width;
      wr.size.height = r.size.height +
        f->output_data.ns->internal_border_width*2 +
        f->output_data.ns->border_height;

      [((EmacsView *) view) setRows: rows andColumns: cols];
//fprintf(stderr,"\trect: (%f, %f)\t%f x %f\n", r.origin.x, r.origin.y, r.size.width, r.size.height);
      /*if (!isOld)*/ [view setFrame:r];

      if (screen)
        {
          wr.origin.x = f->output_data.ns->left;
          wr.origin.y = ([screen frame].size.height -
                         f->output_data.ns->top -
                         wr.size.height);
        }

      wr.origin.x=IN_BOUND(-SCREENMAX,wr.origin.x,SCREENMAX);
      wr.origin.y=IN_BOUND(-SCREENMAX,wr.origin.y,SCREENMAX);
      [window setFrame:wr display:NO];
//fprintf(stderr,"x_set_window_size %d, %d\t%.0f, %.0f\n",rows,cols,r.size.width,r.size.height);

      ns_adjust_size(f);
    }
  else
    {
      change_frame_size(f, rows, cols, 0, 0);
      SET_FRAME_GARBAGED(f);
    }
  reentrant = NO;
}

void x_set_mouse_pixel_position (struct frame *f, int pix_x, int pix_y)
{
  TRACE(x_set_mouse_pixel_position);
  ns_raise_frame(f);
  //PENDING: find a portable, working way to do this
#ifndef GNUSTEP
  [FRAME_NS_VIEW(f) lockFocus];
  PSsetmouse((float)pix_x,(float)pix_y);
  [FRAME_NS_VIEW(f) unlockFocus];
#endif
}

int x_set_mouse_position (struct frame *f, int x, int y)
{
  NSRect r;

  TRACE(x_set_mouse_position);
  ns_chars_to_rect(f,x,y,x,y,&r);
  x_set_mouse_pixel_position(f,r.origin.x+r.size.width/2,
                             r.origin.y+r.size.height/2);
  return 0;
}

/*
  Detects whether the mouse is moving on a different frame from before, or,
  if on the same frame, on a different character.
 */
static void
note_mouse_movement (struct frame *frame, int x, int y)
{
  TRACE(note_mouse_movement);
//fprintf(stderr, "note_mouse_movement, given %d,%d - lastPos %.0f, %.0f\n",x,y,last_mouse_glyph.origin.x,last_mouse_glyph.origin.y);

  if ((frame!=last_mouse_frame)||
      (x<last_mouse_glyph.origin.x)||
      (x>=(last_mouse_glyph.origin.x+last_mouse_glyph.size.width))||
      (y<last_mouse_glyph.origin.y)||
      (y>=(last_mouse_glyph.origin.y+last_mouse_glyph.size.height)))
    {
      frame->mouse_moved = 1;
      last_mouse_frame = frame;
      // mouse-faces are updated elsewhere
      // note_mouse_highlight (frame, x, y);
    }
}

/* This is used for debugging, to turn off note_mouse_highlight.  */
static int disable_mouse_highlight = 0;

/* Take proper action when the mouse has moved to position X, Y on frame F
   as regards highlighting characters that have mouse-face properties.
   Also dehighlighting chars where the mouse was before.  */
static void note_mouse_highlight (struct frame *f, int x, int y)
{
  int row, column, portion;
  Lisp_Object window;
  struct window *w;
  int leftOffset = FRAME_HAS_VERTICAL_SCROLL_BARS_ON_LEFT(f) ?
    FRAME_SCROLL_BAR_COLS(f) : 0;
  TRACE(note_mouse_highlight);

  if (disable_mouse_highlight)
    return;
  
  mouse_face_mouse_x = x;
  mouse_face_mouse_y = y;
  mouse_face_mouse_frame = f;
  
  /* Find out which glyph the mouse is on.  */
  pixel_to_glyph_coords (f, x, y, &column, &row, 0, 1);
  
  mouse_face_mouse_row = row;
  // set mouse_face_mouse_rect to cover current glyph for when we're
  // not in a mouse-face
  ns_chars_to_rect (f, column, row, column, row, &mouse_face_mouse_rect);
//fprintf(stderr, "mfm_rect 1: %d,%d + %d,%d went to %.0f,%.0f + %d, %d\n",
//        column,row,column,row, mouse_face_mouse_rect.origin.x,
//        mouse_face_mouse_rect.origin.y, (column-column+1), (row-row+1));

  if (mouse_face_defer)
    return;
  
  if (gc_in_progress)
    {
      mouse_face_deferred_gc = 1;
      return;
    }
  
  /* Which window is that in?  */
  window = window_from_coordinates (f, column, row, &portion);
  w = XWINDOW (window);

  w->left += leftOffset;
  
  /* If we were displaying active text in another window, clear that.  */
  if (! EQ (window, mouse_face_window))
    clear_mouse_face ();
  
    /* Are we in a window whose display is up to date?
       And verify the buffer's text has not changed.  */
  if (WINDOWP (window) && portion == 0
      && row >= 0 && column >= leftOffset
      && row < FRAME_HEIGHT (f) && column < FRAME_WIDTH (f)+leftOffset
      && EQ (w->window_end_valid, w->buffer)
      && w->last_modified == BUF_MODIFF (XBUFFER (w->buffer)))
    {
      int *ptr = FRAME_CURRENT_GLYPHS (f)->charstarts[row];
      int i, pos;
      
      /* Find which buffer position the mouse corresponds to.  */
      for (i = column; i >= leftOffset; i--)
        if (ptr[i] > leftOffset)
          break;
      pos = ptr[i];
      /* Is it outside the displayed active region (if any)?  */
      if (pos <= leftOffset)
        clear_mouse_face ();
      else if (! (EQ (window, mouse_face_window)
                  && row >= mouse_face_beg_row
                  && row <= mouse_face_end_row
                  && (row > mouse_face_beg_row
                      || column >= mouse_face_beg_col)
                  && (row < mouse_face_end_row
                      || column < mouse_face_end_col
                      || mouse_face_past_end)))
        {
          Lisp_Object mouse_face, overlay, position;
          Lisp_Object *overlay_vec;
          int len, noverlays;
          struct buffer *obuf;
          int obegv, ozv;
          
            /* If we get an out-of-range value, return now; avoid an error.  */
          if (pos > BUF_Z (XBUFFER (w->buffer)))
            {
              w->left -= leftOffset;
              return;
            }
          
          /* Make the window's buffer temporarily current for
               overlays_at and compute_char_face.  */
          obuf = current_buffer;
          current_buffer = XBUFFER (w->buffer);
          obegv = BEGV;
          ozv = ZV;
          BEGV = BEG;
          ZV = Z;
          
            /* Yes.  Clear the display of the old active region, if any.  */
          clear_mouse_face ();
          
          /* Is this char mouse-active?  */
          XSETINT (position, pos);
          
          len = 10;
          overlay_vec = (Lisp_Object *) xmalloc (len * sizeof (Lisp_Object));
          
          /* Put all the overlays we want in a vector in overlay_vec.
               Store the length in len.  */
          noverlays = overlays_at (XINT (pos), 1, &overlay_vec, &len,
                                   NULL, NULL);
          noverlays = sort_overlays (overlay_vec, noverlays, w);
          
            /* Find the highest priority overlay that has a mouse-face prop. */
          overlay = Qnil;
          for (i = 0; i < noverlays; i++)
            {
              mouse_face = Foverlay_get (overlay_vec[i], Qmouse_face);
              if (!NILP (mouse_face))
                {
                  overlay = overlay_vec[i];
                  break;
                }
            }
          free (overlay_vec);
          /* If no overlay applies, get a text property.  */
          if (NILP (overlay))
            mouse_face =
              Fget_text_property (position, Qmouse_face, w->buffer);
          
          /* Handle the overlay case.  */
          if (! NILP (overlay))
            {
              /* Find the range of text around this char that
                 should be active.  */
              Lisp_Object before, after;
              int ignore;
              
              before = Foverlay_start (overlay);
              after = Foverlay_end (overlay);
              /* Record this as the current active region.  */
              fast_find_position (window, before, ozv,
                                  &mouse_face_beg_col,
                                  &mouse_face_beg_row, leftOffset);
              mouse_face_past_end
                = !fast_find_position (window, after, ozv,
                                       &mouse_face_end_col,
                                       &mouse_face_end_row, leftOffset);
              mouse_face_window = window;
              mouse_face_face_id
                = compute_char_face (f, w, pos, 0, 0,
                                     &ignore, pos + 1, 1);
              
              /* Display it as active.  */
              show_mouse_face (1);
            }
          /* Handle the text property case.  */
          else if (! NILP (mouse_face))
            {
              /* Find the range of text around this char that
                 should be active.  */
              Lisp_Object before, after, beginning, end;
              int ignore;
              
              beginning = Fmarker_position (w->start);
              XSETINT (end, (BUF_Z (XBUFFER (w->buffer)) -
                             XFASTINT (w->window_end_pos)));
              
              before
                = Fprevious_single_property_change (make_number (pos + 1),
                                                    Qmouse_face,
                                                    w->buffer, beginning);
              after
                = Fnext_single_property_change (position, Qmouse_face,
                                                w->buffer, end);
              /* Record this as the current active region.  */
              fast_find_position (window, before, ozv,
                                  &mouse_face_beg_col,
                                  &mouse_face_beg_row, leftOffset);
              mouse_face_past_end
                = !fast_find_position (window, after, ozv,
                                       &mouse_face_end_col,
                                       &mouse_face_end_row, leftOffset);
              mouse_face_window = window;
              mouse_face_face_id
                = compute_char_face (f, w, pos, 0, 0,
                                     &ignore, pos + 1, 1);
              
              /* Display it as active.  */
              show_mouse_face (1);
            }
          /* Handle the no-mouse-face case */
          else
            {
              Lisp_Object tbefore, tafter, obefore, oafter;
              Lisp_Object beginning, end;
              int bcol, brow, ecol, erow;
              int width = window_internal_width (w) + w->left;
              int height = window_internal_height (w) + w->top;
              
              beginning = Fmarker_position (w->start);
              XSETINT (end, (BUF_Z (XBUFFER (w->buffer)) -
                             XFASTINT (w->window_end_pos)));
              
              if (pos+1 >= XFASTINT (end))
                {
                  fast_find_position (window, position, ozv,
                                      &bcol, &brow, leftOffset);
                  ecol = width;
                  erow = height;
                  
                  if (bcol > column)
                    {
                      bcol = w->left;
                      brow++;
                    }
                }
              else
                {
                  int after, before;
                  
                  tbefore =
                    Fprevious_single_property_change (make_number(pos+1),
                                                      Qmouse_face,
                                                      w->buffer,beginning);
                  tafter =
                    Fnext_single_property_change (position, Qmouse_face,
                                                  w->buffer, end);
                  
                  obefore = Fprevious_overlay_change (make_number(pos+1));
                  oafter = Fnext_overlay_change (position);
                  
                  if ((before = max (XFASTINT (tbefore),
                                     XFASTINT (obefore)))
                      == XFASTINT (beginning))
                    {
                      bcol = w->left;
                      brow = w->top;
                    }
                  else
                    {
                      fast_find_position (window, before, ozv,
                                          &bcol, &brow, leftOffset);
                      
                      if (brow != row && bcol != (w->left+2))
                        {
                          brow++;
                          bcol = w->left;
                        }
                    }
                  
                  if ((after = min (XFASTINT (tafter),
                                    XFASTINT (oafter))) == XFASTINT (end))
                    {
                      ecol = width;
                      erow = height;
                    }
                  else
                    {
                      fast_find_position (window, after, ozv,
                                          &ecol, &erow, leftOffset);
                      
                      if (erow != row)
                        {
                          erow--;
                          ecol = width;
                        }
                      // we are in the row, but the region apparently
                      // ends before column.  This happens with
                      // invisible text, so we extend the region to the
                      // end of the line.
                      else if (ecol < column)
                        ecol = width;
                    }
                }
              ns_chars_to_rect (f, bcol, brow, ecol-1, erow,
                                &mouse_face_mouse_rect);
// fprintf(stderr, "mfm_rect 2: %d,%d + %d,%d went to %.0f,%.0f + %d, %d\n",
//         bcol,brow,ecol-1,erow, mouse_face_mouse_rect.origin.x,
//         mouse_face_mouse_rect.origin.y, (ecol-1-bcol+1), (erow-brow+1));
            }
          BEGV = obegv;
          ZV = ozv;
          current_buffer = obuf;
        }
    }
  w->left -= leftOffset;
}

/* Find the row and column of position POS in window WINDOW.
   Store them in *COLUMNP and *ROWP.
   This assumes display in WINDOW is up to date.
   If POS is above start of WINDOW, return coords
   of start of first screen line.
   If POS is after end of WINDOW, return coords of end of last screen line.

   Value is 1 if POS is in range, 0 if it was off screen.  */

static int
fast_find_position (Lisp_Object window,
                    int pos, int end, int *columnp, int *rowp,
                    int leftOffset)
{
  struct window *w = XWINDOW (window);
  FRAME_PTR f = XFRAME (WINDOW_FRAME (w));
  int i;
  int row = 0;
  int left = w->left;
  int top = w->top;
  int height = XFASTINT (w->height) - ! MINI_WINDOW_P (w);
  int width = window_internal_width (w);
  int *charstarts;
  int lastcol;
  int maybe_next_line = 0;

  TRACE(fast_find_position);
  /* Find the right row.  */
  for (i = 0;
       i < height;
       i++)
    {
      int linestart = FRAME_CURRENT_GLYPHS (f)->charstarts[top + i][left];
      if (linestart > pos)
        break;
      /* If the position sought is the end of the buffer,
         don't include the blank lines at the bottom of the window.  */
      if (linestart == pos && pos == end) // was BUF_ZV (XBUFFER(w->buffer)))
        {
          maybe_next_line = 1;
          break;
        }
      if (linestart > 0)
        row = i;
    }

  /* Find the right column with in it.  */
  charstarts = FRAME_CURRENT_GLYPHS (f)->charstarts[top + row];
  lastcol = left;
  for (i = leftOffset; i < width+leftOffset; i++)
    {
      if (charstarts[left + i] == pos)
        {
          *rowp = row + top;
          *columnp = i + left;
          return 1;
        }
      else if (charstarts[left + i] > pos)
        break;
      else if (charstarts[left + i] > leftOffset)
        lastcol = left + i;
    }

  /* If we're looking for the end of the buffer,
       and we didn't find it in the line we scanned,
       use the start of the following line.  */
  if (maybe_next_line)
    {
      row++;
      lastcol = left;
    }

  *rowp = row + top;
  *columnp = lastcol;
  return 0;
}

static void show_mouse_face (int hl)
{
  struct window *w = XWINDOW (mouse_face_window);
  int width = window_internal_width (w);
  FRAME_PTR f = XFRAME (WINDOW_FRAME (w));
  int i;
  int cursor_off = 0;
  int old_curs_x = curs_x;
  int old_curs_y = curs_y;

  TRACE(show_mouse_face);

  /* Set these variables temporarily
       so that if we have to turn the cursor off and on again
       we will put it back at the same place.  */
  curs_x = f->phys_cursor_x;
  curs_y = f->phys_cursor_y;

  for (i = mouse_face_beg_row;
       i <= mouse_face_end_row; i++)
    {
      int column = (i == mouse_face_beg_row
                    ? mouse_face_beg_col : w->left);
      int endcolumn = (i == mouse_face_end_row
                       ? mouse_face_end_col : w->left + width);
      endcolumn = min (endcolumn, FRAME_CURRENT_GLYPHS (f)->used[i]);

      /* If the cursor's in the text we are about to rewrite,
         turn the cursor off.  */
      if (i == curs_y
          && curs_x >= column - 1
          && curs_x <= endcolumn)
        {
          ns_dumpcursor (4, f, -1, -1);
          cursor_off = 1;
        }

      ns_dumpglyphs (f, column, i,   // r.origin.x, r.origin.y,
                     FRAME_CURRENT_GLYPHS (f)->glyphs[i] + column,
                     endcolumn - column,
                     /* Highlight with mouse face if hl > 0.  */
                     hl > 0 ? NS_DUMPGLYPH_MOUSEFACE : NS_DUMPGLYPH_NORMAL);
      //PENDING: this was throwing off focus rect mgmt during updates, but
      // maybe it's OK now
      //ns_unfocus(f);

      if (i == mouse_face_mouse_row)
        {
          ns_chars_to_rect (f, column, i, endcolumn-1, i,
                            &mouse_face_mouse_rect);
        }

    }

  /* If we turned the cursor off, turn it back on.  */
  if (cursor_off)
    {
      ns_dumpcursor(5, f,f->cursor_x,f->cursor_y);
    }

  curs_x = old_curs_x;
  curs_y = old_curs_y;
  ns_unfocus(f);
}

static void clear_mouse_face ()
{
  TRACE(clear_mouse_face);
  if (! NILP (mouse_face_window))
    {
      struct window *w = XWINDOW (mouse_face_window);
      int leftOffset
        = FRAME_HAS_VERTICAL_SCROLL_BARS_ON_LEFT(mouse_face_mouse_frame)
        ? FRAME_SCROLL_BAR_COLS(mouse_face_mouse_frame) : 0;
      w->left += leftOffset;
      show_mouse_face (0);
      w->left -= leftOffset;
    }

  mouse_face_beg_row = mouse_face_beg_col = -1;
  mouse_face_end_row = mouse_face_end_col = -1;
  mouse_face_window = Qnil;
}

// called to adjust frame size when main frame font changed
Lisp_Object ns_new_font (struct frame *f, char *fontname)
{
  struct font_info *fontp =
    fs_load_font (f, FRAME_NS_FONT_TABLE (f), CHARSET_ASCII, fontname,-1);
  int fw;

  TRACE(ns_new_font);

  if (!fontp)
    return Qnil;

  f->output_data.ns->font = (XFontStruct *) (fontp->font);

  /* Compute the scroll bar width in character columns.  */
  fw = NS_FONT_WIDTH(f);
  if (f->scroll_bar_pixel_width > 0)
    f->scroll_bar_cols = (f->scroll_bar_pixel_width + fw - 1) / fw;
  else
    f->scroll_bar_cols = (NS_SCROLL_BAR_WIDTH_DEFAULT + fw - 1) / fw;

    /* Now make the frame display the given font.  */
  if (FRAME_NS_VIEW (f) != 0)
    {
      ns_frame_update_line_height(f);
      // do this whether or not height changed: width may have changed
      x_set_window_size (f, 0, f->width, f->height);
    }
  else
    /* If we are setting a new frame's font for the first time,
           there are no faces yet, so this font's height is the line height. */
    f->output_data.ns->line_height = FONT_HEIGHT (f->output_data.ns->font);

  return build_string (fontp->full_name);
}

static void ns_clear_frame(void)
{
  struct frame *f=(updating_frame ? updating_frame : selected_frame);
  NSView *view = FRAME_NS_VIEW(f);
  NSRect r;
  int onLeft = FRAME_HAS_VERTICAL_SCROLL_BARS_ON_LEFT(f);
  int onRight = FRAME_HAS_VERTICAL_SCROLL_BARS_ON_RIGHT(f);
  int sWidth = (onLeft || onRight) ? NS_SCROLL_BAR_WIDTH(f) : 0;
  int i;
  NSArray *subs;

  TRACE(ns_clear_frame);
  ns_dumpcursor(6, f,-1,-1);

  curs_x=curs_y=0;

  [[view window] display];
  r = [view bounds];

  if (onLeft)
    r.origin.x += sWidth;
  else if (onRight)
    r.size.width -= sWidth;

  ns_focus(f, &r);
  [NS_FACE_BACKGROUND(FRAME_DEFAULT_FACE(f)) set];
  NSRectFill(r);
  ns_unfocus(f);

  subs = [[[view window] contentView] subviews];
  for (i=[subs count]-1; i>=0; i--)
    if ([[subs objectAtIndex: i] isKindOfClass: [EmacsScroller class]])
      [[subs objectAtIndex: i] display];
}

static void ns_clear_end_of_line(int first_unused)
{
  struct frame *f = (updating_frame ? updating_frame : selected_frame);
  BOOL left = FRAME_HAS_VERTICAL_SCROLL_BARS_ON_LEFT(f);
  BOOL inMiniBuf = curs_y >= XFASTINT(XWINDOW(FRAME_MINIBUF_WINDOW(f))->top);
  BOOL needsCursor = NO;

  TRACE(ns_clear_end_of_line);
  // note, first_unused refers to never-used columns at right edge

  if (curs_y < 0 || curs_y >= f->height)
    return;
  if (first_unused <= 0)
    return;
  first_unused++;  // chop off italic overhang
  if (first_unused > f->width)
    first_unused = f->width;

  if (left)
    first_unused += 2 * FIRST_CHAR_POSITION(f);
  else if (inMiniBuf)
    first_unused += FRAME_SCROLL_BAR_COLS(f);

  if (curs_y == f->phys_cursor_y
      && curs_x <= f->phys_cursor_x && f->phys_cursor_x < first_unused)
    {
      ns_dumpcursor(7, f,-1,-1);
      needsCursor = YES;
    }

  ns_dumpglyphs(f, curs_x, curs_y,
                NULL, first_unused - curs_x,
                NS_DUMPGLYPH_BACKGROUND);


  if (needsCursor && f != updating_frame)
    ns_dumpcursor(8, f,f->cursor_x,f->cursor_y);
  ns_unfocus(f);
}

#if defined(MACOSX) || defined(GNUSTEP)
enum NSAppKitPatch { NSNullObject = 0 };
#endif

static void ns_ins_del_lines(int vpos, int n)
// ----------------------------------------------------------------------------
//   Insert or delete n lines at line vpos
// ----------------------------------------------------------------------------
{
  struct frame *f = (updating_frame ? updating_frame : selected_frame);
  int height = n * NS_FRAME_LINE_HEIGHT(f);
  NSPoint destOrigin;
  NSRect srcRect, blkRect, modRect;
  // blkRect = part of copy src not overlapped by copy destination
  BOOL left = FRAME_HAS_VERTICAL_SCROLL_BARS_ON_LEFT(f);

  TRACE(ns_ins_del_lines);
  curs_x = left ? FIRST_CHAR_POSITION(f) : 0;
  curs_y = vpos;

  if (curs_y >= flexlines)
    return;

  ns_chars_to_rect(f, curs_x, curs_y, f->width + (left ? 1 : -1),
                   flexlines - 1, &srcRect);
#ifdef MACOSX
  srcRect.size.width += 2;
#else
  srcRect.size.width++;
#endif
  destOrigin = srcRect.origin;
  blkRect = srcRect;
  
  if (n > 0)
    {
      srcRect.size.height -= height;
      destOrigin.y        += height;
    }
  else if (n < 0)
    {
      height = -height;

      blkRect.origin.y     = srcRect.origin.y + srcRect.size.height - height;
      srcRect.size.height -= height;
      srcRect.origin.y    += height;
    }
  blkRect.size.height  = height;

  ns_dumpcursor(9, f,-1,-1);

  modRect = NSUnionRect(NSMakeRect(destOrigin.x, destOrigin.y,
                                   srcRect.size.width, srcRect.size.height),
                        blkRect);
  ns_focus(f, &modRect);

  if (srcRect.size.height > 0)
    NSCopyBits(NSNullObject, srcRect , destOrigin);

  if (blkRect.size.height > 0)
    {
      [NS_FACE_BACKGROUND(FRAME_DEFAULT_FACE(f)) set];
      //PENDING: need to remove scrollbar frames
      NSRectFill(blkRect);
    }

  ns_unfocus(f);
}

static void ns_change_line_highlight(int new_highlight, int vpos,
                                     int first_unused_hpos)
{
  TRACE(ns_change_line_highlight);
  highlight = new_highlight;
  curs_x = 0;
  curs_y = vpos;
  ns_clear_end_of_line (updating_frame->width);
}

static void ns_insert_glyphs(GLYPH *start, int len)
{
  TRACE(ns_insert_glyphs);
  abort();
}

static void ns_write_glyphs(GLYPH *start, int len)
{
  struct frame *f = (updating_frame ? updating_frame : selected_frame);

  TRACE(ns_write_glyphs);

  if (f != updating_frame)
    {
      curs_x = f->cursor_x;
      curs_y = f->cursor_y;
      ns_dumpcursor(11, f,f->cursor_x+len,f->cursor_y);
      ns_dumpglyphs(f, curs_x, curs_y, start, len,
                    highlight ? NS_DUMPGLYPH_HIGHLIGHT : NS_DUMPGLYPH_NORMAL);
    }
  else
    {
      if (curs_y == f->phys_cursor_y && curs_x <= f->phys_cursor_x
          && curs_x + len > f->phys_cursor_x)
        ns_dumpcursor(12, f,-1,-1);
      ns_dumpglyphs(f, curs_x, curs_y, start, len,
                    highlight ? NS_DUMPGLYPH_HIGHLIGHT : NS_DUMPGLYPH_NORMAL);
      curs_x += len;
    }

  ns_unfocus(f);
}

static void ns_delete_glyphs(int len)
{
  TRACE(ns_delete_glyphs);
  abort();
}

static void ns_ring_bell(void)
{
  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
  TRACE(ns_ring_bell);
  if (visible_bell)
    {
      struct frame *frame = 
        (updating_frame ? updating_frame : selected_frame);
      NSView *view;

      ns_focus(frame, NULL);
      view = FRAME_NS_VIEW(frame);
      if (view != nil)
        {
          NSRect r;
          NSPoint dim;

          BLOCK_INPUT;
          r = [view bounds];
          dim.x = 128;
          dim.y = 128;
          r.origin.x += (r.size.width - dim.x) / 2;
          r.origin.y += (r.size.height - dim.y) / 2;
          r.size.width = dim.x;
          r.size.height = dim.y;
          [[view window] cacheImageInRect: NSInsetRect(r, -2, -2)];
          [NS_FACE_FOREGROUND(FRAME_DEFAULT_FACE(frame)) set];
          NSRectFill(r);
          [[view window] flushWindow];
          ns_timeout(150000);
          [[view window] restoreCachedImage];
          [[view window] flushWindow];
          UNBLOCK_INPUT;
        }
      ns_unfocus(frame);
    }
  else
    {
      NSBeep();
    }
  [pool release];
}

static void ns_reset_terminal_modes(void)
{
  TRACE(ns_reset_terminal_modes);
}

static void ns_set_terminal_modes(void)
{
  TRACE(ns_set_terminal_modes);
}

static void ns_update_begin(struct frame *f)
{
  NSView *view = FRAME_NS_VIEW(f);
  [view lockFocus];

  TRACE(ns_update_begin);

#ifdef GNUSTEP
  uRect = NSMakeRect(0,0,0,0);
#endif

  flexlines = f->height;
  highlight = 0;
  if (f == mouse_face_mouse_frame)
    {
      /* Don't do highlighting for mouse motion during the update.  */
      mouse_face_defer = 1;

        /* If the frame needs to be redrawn,
           simply forget about any prior mouse highlighting.  */
      if (FRAME_GARBAGED_P (f))
        mouse_face_window = Qnil;

      if (!NILP (mouse_face_window))
        {
          int firstline, lastline, i;
          struct window *w = XWINDOW (mouse_face_window);

          /* Find the first, and the last+1, lines affected by redisplay.  */
          for (firstline = 0; firstline < f->height; firstline++)
            if (FRAME_DESIRED_GLYPHS (f)->enable[firstline])
              break;

          lastline = f->height;
          for (i = f->height - 1; i >= 0; i--)
            {
              if (FRAME_DESIRED_GLYPHS (f)->enable[i])
                break;
              else
                lastline = i;
            }

          /* Can we tell that this update does not affect the window
             where the mouse highlight is?  If so, no need to turn off.
             Likewise, don't do anything if the frame is garbaged;
             in that case, the FRAME_CURRENT_GLYPHS that we would use
             are all wrong, and we will redisplay that line anyway.  */
          if (! (firstline > (XFASTINT (w->top) + window_internal_height (w))
                 || lastline < XFASTINT (w->top)))
            /* Otherwise turn off the mouse highlight now.  */
            clear_mouse_face ();
        }
    }
}

static void ns_update_end(struct frame *f)
{
  NSView *view = FRAME_NS_VIEW(f);
  TRACE(ns_update_end);
  ns_dumpcursor(13, f,curs_x,curs_y);

#ifdef GNUSTEP
  // trigger flush only in the rectangle we tracked as being drawn
  [view unlockFocusNeedsFlush: NO];
  [view lockFocusInRect: uRect];
//fprintf(stderr," (%.0f, %.0f : %.0f x %.0f)", uRect.origin.x, uRect.origin.y, uRect.size.width, uRect.size.height);
#endif
  [view unlockFocus];
  [[view window] flushWindow];

  if (f == mouse_face_mouse_frame)
    mouse_face_defer = 0;
}

static void ns_frame_up_to_date(struct frame *f)
{
  TRACE(ns_frame_up_to_date);
// This is run to fix up mouse highlighting right after a full update
// (ns_update_begin/end).  Some highlighting was deferred if GC was happening
// during note_mouse_highlight, while other highlighting was deferred for
// the full update.

/// temporarily disable: performance and stability issues
#if 0
  if (mouse_face_deferred_gc || (f == mouse_face_mouse_frame))
    {
      NSRect r;
      NSView *view = FRAME_NS_VIEW(mouse_face_mouse_frame);
      NSPoint position;

      mouse_face_deferred_gc = 0;

      position = [[view window] mouseLocationOutsideOfEventStream];
      position = [view convertPoint:position fromView:nil];
      r = [view bounds];
      if ([view mouse:position inRect:r] && !f->iconified &&
          !f->async_iconified)
        {
          note_mouse_highlight (mouse_face_mouse_frame,
                                mouse_face_mouse_x,
                                mouse_face_mouse_y);
          [(EmacsView *)view setMouseTracking: mouse_face_mouse_rect];
        }
      else
        {
          clear_mouse_face ();
          [(EmacsView *)view removeMouseTracking];
        }
    }
#endif
}

static void ns_set_terminal_window(int n)
{
  struct frame *f=(updating_frame ? updating_frame : selected_frame);
  TRACE(ns_set_terminal_window);
  flexlines = ((n > 0) && (n <= f->height)) ? n : f->height;
}

void ns_send_appdefined (int value)
{
  //TRACE(ns_send_appdefined);

  /* Only post this event if we haven't already posted one.  This will end
       the [NXApp run] main loop after having processed all events queued at
       this moment.  */
  if (send_appdefined)
    {
      NSEvent *nxev;

      /* We only need one NX_APPDEFINED event to stop NXApp from running.  */
      send_appdefined = NO;

      /* Don't need wakeup timer any more */
      if (timed_entry)
        {
          [timed_entry invalidate];
          [timed_entry release];
          timed_entry = 0;
        }

      /* Ditto for file descriptor poller */
      if (fd_entry)
        {
          [fd_entry invalidate];
          [fd_entry release];
          fd_entry = 0;
        }

      nxev = [NSEvent otherEventWithType: NSApplicationDefined
                                location: NSMakePoint(0, 0)
                           modifierFlags: 0
                               timestamp: 0
                            windowNumber: [[NSApp mainWindow] windowNumber]
                                 context: [NSApp context]
                                 subtype: 0
                                   data1: value
                                   data2: 0];
      
      /* Post an application defined event on the event queue.  When this is
         received the [NXApp run] will return, thus having processed all
         events which are currently queued.  */
      [NSApp postEvent: nxev atStart: NO];
    }
}

/* Post an event to ourself and keep reading events until we read it back
   again.  In effect process all events which were waiting.  */
static int ns_read_socket (int sd, struct input_event *bufp,
                           int numchars, int expected)
{
  int nevents;
  //TRACE(ns_read_socket);

  // we manage autorelease pools by allocate/reallocate each time around
  // the loop; strict nesting is occasionally violated but seems not to
  // matter.. earlier methods using full nesting caused major memory leaks
  [outerpool release];
  outerpool = [[NSAutoreleasePool alloc] init];

    /* We must always send one NX_APPDEFINED event to ourself, otherwise
       [NXApp run] will never exit.  */
  send_appdefined = YES;

  if (!expected)
    {
      /* Post an application defined event on the event queue.  When this is
         received the [NXApp run] will return, thus having processed all
         events which are currently queued, if any.  */
      ns_send_appdefined (-1);
    }

  events = bufp;
  eventsleft = numchars;

  [NSApp run];

  nevents = numchars - eventsleft;
  eventsleft = 0;
  events = 0;
  return nevents;
}

static void ns_reassert_line_highlight(int new, int vpos)
{
  TRACE(ns_reassert_line_highlight);
  highlight = new;
}

static void ns_mouse_position (struct frame **fp, int insist,
                               Lisp_Object *bar_window,
                               enum scroll_bar_part *part,
                               Lisp_Object *x, Lisp_Object *y,
                               unsigned long *time)
{
  id view;
  NSPoint position;
  int xchar, ychar;
  Lisp_Object frame, tail;

  TRACE(ns_mouse_position);

    /* Clear the mouse-moved flag for every frame on this display.  */
  FOR_EACH_FRAME (tail, frame)
    if (FRAME_NS_DISPLAY (XFRAME (frame)) == FRAME_NS_DISPLAY (*fp))
      XFRAME (frame)->mouse_moved = 0;
   
  if (last_mouse_frame)
    *fp=last_mouse_frame;
  view = FRAME_NS_VIEW(*fp);
   
  position = [[view window] mouseLocationOutsideOfEventStream];
  position = [view convertPoint:position fromView:nil];
  pixel_to_glyph_coords(*fp,
                        (int)rint(position.x), (int)rint(position.y),
                        &xchar, &ychar, 0, 1);
  ns_chars_to_rect(*fp, xchar, ychar, xchar, ychar, &last_mouse_glyph);
  if (bar_window)
    *bar_window = Qnil;
  if (part)
    *part = 0;
  if (x)
    XSET (*x, Lisp_Int, (int)rint(position.x));
  if (y)
    XSET (*y, Lisp_Int, (int)rint(position.y));
  if (time)
    *time = last_mouse_movement_time;
}

static void ns_frame_rehighlight(struct frame *frame)
{
  struct frame *f=0;

  TRACE(ns_frame_rehighlight);
  if (ns_focus_frame)
    {
      f = (GC_FRAMEP (FRAME_FOCUS_FRAME (ns_focus_frame))
           ? XFRAME (FRAME_FOCUS_FRAME (ns_focus_frame))
           : ns_focus_frame);
      if (!FRAME_LIVE_P (f))
        {
          FRAME_FOCUS_FRAME (ns_focus_frame) = Qnil;
          f = ns_focus_frame;
        }
    }

  if (f != ns_highlight_frame)
    {
      if (f)
        [[FRAME_NS_VIEW(f) window] makeKeyAndOrderFront:NSApp];
    }
}

static void ns_frame_raise_lower(struct frame *f, int raise)
{
  TRACE(ns_frame_raise_lower);
  if (raise)
    ns_raise_frame(f);
  else
    ns_lower_frame(f);
}

// update or add scrollbar
static void ns_set_vertical_scroll_bar(struct window *window,
                                       int portion, int whole, int position)
{
  Lisp_Object win;
  NSRect r, v;
  struct frame *f= XFRAME (WINDOW_FRAME (window));
  EmacsView *view = FRAME_NS_VIEW(f);
  int top = XINT (window->top);
  int left = WINDOW_VERTICAL_SCROLL_BAR_COLUMN (window);
  int height = WINDOW_VERTICAL_SCROLL_BAR_HEIGHT (window);
  EmacsScroller *bar;

  // optimization; display engine sends WAY too many of these..
  if (!NILP (window->vertical_scroll_bar))
    {
      bar = XNS_SCROLL_BAR(window->vertical_scroll_bar);
      if ([bar checkSamePosition:position portion:portion whole:whole])
        {
          if (view->scrollbarsNeedingUpdate == 0)
            return;
          else
            view->scrollbarsNeedingUpdate--;
        }
    }

  TRACE(ns_set_vertical_scroll_bar);

  XSETWINDOW (win, window);
  ns_chars_to_rect (f, left, top,
                    left + FRAME_SCROLL_BAR_COLS (f) - 1, top + height - 1, &r);
  if (FRAME_SCROLL_BAR_PIXEL_WIDTH (f) > 0)
    r.size.width = FRAME_SCROLL_BAR_PIXEL_WIDTH (f);

    // the parent view is flipped, so we need to flip y value
  v = [view frame];
  r.origin.y = (v.size.height + f->output_data.ns->internal_border_width -
                r.size.height - r.origin.y);
  // minor adjustment..
  if (FRAME_HAS_VERTICAL_SCROLL_BARS_ON_RIGHT (f))
    r.origin.x += 2*f->output_data.ns->internal_border_width;

  if (v.size.height - (r.origin.y+r.size.height) < 5) // top scrollbar
    r.size.height += f->output_data.ns->internal_border_width;

    // we want at least 3 lines to display a scrollbar
  if (window->height < 3)
    {
      if (!NILP (window->vertical_scroll_bar))
        {
          bar = XNS_SCROLL_BAR(window->vertical_scroll_bar);
          [bar removeFromSuperview];
          window->vertical_scroll_bar = Qnil;
        }
      return;
    }

  if (NILP (window->vertical_scroll_bar))
    {
      bar=[[EmacsScroller alloc] initFrame:r window:win];
      VOID_TO_LISP (window->vertical_scroll_bar, bar);
    }
  else
    {
      bar= XNS_SCROLL_BAR(window->vertical_scroll_bar);
      [bar setFrame:r];
    }

  [bar setPosition:position portion:portion whole:whole];
}

/// NOTE: most of the following comments are copied from xterm/term.c

/* The following three hooks are used when we're doing a thorough
   redisplay of the frame.  We don't explicitly know which scroll bars
   are going to be deleted, because keeping track of when windows go
   away is a real pain - "Can you say set-window-configuration, boys
   and girls?"  Instead, we just assert at the beginning of redisplay
   that *all* scroll bars are to be removed, and then save a scroll bar
   from the fiery pit when we actually redisplay its window.  */

/* Arrange for all scroll bars on FRAME to be removed at the next call
   to `*judge_scroll_bars_hook'.  A scroll bar may be spared if
   `*redeem_scroll_bar_hook' is applied to its window before the judgement. */
static void ns_condemn_scroll_bars(struct frame *f)
{
  int i;
  id view;
  NSArray *subviews = [[FRAME_NS_VIEW(f) superview] subviews];
  TRACE(ns_condemn_scroll_bars);
  for (i=[subviews count]-1; i>=0; i--)
    {
      view=[subviews objectAtIndex:i];
      if (![view isKindOfClass:[EmacsScroller class]])
        continue;
      [view condemn];
    }
}

/* Unmark WINDOW's scroll bar for deletion in this judgement cycle.
   Note that WINDOW isn't necessarily condemned at all.  */
static void ns_redeem_scroll_bar(struct window *window)
{
  id bar;
  TRACE(ns_redeem_scroll_bar);
  if (!NILP(window->vertical_scroll_bar))
    {
      bar=XNS_SCROLL_BAR (window->vertical_scroll_bar);
      [bar reprieve];
    }
}

/* Remove all scroll bars on FRAME that haven't been saved since the
   last call to `*condemn_scroll_bars_hook'.  */
static void ns_judge_scroll_bars(struct frame *f)
{
  int i;
  id view;
  NSArray *subviews = [[FRAME_NS_VIEW(f) superview] subviews];
  TRACE(ns_judge_scroll_bars);
  for (i=[subviews count]-1; i>=0; i--)
    {
      view=[subviews objectAtIndex:i];
      if (![view isKindOfClass:[EmacsScroller class]]) continue;
      [view judge];
    }
}

static Lisp_Object append2(Lisp_Object list, Lisp_Object item)
{
  Lisp_Object array[2];
  array[0]=list;
  array[1]=Fcons(item,Qnil);
  return Fnconc(2,&array[0]);
}

static void ns_set_default_prefs()
{
  ns_alternate_is_meta = Qt;
  ns_cursor_blink_rate = Qnil;
  ns_shrink_space = make_float(1.0);
  ns_antialias_text = YES;
  ns_antialias_threshold = 10.0;
  ns_use_qd_smoothing = NO;
  ns_use_system_highlight_color = YES;
}

static void ns_default(const char *parameter,
                       Lisp_Object *result,
                       Lisp_Object yesval,
                       Lisp_Object noval,
                       Lisp_Object leftval,
                       Lisp_Object rightval,
                       int floatval)
// ----------------------------------------------------------------------------
//    Check parameter values in user's preferences
// ----------------------------------------------------------------------------
{
  const char *value;

  if ( (value=[[[NSUserDefaults standardUserDefaults]
                   stringForKey:[NSString stringWithUTF8String: parameter]]
                UTF8String]) )
    {
      double f;
      if (strcasecmp(value,"YES") == 0)
        *result = yesval;
      else if (strcasecmp(value,"NO") == 0)
        *result = noval;
      else if (leftval && strcasecmp(value,"LEFT") == 0)
        *result = leftval;
      else if (rightval && strcasecmp(value,"RIGHT") == 0)
        *result = noval;
      else if (floatval && (f = atof(value)) > 0.0)
        *result = make_float(f);
      else fprintf(stderr,
                   "Bad value for default \"%s\": \"%s\"\n", parameter,value);
    }
}


/* Set up use of NS before we make the first connection.  */
void ns_initialize ()
{
  TRACE(ns_initialize);
  clear_frame_hook = ns_clear_frame;
  clear_end_of_line_hook = ns_clear_end_of_line;
  ins_del_lines_hook = ns_ins_del_lines;
  change_line_highlight_hook = ns_change_line_highlight;
  insert_glyphs_hook = ns_insert_glyphs;
  write_glyphs_hook = ns_write_glyphs;
  delete_glyphs_hook = ns_delete_glyphs;
  update_begin_hook = ns_update_begin;
  update_end_hook = ns_update_end;
  ring_bell_hook = ns_ring_bell;
  reset_terminal_modes_hook = ns_reset_terminal_modes;
  set_terminal_modes_hook = ns_set_terminal_modes;
  set_terminal_window_hook = ns_set_terminal_window;
  read_socket_hook = ns_read_socket;
  frame_up_to_date_hook = ns_frame_up_to_date;
  cursor_to_hook = ns_cursor_to;
  reassert_line_highlight_hook = ns_reassert_line_highlight;
  mouse_position_hook = ns_mouse_position;
  frame_rehighlight_hook = ns_frame_rehighlight;
  frame_raise_lower_hook = ns_frame_raise_lower;
  set_vertical_scroll_bar_hook = ns_set_vertical_scroll_bar;
  condemn_scroll_bars_hook = ns_condemn_scroll_bars;
  redeem_scroll_bar_hook = ns_redeem_scroll_bar;
  judge_scroll_bars_hook = ns_judge_scroll_bars;

  scroll_region_ok = 1;
  char_ins_del_ok = 0;
  line_ins_del_ok = 1;
  fast_clear_end_of_line = 1;
  memory_below_frame = 0;
  baud_rate = 38400;

    /* No interrupt input under NS */
  Fset_input_mode (Qnil, Qnil, Qt, Qnil);
}

/* Start the Application and get things rolling.  */
struct ns_display_info *
ns_term_init (Lisp_Object display_name)
{
  extern Lisp_Object Fset_input_mode(Lisp_Object,Lisp_Object,
                                     Lisp_Object,Lisp_Object);
  struct ns_display_info *dpyinfo;
  static int ns_initialized = 0;
  Lisp_Object tmp;

  TRACE(ns_term_init);
  // count object allocs (About, click icon); on OS X use ObjectAlloc tool
  //GSDebugAllocationActive(YES);

  if (!ns_initialized)
    {
      ns_initialize ();
      ns_initialized = 1;
    }

  dpyinfo = (struct ns_display_info *)
    xmalloc (sizeof (struct ns_display_info));
  bzero (dpyinfo, sizeof (struct ns_display_info));

#ifdef MULTI_KBOARD
  dpyinfo->kboard = all_kboards;
#endif

  dpyinfo->next = ns_display_list;
  ns_display_list = dpyinfo;

  /* Put it on ns_display_name_list */
  ns_display_name_list = Fcons (Fcons (display_name, Qnil),
                                ns_display_name_list);
  dpyinfo->name_list_element = XCONS (ns_display_name_list)->car;

    // Read various user defaults.
  ns_set_default_prefs();
  ns_default("AlternateIsMeta", &ns_alternate_is_meta,
             Qt, Qnil, Qt, Qt, 0);
  ns_default("CursorBlinkRate", &ns_cursor_blink_rate,
             make_float(0.5), Qnil, 0, 0, 1);
  ns_default("ShrinkSpace", &ns_shrink_space,
             make_float(0.75), make_float(1.0), 0, 0, 1);
  ns_default("GSFontAntiAlias", &ns_antialias_text,
             YES, NO, 0, 0, 0);
  tmp = Qnil;
  ns_default("AppleAntiAliasingThreshold", &tmp,
             make_float(10.0), make_float(6.0), 0, 0, 1);
  ns_antialias_threshold = NILP(tmp) ? 10.0 : QXFLOAT(tmp);
  ns_default("UseQuickdrawSmoothing", &ns_use_qd_smoothing,
             YES, NO, 0, 0, 0);

  ns_default("UseSystemHighlightColor", &ns_use_system_highlight_color,
             YES, NO, 0, 0, 0);
  if (ns_use_system_highlight_color == YES)
    {
      ns_selection_color = [[NSUserDefaults standardUserDefaults]
                               stringForKey: @"AppleHighlightColor"];
      if (ns_selection_color == nil)
        ns_selection_color = @"grey";
    }
  else
    ns_selection_color = @"grey";

  {
    id cl;
    Lisp_Object tem, tem1;
    extern Lisp_Object Vsource_directory;

    cl = [NSColorList colorListNamed:@"Emacs"];

    if ( cl == nil )
      {
        // first try data_dir, then invocation-dir
        // and finally source-directory/etc
        tem1 = tem =
          Fexpand_file_name (build_string("Emacs.clr"), Vdata_directory);
        if (NILP (Ffile_exists_p (tem)))
          {
            tem = Fexpand_file_name (build_string ("Emacs.clr"),
                                     Vinvocation_directory);
            if (NILP (Ffile_exists_p (tem)))
              {
                Lisp_Object newdir =
                  Fexpand_file_name (build_string ("etc/"),
                                     Vsource_directory);
                tem = Fexpand_file_name (build_string ("Emacs.clr"),
                                         newdir);
              }
          }

        cl = [[NSColorList alloc]
               initWithName: @"Emacs"
                   fromFile: [NSString stringWithCString:XSTRING (tem)->data]];
        if (cl==nil)
          fatal ("Could not find %s.\n", XSTRING (tem1)->data);
        [cl writeToFile: nil];
      }


  }

  {
    char c[1024];
#ifndef GNUSTEP
    PSnextrelease(1024,c);
#else
    strcpy(c, gnustep_base_version);
#endif
    Vwindow_system_version = build_string(c);
  }

  mouse_face_mouse_frame = 0;
  mouse_face_deferred_gc = 0;
  mouse_face_beg_row = mouse_face_beg_col = -1;
  mouse_face_end_row = mouse_face_end_col = -1;
  mouse_face_face_id = 0;
  mouse_face_window = Qnil;
  mouse_face_mouse_x = mouse_face_mouse_y = 0;
  mouse_face_defer = 0;
  mouse_face_mouse_row = -1;
  mouse_face_tracked_view = nil;

  delete_keyboard_wait_descriptor (0);

    // Start app and create the main menu, window, view.
    // The view will then ask the NSApp to stop and return to Emacs
  [EmacsApp sharedApplication];
  if (NSApp == 0)
    return 0;
  [NSApp setDelegate: NSApp];
    
#ifdef MACOSX
  {
    NSMenu *appMenu;
    id<NSMenuItem> item;
    // set up the application menu
    svcsMenu = [[EmacsMenu alloc] initWithTitle: @"Services"];
    [svcsMenu setAutoenablesItems: NO];
    appMenu = [[EmacsMenu alloc] initWithTitle: @"Emacs"];
    [appMenu setAutoenablesItems: NO];
    mainMenu = [[EmacsMenu alloc] initWithTitle: @""];

    [appMenu insertItemWithTitle: @"About Emacs"
                          action: @selector(orderFrontStandardAboutPanel:) 
                   keyEquivalent: @""
                         atIndex: 0];
    [appMenu insertItem: [NSMenuItem separatorItem] atIndex: 1];
    [appMenu insertItemWithTitle: @"Preferences"
                          action: @selector(showPreferencesWindow:) 
                   keyEquivalent: @","
                         atIndex: 2];
    [appMenu insertItem: [NSMenuItem separatorItem] atIndex: 3];
    item = [appMenu insertItemWithTitle: @"Services"
                                 action: @selector(menuDown:)
                          keyEquivalent: @""
                                atIndex: 4];
    [appMenu setSubmenu: svcsMenu forItem: item];
    [svcsMenu setSupercell: item];
    [appMenu insertItem: [NSMenuItem separatorItem] atIndex: 5];
    [appMenu insertItemWithTitle: @"Hide Emacs"
                          action: @selector(hide:) 
                   keyEquivalent: @"h" 
                         atIndex: 6];
    [appMenu insertItemWithTitle: @"Hide Others"
                          action: @selector(hideOtherApplications:) 
                   keyEquivalent: @"H" 
                         atIndex: 7];
    [appMenu insertItem: [NSMenuItem separatorItem] atIndex: 8];
    [appMenu insertItemWithTitle: @"Quit Emacs"
                          action: @selector(terminate:) 
                   keyEquivalent: @"q" 
                         atIndex: 9];

    item = [mainMenu insertItemWithTitle: @"Emacs"
                                  action: @selector(menuDown:)
                           keyEquivalent: @""
                                 atIndex: 0];
    [mainMenu setSubmenu: appMenu forItem: item];

    [NSApp setMainMenu: mainMenu];
    [NSApp setAppleMenu: appMenu];
    [NSApp setServicesMenu: svcsMenu];
  }
#endif  // MACOSX menu setup
    
  [NSApp run];

  return dpyinfo;
}

extern Lisp_Object Vauto_save_list_file_name;
void ns_term_shutdown()
{
  // not reached in emacs.c after this is called by shut_down_emacs
  if (STRINGP (Vauto_save_list_file_name))
    unlink (XSTRING (Vauto_save_list_file_name)->data);
  ns_shutdown_properly = YES;
  [NSApp terminate: NSApp];
}

char *x_get_keysym_name (int keysym)
{
  /* Make static so we can always return it */
  static char value[100];
  TRACE(x_get_keysym_name);
  sprintf(value, "%c", keysym);
  return value;
}

void syms_of_nsterm ()
{
  TRACE(syms_of_nsterm);
  DEFVAR_LISP("ns-input-file", &ns_input_file,
              "The file specified in the last NS event.");
  ns_input_file=Qnil;

  DEFVAR_LISP("ns-input-ascii", &ns_input_ascii,
              "The data received in the last NS drag event..");
  ns_input_ascii=Qnil;

  DEFVAR_LISP("ns-input-font", &ns_input_font,
              "The font specified in the last NS event.");
  ns_input_font=Qnil;

  DEFVAR_LISP("ns-input-fontsize", &ns_input_fontsize,
              "The fontsize specified in the last NS event.");
  ns_input_fontsize=Qnil;

  DEFVAR_LISP ("ns-input-line", &ns_input_line,
               "The line specified in the last NS event.");
  ns_input_line=Qnil;

  DEFVAR_LISP ("ns-input-color", &ns_input_color,
               "The color specified in the last NS event.");
  ns_input_color=Qnil;

  DEFVAR_LISP ("ns-alternate-is-meta", &ns_alternate_is_meta,
               "This variable describes what the effect of the alternate key is under NS.\n\
nil means that the alternate key is not interpreted by Emacs at all,\n\
t means that the alternate key is used as meta key.\n\
(This variable should only be read, never set.)");

  DEFVAR_LISP ("ns-cursor-blink-rate", &ns_cursor_blink_rate,
               "Rate at which the Emacs cursor blinks (in seconds).\n\
Set to nil to disable blinking.");

  DEFVAR_LISP ("ns-shrink-space", &ns_shrink_space,
               "Amount by which spacing between lines is compressed.\n\
(This variable should only be read, never set.)");

  DEFVAR_BOOL ("ns-antialias-text", &ns_antialias_text,
               "*Non-nil (the default) means to render text antialiased. Only has an effect on OS X Panther and above.");

  DEFVAR_BOOL ("ns-use-qd-smoothing", &ns_use_qd_smoothing,
               "*Nil (the default) means to render text using QuickDraw (less heavy) antialiasing. Only has an effect on OS X Panther and above.");

  DEFVAR_BOOL ("ns-use-system-highlight-color",
               &ns_use_system_highlight_color,
               "*t (the default) means to use the system default (on OS X only) for the highlight color.  Nil means to use standard emacs 'grey'.");

  staticpro (&ns_display_name_list);
  ns_display_name_list = Qnil;
}

int ns_select (int nfds, fd_set *readfds, fd_set *writefds,
               fd_set *exceptfds, struct timeval *timeout)
// ----------------------------------------------------------------------------
//   Replacement for select, checking for events
// ----------------------------------------------------------------------------
{
  int result;
  double time;
  NSEvent *ev;
  //TRACE(ns_select);  // (swamps the other trace messages)

  // Save file descriptor set, which gets overwritten in calls to select()
  // Note, this is called from process.c, and only readfds is ever set
  if (readfds)
    {
      memcpy(&select_readfds,readfds,sizeof(fd_set));
      select_nfds = nfds;
    }
  else
    select_nfds = 0;

    // Try an initial select for pending data on input files
  select_timeout.tv_sec = select_timeout.tv_usec = 0;
  result = select(nfds, readfds, writefds, exceptfds, &select_timeout);
  if (result)
    return result;

  if (!timeout || timed_entry || fd_entry)
    fprintf(stderr, "assertion failed: timeout null or timed_entry/fd_entry non-null in ns_select\n");

    // set a timeout and run the main AppKit event loop while continuing
    // to monitor the files
  time = ((double) timeout->tv_sec) + ((double) timeout->tv_usec)/1000000.0;
  timed_entry = [[NSTimer scheduledTimerWithTimeInterval: time
                                                  target: NSApp
                                                selector: @selector(timeout_handler:)
                                                userInfo: 0
                                                 repeats: YES]  // ensure safe removal..
                                                                retain];

  // set a periodic task to try the select() again
  fd_entry = [[NSTimer scheduledTimerWithTimeInterval: 0.1
                                               target: NSApp
                                             selector: @selector(fd_handler:)
                                             userInfo: 0
                                              repeats: YES]
               retain];

  if (NUMBERP(ns_cursor_blink_rate) && !cursor_blink_entry)
    {
      cursor_blink_entry =
        [[NSTimer scheduledTimerWithTimeInterval:
                    QXFLOAT(ns_cursor_blink_rate)
                                          target: NSApp
                                        selector: @selector(cursor_blink_handler:)
                                        userInfo:0
                                         repeats:YES]
          retain];
    }
  else if (!NUMBERP(ns_cursor_blink_rate) && cursor_blink_entry)
    {
      [cursor_blink_entry invalidate];
      [cursor_blink_entry release];
      cursor_blink_entry = 0;
      if (ns_highlight_frame)
        {
          Lisp_Object tem=get_frame_param(ns_highlight_frame, Qcursor_type);
          ns_highlight_frame->output_data.ns->desired_cursor =
            ns_lisp_to_cursor_type(tem);
        }
    }

  /* Let Application dispatch events until it receives an event of the type
       NX_APPDEFINED, which should only be sent by timeout_handler.  */
  gobble_input (timeout ? 1 : 0);
  ev = last_appdefined_event;

  if (ev)
    {
      int t;
      if ([ev type] != NSApplicationDefined)
        abort ();
      
      t = [ev data1];
      last_appdefined_event = 0;

      if (t == -2)
        {
          /* The NX_APPDEFINED event we received was a timeout. */
          return 0;
        }
      else if (t == -1)
        {
          /* The NX_APPDEFINED event we received was the result of
             at least one real input event arriving.  */
          errno = EINTR;
          return -1;
        }
      else
        {
          // Received back from select() in fd_handler; copy the results
          if (readfds) memcpy(readfds, &select_readfds, sizeof(fd_set));
          return t;
        }
    }
  // never reached, shut compiler up
  return 0;
}



// ****************************************************************************
//
//    EmacsApp implementation
//
// ****************************************************************************

@implementation EmacsApp

- (void)sendEvent:(NSEvent *)theEvent
// ----------------------------------------------------------------------------
//   Events posted by ns_send_appdefined interrupt the run loop here
// ----------------------------------------------------------------------------
{
  //TRACE(sendEvent);
  if ([theEvent type] == NSApplicationDefined)
    {
      last_appdefined_event = theEvent;
      [self stop:self];
    }
  [super sendEvent:theEvent];
}

- (void)showPreferencesWindow:(id)sender
{
  if (prefsController == nil)
    prefsController = [[EmacsPrefsController alloc] init];
  [prefsController showForFrame: selected_frame];
}

// ****************************************************************************
//
//    EmacsApp delegate implementation
//
// ****************************************************************************

- (void)applicationDidFinishLaunching:(NSNotification *)notification
// ----------------------------------------------------------------------------//   When application is loaded, terminate event loop in ns_term_init
// ----------------------------------------------------------------------------
{
  TRACE(applicationDidFinishLaunching);
  ns_send_appdefined (-2);
}

- (void) terminate: (id)sender
{
  if (ns_shutdown_properly)
    [super terminate: sender];
  else {
      //PENDING: this still fails if you ask to not quit anyway
      Feval(Fcons(intern("save-buffers-kill-emacs"),Qnil));
      ns_send_appdefined(-1);
  }
}

// Open a file (used by below)
-(BOOL) openFile: (char *)fileName
        fromLine: (int)startLine
          toLine: (int)endLine
{
  struct frame *emacsframe = selected_frame;
  NSEvent *theEvent = [NSApp currentEvent];

  if (eventsleft <= 0)
    return NO;

  events->kind=non_ascii_keystroke;
  events->code=KEY_NS_OPEN_FILE_LINE;
  ns_input_file=append2(ns_input_file, build_string(fileName));
  if (startLine==endLine)
    ns_input_line= (startLine >= 0) ? make_number(startLine) : Qnil;
  else
    ns_input_line= Fcons(make_number(startLine), make_number(endLine));
  events->modifiers=0;
  EV_TRAILER(theEvent);

  return YES;
}

//   Notification from the Workspace to open a file
- (BOOL)application: sender openFile: (NSString *)fileName
{
  return [self openFile: (char *) [fileName UTF8String] fromLine:-1 toLine:-1];
}

//   Open a file as a temporary file
- (BOOL)application:sender openTempFile:(NSString *)file
{
  return [self application: sender openFile: file];
}

//   Notification from the Workspace to open a file noninteractively (?)
- (BOOL)application: sender openFileWithoutUI: (NSString *)fileName
{
  return [self openFile: (char *) [fileName UTF8String] fromLine:-1 toLine:-1];
}

//   Notification from the Workspace to open multiple files
- (void)application: sender openFiles: (NSArray *)fileList
{
  NSEnumerator *filenames = [fileList objectEnumerator];
  NSString *fileName;
  while ((fileName = [filenames nextObject]) != nil)
    if ([self openFile: (char *)[fileName UTF8String] fromLine:-1 toLine:-1]
        == NO)
      return;
}


// ****************************************************************************
//
//    EmacsApp aux handlers for managing event loop
//
// ****************************************************************************

- (void)timeout_handler: (NSTimer *)timedEntry
// ----------------------------------------------------------------------------
//   When a timer expires, exit the application loop
// ----------------------------------------------------------------------------
{
  //TRACE(timeout_handler);
  /* The timeout specified to ns_select has passed.  */
  ns_send_appdefined (-2);
}


- (void)cursor_blink_handler: (NSTimer *)timedEntry
// ----------------------------------------------------------------------------
//   Flash the cursor
// ----------------------------------------------------------------------------
{
  TRACE(cursor_blink_handler);
  if (!ns_highlight_frame)
    return;
  if (ns_highlight_frame->output_data.ns->current_cursor==no_highlight)
    {
      Lisp_Object tem=get_frame_param(ns_highlight_frame, Qcursor_type);
      ns_highlight_frame->output_data.ns->desired_cursor =
        ns_lisp_to_cursor_type(tem);
    }
  else
    {
      ns_highlight_frame->output_data.ns->desired_cursor = no_highlight;
    }

  ns_dumpcursor(14,
                ns_highlight_frame,
                ns_highlight_frame->cursor_x,
                ns_highlight_frame->cursor_y);
}


- (void)fd_handler: (NSTimer *) fdEntry
// ----------------------------------------------------------------------------
//   Check data waiting on file descriptors and terminate if so
// ----------------------------------------------------------------------------
{
  int result;
  // TRACE(fd_handler);

  if (select_nfds == 0)
    return;

  memcpy(&t_readfds, &select_readfds, sizeof(fd_set));

  select_timeout.tv_sec = select_timeout.tv_usec = 0;
  result = select(select_nfds, &t_readfds, (SELECT_TYPE *)0,(SELECT_TYPE *)0,
                  &select_timeout);
  if (result)
    {
      memcpy(&select_readfds, &t_readfds, sizeof(fd_set));
      ns_send_appdefined (result);
    }
}

@end


// ****************************************************************************
//
//   EmacsView implementation
//
// ****************************************************************************

@implementation EmacsView

// needed to inform when window closed from LISP
- (void) setWindowClosing: (BOOL)closing
{
  windowClosing = closing;
}

// called on font panel selection
- (void)changeFont:(id)sender
{
  NSEvent *e=[[self window] currentEvent];
  struct face *face=FRAME_DEFAULT_FACE(emacsframe);
  id newFont;
  float size;

  TRACE(changeFont);
  if (eventsleft<=0)
    return;

  if ( (newFont=[sender convertFont:face->font->nsfont]) )
    {
      events->kind=non_ascii_keystroke;
      events->modifiers=0;
      events->code=KEY_NS_CHANGE_FONT;

      size=[newFont pointSize];
      if (size==rint(size))
        ns_input_fontsize=make_number((int)rint(size));
      else
        ns_input_fontsize=make_float(size);
      ns_input_font=build_string([[newFont fontName] UTF8String]);
      EV_TRAILER(e);
    }
}

- (BOOL)acceptsFirstResponder
{
  TRACE(acceptsFirstResponder);
  return YES;
}

- (void)resetCursorRects
{
  NSRect visible = [self visibleRect];
  // adjust for scrollbar, which overlaps this
  visible.origin.x +=
    (FRAME_HAS_VERTICAL_SCROLL_BARS_ON_LEFT(emacsframe) ? 1 : -1)
    * NS_SCROLL_BAR_WIDTH_DEFAULT;
  TRACE(resetCursorRects);

  if (!NSIsEmptyRect(visible))
    [self addCursorRect:visible cursor:[NSCursor IBeamCursor]];
  [[NSCursor IBeamCursor] setOnMouseEntered: YES];
}

- (void)keyDown:(NSEvent *)theEvent
{
  int code;
  unsigned keysym;
  enum event_kind kind;
  int flags;
  int mapped_function_key = NO;
  static NSMutableArray *nsEvArray;
  static BOOL inDeadKey = NO;

  TRACE(keyDown);

  if (nsEvArray == nil)
    nsEvArray = [[NSMutableArray alloc] initWithCapacity: 2];

  // Rhapsody and OS X give up and down events for the arrow keys
  if ([theEvent type] != NSKeyDown)
    return;

  if (eventsleft <= 0)
    return;

  [NSCursor setHiddenUntilMouseMoves:YES];

  flags=[theEvent modifierFlags];

  if ([[theEvent characters] length] == 0
      && (!(flags & NSAlternateKeyMask) || NILP(ns_alternate_is_meta)))
    {
//  fprintf(stderr, "Received a dead key with code %d.\n",[theEvent keyCode]);
      //store dead key, and call [self interpretKeyEvents:(NSArray *)] on
      // next keyDown and wait for an insertText:(NSString *) to follow
      inDeadKey = YES;
      [nsEvArray addObject: theEvent];
      [self interpretKeyEvents: nsEvArray];
      return;
    }

  code=[[theEvent charactersIgnoringModifiers] characterAtIndex:0];
  kind=non_ascii_keystroke;

  events->modifiers=0;

  if (inDeadKey)
    {
//  fprintf(stderr, "Received a key code %d while processing dead key.\n",[theEvent keyCode]);
      [nsEvArray addObject: theEvent];
      [self interpretKeyEvents: nsEvArray];
      [nsEvArray removeAllObjects];
      inDeadKey = NO;
      return;
    }

  // PENDING: Mapping of NS-keys to X-compatible function key numbers
  // Note: an array would be better here, but not as easy to read
  // Ultimately, we want to have an 'ns' based table in keyboard.c
  keysym = ns_convert_key(code);
  if (keysym)
    {
      mapped_function_key = YES;
      code = keysym;
    }
  if (flags & NSHelpKeyMask)
    {
      events->modifiers |= hyper_modifier;
    }
  if (flags & NSCommandKeyMask)
    {
      events->modifiers |= super_modifier;
    }
  if (!NILP(ns_alternate_is_meta) && (flags & NSAlternateKeyMask))
    {
      events->modifiers |= meta_modifier;
      flags&=~NSAlternateKeyMask;
    }
  else if (NILP(ns_alternate_is_meta) && (flags & NSAlternateKeyMask))
    {
      code = [[theEvent characters] characterAtIndex: 0];
    }
  if ((flags & NSControlKeyMask) && (flags & NSShiftKeyMask))
    {
      events->modifiers |= ctrl_modifier;
      code |= 64;
      flags &= ~ (NSControlKeyMask | NSShiftKeyMask);
    }
  if (flags & NSControlKeyMask)
    {
      events->modifiers |= ctrl_modifier;
      flags&=~NSControlKeyMask;
    }
  if (flags & NSShiftKeyMask)
    {
      events->modifiers |= shift_modifier;
      flags&=~NSShiftKeyMask;
    }

//fprintf(stderr,"%d\t%d\n",flags,events->modifiers);

  if ((code<0x20)&&(([theEvent modifierFlags]&NSControlKeyMask) == 0))
    code |= (1<<28)|(3<<16);
  else if (code == 0x7f)
    code |= (1<<28)|(3<<16);
  else if (mapped_function_key == NO)
    kind=ascii_keystroke;

  events->kind=kind;
  events->code=code;
  EV_TRAILER (theEvent);
}

- (void)insertText:(NSString *)aString
{
  int code;

//NSLog(@"Received order to insert '%@'\t%d",aString,[aString characterAtIndex: 0]);
  if (eventsleft <= 0 || [aString length] != 1)
    return;
  code = [aString characterAtIndex: 0];
  if (code == 732)
    code = '~'; // 126
  events->modifiers = 0;
  events->kind = ascii_keystroke;
  events->code = code;
  XSETFRAME (events->frame_or_window, emacsframe);
  //events->timestamp = EV_TIMESTAMP (ns_last_event);
  events++;
  eventsleft--;
  ns_send_appdefined (-1);
}

/*****************************************************************************/
/* Begin dummy NSTextInput impl (only insertText is needed. */
- (void)doCommandBySelector:(SEL)aSelector
{
}
- (NSArray *)validAttributesForMarkedText
{
  static NSArray *arr = nil;
  if (arr = nil) arr = [NSArray new];
  return arr;
}
- (void)unmarkText
{
}
- (void)setMarkedText:(id)aString selectedRange:(NSRange)selRange
{
}
- (NSRange)selectedRange
{
  static NSRange rng;
  if (rng.length != 0) rng.location = rng.length = 0;
  return rng;
}
- (NSRange)markedRange
{
  static NSRange rng = { 0, 0 };
  return rng;
}
- (BOOL)hasMarkedText
{
  return NO;
}
- (NSRect)firstRectForCharacterRange:(NSRange)theRange
{
  static NSRect rect = { 0,0,0,0 };
  return rect;
}
- (long)conversationIdentifier
{
  return (long)self;
}
- (unsigned int)characterIndexForPoint:(NSPoint)thePoint
{
  return 0;
}
- (NSAttributedString *)attributedSubstringFromRange:(NSRange)theRange
{
  static NSAttributedString *str = nil;
  if (str == nil) str = [NSAttributedString new];
  return str;
}
/* End dummy NSTextInput impl (only insertText is needed. */
/*****************************************************************************/


/* This is what happens when the user presses the mouse button.  */
- (void)mouseDown:(NSEvent *)theEvent
{
  NSPoint p = [self convertPoint: [theEvent locationInWindow] fromView: nil];

  TRACE(mouseDown);
  if (eventsleft <= 0)
    return;

  last_mouse_frame=emacsframe;
  events->kind = mouse_click;
  events->code = EV_BUTTON(theEvent);
  XSET (events->x, Lisp_Int, (int)rint(p.x));
  XSET (events->y, Lisp_Int, (int)rint(p.y));
  events->modifiers = EV_MODIFIERS (theEvent) | EV_UDMODIFIERS (theEvent);
  EV_TRAILER (theEvent);
}

/* This is what happens when the user releases the mouse button.  */
- (void)mouseUp:(NSEvent *)theEvent
{
  TRACE(mouseUp);
  [self mouseDown:theEvent];
}

- (void)rightMouseDown:(NSEvent *)theEvent
{
  TRACE(rightMouseDown);
  [self mouseDown:theEvent];
}

- (void)rightMouseUp:(NSEvent *)theEvent
{
  TRACE(rightMouseUp);
  [self mouseDown:theEvent];
}

- (void) scrollWheel: (NSEvent *)theEvent
{
  TRACE(scrollWheel);
  [self mouseDown:theEvent];
}

/* Tell emacs the mouse has moved. */
- (void)mouseMoved:(NSEvent *)e
{
  NSPoint p = [self convertPoint: [e locationInWindow] fromView: nil];

  TRACE(mouseMoved);
  last_mouse_movement_time = EV_TIMESTAMP (e);

    // update any mouse face while dragging.
  note_mouse_movement (emacsframe, p.x, p.y);
  note_mouse_highlight (emacsframe, p.x, p.y);

  if (emacsframe->mouse_moved && send_appdefined)
    ns_send_appdefined (-1);
}

- (void)mouseDragged:(NSEvent *)e
{
  TRACE(mouseDragged);
  [self mouseMoved:e];
}

- (void)rightMouseDragged:(NSEvent *)e
{
  TRACE(rightMouseDragged);
  [self mouseMoved:e];
}

- (BOOL)windowShouldClose:(id)sender
{
  NSEvent *e=[[self window] currentEvent];

  TRACE(windowShouldClose);
  windowClosing = YES;
  if (ns_window_num<=1)
    return NO;
  if (eventsleft <= 0)
    return NO;
  events->kind=delete_window_event;
  events->modifiers=0;
  events->code=0;
  EV_TRAILER (e);
  /* Don't close this window, let this be done from lisp code.  */
  return NO;
}

- (NSSize)windowWillResize:(NSWindow *)sender toSize:(NSSize)frameSize
{
  int vbextra = NS_SCROLL_BAR_WIDTH(emacsframe);
  TRACE(windowWillResize);

//fprintf(stderr,"Window will resize: %f x %f\n",frameSize.width,frameSize.height);
  cols = rint((frameSize.width -
               emacsframe->output_data.ns->border_width -
               2 * emacsframe->output_data.ns->internal_border_width -
               vbextra) / NS_FONT_WIDTH(emacsframe));
  if (cols < MINWIDTH)
    cols=MINWIDTH;
  frameSize.width = (cols*NS_FONT_WIDTH(emacsframe) +
                     emacsframe->output_data.ns->border_width +
                     2*emacsframe->output_data.ns->internal_border_width +
                     vbextra);

  rows = rint((frameSize.height -
               emacsframe->output_data.ns->border_height -
               2*emacsframe->output_data.ns->internal_border_width) /
              emacsframe->output_data.ns->line_height);
  if (rows < MINHEIGHT)
    rows = MINHEIGHT;
//fprintf(stderr, "cols,rows = %d, %d\n",cols,rows);
  frameSize.height=(rows*emacsframe->output_data.ns->line_height +
                    emacsframe->output_data.ns->border_height +
                    2*emacsframe->output_data.ns->internal_border_width);

#ifdef MACOSX
  {
    // this sets window title to have size in it; the wm does this under GS
    NSRect r = [[self window] frame];
    if (r.size.height == frameSize.height && r.size.width == frameSize.width)
      {
        if (old_title != 0)
          {
            xfree(old_title);
            old_title = 0;
          }
      }
    else
      {
        char *size_title;
        if (old_title == 0)
          {
            const char *t = [[[self window] title] UTF8String];
            char *pos = strstr(t, "  —  ");
            if (pos)
              *pos = '\0';
            old_title = (char *) xmalloc(strlen(t) + 1);
            strcpy(old_title, t);
          }
        size_title = xmalloc(strlen(old_title) + 40);
        sprintf(size_title, "%s  —  (%d x %d)", old_title, cols, rows);
        [[self window] setTitle: [NSString stringWithUTF8String: size_title]];
        xfree(size_title);
      }
  }
#endif // MACOSX
//fprintf(stderr,"    ...size became %f x %f  (%d x %d)\n",frameSize.width,frameSize.height,cols,rows);

  return frameSize;
}

- (void)windowDidResize:(NSNotification *)notification
{
  NSWindow *theWindow = [notification object];

#ifdef GNUSTEP
    // in GNUstep, at least currently, it's possible to get a didResize
    // without getting a willResize.. therefore we need to act as if we got
    // the willResize now
//fprintf(stderr,"windowDidResize\n");
  NSSize fz = [theWindow frame].size;
  fz = [self windowWillResize: theWindow toSize: fz];
#endif // GNUSTEP

  TRACE(windowDidResize);

#ifdef MACOSX
  if (old_title != 0)
    {
      xfree(old_title);
      old_title = 0;
    }
#endif // MACOSX

#if defined(RHAPSODY) || defined(MACOSX) || defined(GNUSTEP)
  // The following is necessary to show the resize on Rhapsody / Aqua
  if (cols && rows)
    x_set_window_size(emacsframe, 0, cols, rows);
#endif
  ns_send_appdefined (-1);
}

- (void)windowDidBecomeKey:(NSNotification *)notification
{
  int val = ns_lisp_to_cursor_type(get_frame_param(emacsframe, Qcursor_type));
  TRACE(windowDidBecomeKey);
  ns_highlight_frame = emacsframe;
  if (val>=0)
    {
      FRAME_NEW_CURSOR(emacsframe) = val;
      ns_dumpcursor(16, emacsframe,
                    emacsframe->cursor_x,
                    emacsframe->cursor_y);
    }
  /// in xterm.c x_new_focus_frame (dpyinfo, frame), this is ifdef'd out
  // saying lisp code will get confused.. maybe the event loop is different
  // here, since this appears to be needed to get the focus shift to "take"
  selected_frame = emacsframe;
  XSETFRAME(XWINDOW(selected_frame->selected_window)->frame, selected_frame);
  Fselect_window(selected_frame->selected_window);
  choose_minibuf_frame();
  set_frame_menubar(emacsframe, 0, 1); // else ~1s delay..
  //ns_send_appdefined(-1);
}

- (void)windowDidResignKey:(NSNotification *)notification
{
  TRACE(windowDidResignKey);
  if (!windowClosing && [[self window] isVisible] == YES)
    {
      FRAME_NEW_CURSOR(emacsframe) = hollow_box;
      ns_dumpcursor(17, emacsframe,emacsframe->cursor_x,emacsframe->cursor_y);
    }
  if (ns_highlight_frame==emacsframe)
    ns_highlight_frame=0;
}

- (void)windowWillMiniaturize:sender
{
  TRACE(windowWillMiniaturize);
  [self setMiniwindowImage];
}

- (BOOL)isFlipped
{
  return YES;
}

- (BOOL)isOpaque
{
  return NO;
}

- initFrameFromEmacs:(struct frame *)f
{
  NSRect r, wr;
  Lisp_Object tem;
  NSWindow *win;
  int vbextra = NS_SCROLL_BAR_WIDTH(f);
  NSSize sz;
  NSColor *col;

  TRACE(initFrameFromEmacs);
  windowClosing = NO;
  scrollbarsNeedingUpdate = 0;
  windowTrackNum = 0;
  frameTrackNum = 0;
  faceTrackNum = 0;

  sbRects = (NSRectArray)xmalloc(50 * sizeof(NSRect));
  nsbRects = 0;

  r.origin.x   = f->output_data.ns->internal_border_width;
  r.origin.y   = f->output_data.ns->internal_border_width;
  r.size.width = rint(NS_FONT_WIDTH(f)*f->width) + vbextra;
  r.size.height= rint(f->output_data.ns->line_height*f->height);
  [self initWithFrame:r];

  FRAME_NS_VIEW(f) = self;
  emacsframe=f;
  old_title=0;

  r.origin.x=r.origin.y=0;
  r.size.width +=2*f->output_data.ns->internal_border_width;//+vbextra;
  r.size.height+=2*f->output_data.ns->internal_border_width;

  win = [[NSWindow alloc]
            initWithContentRect: r
                      styleMask: (NSResizableWindowMask |
                                  NSMiniaturizableWindowMask |
                                  NSClosableWindowMask)
                        backing: NSBackingStoreBuffered
                          defer: YES];

  wr = [win frame];
  f->output_data.ns->border_width=wr.size.width-r.size.width;
  f->output_data.ns->border_height=wr.size.height-r.size.height;

  [win setAcceptsMouseMovedEvents:YES];
  [win setDelegate:self];
  [win useOptimizedDrawing:YES];

  sz.width = NS_FONT_WIDTH(f);
  sz.height = FRAME_LINE_HEIGHT(f);
  [win setResizeIncrements: sz];

  [[win contentView] addSubview:self];

  if (ns_drag_types)
    {
      [self registerForDraggedTypes:ns_drag_types];
    }

  tem=f->name;
  if (!NILP(tem))
    [win setTitle:[NSString stringWithUTF8String:XSTRING(tem)->data]];

  tem=f->icon_name;
  if (!NILP(tem))
    [win setMiniwindowTitle:[NSString stringWithUTF8String:XSTRING(tem)->data]];

  {
    NSScreen *screen=[win screen];

    if (screen!=0)
      [win setFrameTopLeftPoint:NSMakePoint
           (IN_BOUND(-SCREENMAX,f->output_data.ns->left,SCREENMAX),
            IN_BOUND(-SCREENMAX,
                     [screen frame].size.height -
                     f->output_data.ns->top,SCREENMAX))];
  }

  [win makeFirstResponder:self];

  col = NS_FACE_BACKGROUND(FRAME_DEFAULT_FACE(emacsframe));
  [win setBackgroundColor: col];
  if ([col alphaComponent] != 1.0)
    [win setOpaque: NO];

  [self allocateGState];

  ns_adjust_size(emacsframe);
  ns_window_num++;
  return self;
}

- (void)windowDidMove:sender
{
  NSWindow *win=[self window];
  NSRect r;
  NSScreen *screen;

  TRACE(windowDidMove);
  if (!emacsframe->output_data.ns)
    return;
  r = [win frame];
  screen=[win screen];
  if (screen!=0)
    {
      emacsframe->output_data.ns->left=r.origin.x;
      emacsframe->output_data.ns->top=[screen frame].size.height-
        (r.origin.y+r.size.height);
    }
  /* Terminate the event loop.  */
  ns_send_appdefined (-1);
}

- (void)windowDidDeminiaturize:sender
{
  TRACE(windowDidDeminiaturize);
  if (!emacsframe->output_data.ns)
    return;
  emacsframe->async_visible   = 1;
  emacsframe->async_iconified = 0;
  windows_or_buffers_changed++;
  if ([[self window] windowNumber] > 0)
    [self adjustTrackingRects];
}

- (void)windowDidExpose:sender
{
  TRACE(windowDidExpose);
  if (!emacsframe->output_data.ns)
    return;
  emacsframe->async_visible = 1;
  SET_FRAME_GARBAGED (emacsframe);
  /* Terminate the event loop.  */
  if (send_appdefined) ns_send_appdefined (-1);
}

- (void)windowDidMiniaturize:sender
{
  TRACE(windowDidMiniaturize);
  if (!emacsframe->output_data.ns)
    return;
  if (FRAME_NS_VIEW(emacsframe) == mouse_face_tracked_view)
    [self removeMouseTracking];
  emacsframe->async_iconified = 1;
  /* Terminate the event loop.  */
  if (send_appdefined) ns_send_appdefined (-1);
}

- (void)adjustTrackingRects
{
  NSRect r;
  NSWindow *win = [self window];

  TRACE(EmacsView_adjustTrackingRects);

  //PENDING: need this?
  [self removeMouseTracking];

  r = [self bounds];
  [self convertRect:r toView:nil];

  if (windowTrackNum != 0)
    {
      [self removeTrackingRect: windowTrackNum];
      [self removeTrackingRect: frameTrackNum];
    }

  // Set inside to NO: if we are inside, we'll immediately get an
  // event and everything will be ok
  windowTrackNum = [self addTrackingRect:r
                                   owner:self
                                userData:nil
                            assumeInside:NO];

  r = [win frame];
  r.origin.x=r.origin.y=0;
  frameTrackNum = [self addTrackingRect:r
                                  owner:self
                               userData:nil
                           assumeInside:NO];
}

- (void) setMouseTracking:(NSRect)rect
{
  TRACE(EmacsView_setMouseTracking);
//fprintf(stderr, "setMouseTracking: %.0f, %.0f: %.0f x %.0f\n",rect.origin.x/7,rect.origin.y/15,rect.size.width/7,rect.size.height/15);

  if (faceTrackNum != 0)
    {
      TRACKREC (trackrec, self);
      [self removeTrackingRect: faceTrackNum];
    }
  [self convertRect:rect toView:nil];
  TRACKREC (rect, self);
  faceTrackNum = [self addTrackingRect:rect
                                 owner:self
                              userData:nil
                          assumeInside:YES];
}

- (void) removeMouseTracking
{
  TRACE(EmacsView_removeMouseTracking);

  if (faceTrackNum != 0)
    {
      TRACKREC (trackrec, self);
      [self removeTrackingRect: faceTrackNum];
      faceTrackNum = 0;
    }
}


- (void)mouseEntered:(NSEvent *)theEvent
{
  NSPoint p = [self convertPoint: [theEvent locationInWindow] fromView: nil];

  TRACE(mouseEntered);

  last_mouse_movement_time = EV_TIMESTAMP (theEvent);

  if ([theEvent trackingNumber] == frameTrackNum )
    {
      if (emacsframe && emacsframe->auto_raise
          && !emacsframe->iconified && !emacsframe->async_iconified)
        {
          [[self window] makeKeyAndOrderFront: NSApp];
        }
    }
  else if ([theEvent trackingNumber] == windowTrackNum )
    {
      note_mouse_movement (emacsframe, p.x, p.y);
      note_mouse_highlight (emacsframe, p.x, p.y);
      [self setMouseTracking: mouse_face_mouse_rect];
    }
  else if ([theEvent trackingNumber] == faceTrackNum )
    {
      note_mouse_movement (emacsframe, p.x, p.y);
      note_mouse_highlight (emacsframe, p.x, p.y);
      [self setMouseTracking: mouse_face_mouse_rect];
    }
}

- (void)mouseExited:(NSEvent *)theEvent
{
  NSPoint p = [self convertPoint: [theEvent locationInWindow] fromView: nil];
  NSRect r;

  TRACE(mouseExited);

  last_mouse_movement_time = EV_TIMESTAMP (theEvent);

  if ([theEvent trackingNumber] == frameTrackNum )
    {
      if (emacsframe && emacsframe->auto_lower
          && !emacsframe->iconified && !emacsframe->async_iconified)
        {
          [[self window] orderBack: NSApp]; /* XXX unfocus ! */
        }
    }
  else if ([theEvent trackingNumber] == windowTrackNum )
    {
      [self removeMouseTracking];
      if (emacsframe == mouse_face_mouse_frame)
        clear_mouse_face ();
    }
  else if ([theEvent trackingNumber] == faceTrackNum )
    {
      note_mouse_movement (emacsframe, p.x, p.y);
      note_mouse_highlight (emacsframe, p.x, p.y);

      r = [self bounds];
      if (emacsframe == mouse_face_mouse_frame &&
          [self mouse:p inRect:r] && !mouse_face_defer &&
          !mouse_face_deferred_gc)
        [self setMouseTracking: mouse_face_mouse_rect];
      else if (self == mouse_face_tracked_view)
        [self removeMouseTracking];
    }
}

- menuDown: sender
{
  NSEvent *theEvent;
  NSPoint p;
  EmacsMenuObject *obj;

  TRACE(menuDown);

  if (eventsleft <= 0)
    return self;
  theEvent=[[self window] currentEvent];
  ns_menu_path = Qnil;

  for (obj = [sender representedObject];
       obj != nil;
       obj=[[obj supercell] representedObject])
    {
      ns_menu_path = Fcons([obj value], ns_menu_path);
//NSLog(@"supercell is '%@'", [obj supercell]);
    }
//debug_print(ns_menu_path);

  p = [[self window] mouseLocationOutsideOfEventStream];
  p = [self convertPoint:p fromView:nil];
  
  events->kind = mouse_click;
  XSET (events->code, Lisp_Int, 0);
  events->modifiers=EV_MODIFIERS(theEvent)|down_modifier;
  XSET (events->x, Lisp_Int, (int)rint(p.x));
  XSET (events->y, Lisp_Int, -1);
  EV_TRAILER(theEvent);
  return self;
}

- (void)drawRect:(NSRect)rect
{
  TRACE(drawRect);

  if (!emacsframe || !emacsframe->output_data.ns)
    return;

  emacsframe->async_visible = 1;
  emacsframe->async_iconified = 0;
  SET_FRAME_GARBAGED(emacsframe);
}

/* NSDraggingDestination protocol methods.  Actually this is not really a
   protocol, but a category of Object.  O well...  */

-(unsigned int) draggingEntered:(id <NSDraggingInfo>) sender
{
  TRACE(draggingEntered);
  return NSDragOperationGeneric;
}

-(BOOL)prepareForDragOperation:(id <NSDraggingInfo>) sender
{
  return YES;
}

-(BOOL)performDragOperation:(id <NSDraggingInfo>) sender
{
  id pb;
  int x,y;
  NSString *type;
  NSEvent *theEvent=[[self window] currentEvent];
  NSPoint position;

  TRACE(performDragOperation);
  position=[theEvent locationInWindow];
  position = [self convertPoint:position fromView:nil];
  pixel_to_glyph_coords (emacsframe, (int)rint(position.x),
                         (int)rint(position.y), &x, &y, 0, 1);

  pb = [sender draggingPasteboard];
  type = [pb availableTypeFromArray:ns_drag_types];
  if (type == 0)
    {
      return NO;
    }
  else if ([type isEqualToString:NSFilenamesPboardType])
    {
      NSArray *files;
      NSEnumerator *fenum;
      NSString *file;

      if (!(files = [pb propertyListForType:type]))
        return NO;

      fenum = [files objectEnumerator];
      while ( (file = [fenum nextObject]) )
        {
          events->kind=non_ascii_keystroke;
          events->code=KEY_NS_DRAG_FILE;
          XSET(events->x, Lisp_Int, x);
          XSET(events->y, Lisp_Int, y);
          ns_input_file=append2(ns_input_file,build_string([file UTF8String]));
          events->modifiers=EV_MODIFIERS(theEvent);
          EV_TRAILER(theEvent);
        }


      return YES;
    }
  else if ([type isEqualToString:NSURLPboardType])
    {
      NSString *file;
      NSURL *fileURL;

      if (!(fileURL = [NSURL URLFromPasteboard:pb]) ||
          [fileURL isFileURL] == NO)
        return NO;

      file = [fileURL path];
      events->kind=non_ascii_keystroke;
      events->code=KEY_NS_DRAG_FILE;
      XSET(events->x, Lisp_Int, x);
      XSET(events->y, Lisp_Int, y);
      ns_input_file=append2(ns_input_file, build_string([file UTF8String]));
      events->modifiers=EV_MODIFIERS(theEvent);
      EV_TRAILER(theEvent);

      return YES;
    }
  else if ([type isEqualToString:NSStringPboardType]
           || [type isEqualToString:NSTabularTextPboardType])
    {
      NSString *data;

      if (! (data = [pb stringForType:type]))
        return NO;

      events->kind=non_ascii_keystroke;
      events->code=KEY_NS_DRAG_ASCII;
      XSET(events->x, Lisp_Int, x);
      XSET(events->y, Lisp_Int, y);
      ns_input_ascii=build_string([data UTF8String]);
      events->modifiers=EV_MODIFIERS(theEvent);
      EV_TRAILER(theEvent);

      return YES;
    }
  else if ([type isEqualToString:NSColorPboardType])
    {
      NSColor * c=[NSColor colorFromPasteboard:pb];
      events->kind=non_ascii_keystroke;
      events->code=KEY_NS_DRAG_COLOR;
      XSET(events->x, Lisp_Int, x);
      XSET(events->y, Lisp_Int, y);
      ns_input_color=ns_color_to_lisp(c);
      events->modifiers=EV_MODIFIERS(theEvent);
      EV_TRAILER(theEvent);
      return YES;
    }
  else if ([type isEqualToString:NSFontPboardType])
    {
      // impl based on GNUstep NSTextView.m
      NSData *data = [pb dataForType: NSFontPboardType];
      NSDictionary *dict = [NSUnarchiver unarchiveObjectWithData: data];
      NSFont *font = [dict objectForKey: NSFontAttributeName];
      char fontSize[10];

      if (font == nil)
        return NO;

      events->kind=non_ascii_keystroke;
      events->code=KEY_NS_CHANGE_FONT;
      XSET(events->x, Lisp_Int, x);
      XSET(events->y, Lisp_Int, y);
      ns_input_font = build_string([[font fontName] UTF8String]);
      snprintf(fontSize, 10, "%f", [font pointSize]);
      ns_input_fontsize = build_string(fontSize);
      events->modifiers=EV_MODIFIERS(theEvent);
      EV_TRAILER(theEvent);
      return YES;
    }
  else
    {
      error("Invalid data type in dragging pasteboard.");
      return NO;
    }
}

- validRequestorForSendType: (NSString *)typeSent
                 returnType: (NSString *)typeReturned
{
  TRACE(validRequestorForSendType);
  if ([ns_send_types indexOfObjectIdenticalTo:typeSent] != NSNotFound &&
      [ns_return_types indexOfObjectIdenticalTo:typeSent] != NSNotFound)
    return self;

  return [super validRequestorForSendType: typeSent
                               returnType: typeReturned];
}

- setMiniwindowImage
{
  id image = [[self window] miniwindowImage];
  TRACE(setMiniwindowImage);
  if (image == emacsframe->output_data.ns->miniimage)
    return self;
  if (image && [image isKindOfClass:[EmacsImage class]])
    [image release];
  [[self window] setMiniwindowImage:emacsframe->output_data.ns->miniimage];
  return self;
}

- (void) setRows: (int) r andColumns: (int) c
{
  rows = r;
  cols = c;
}

- (void) addSBRect:(NSRect)r
{
  int i, j; 
  for (i=0; (sbRects[i].origin.x < r.origin.x) && (i < nsbRects); i++);
  for (j=nsbRects-1; j>=i; j--)
    sbRects[j+1] = sbRects[j];
  sbRects[i] = r;
  nsbRects++;
//fprintf(stderr, "inserted rect #%d (%.0f, %.0f): %.0f x %.0f at position %d\n", nsbRects, r.origin.x, r.origin.y, r.size.width, r.size.height, i);
}

- (void) dropSBRect:(NSRect)r
{
  int i;
  for (i=0; i<nsbRects; i++)
    {
      if (NSEqualRects(sbRects[i], r))
        {
          int j;
          for (j=i; j<nsbRects-1; j++)
            sbRects[j] = sbRects[j+1];
          nsbRects--;
//fprintf(stderr, "removed rect (%.0f, %.0f): %.0f x %.0f from position %d\n", r.origin.x, r.origin.y, r.size.width, r.size.height, i);
          break;
        }
    }
}


@end


@implementation EmacsScroller
+ (float) scrollerWidth
{
  return [NSScroller scrollerWidth]
#ifdef GNUSTEP
    - 3
#endif
    ;
}

- initFrame:(NSRect )r window:(Lisp_Object)nwin
{
  TRACE(EmacsScroller_initFrame);

  [super initWithFrame:NSMakeRect(0,0,0,0)];
  [self setTarget:self];
  [self setAction:@selector(scrollerMoved:)];
  [self setContinuous:YES];
  [self setEnabled:YES];

    // Ensure auto resizing of scrollbars occurs within the emacs frame's view
    // locked against the right, top and bottom edges.
  [self setAutoresizingMask: NSViewMinXMargin | NSViewHeightSizable];

  win = nwin;
  condemned = NO;
  fake_event_p = NO;
  repeat = NO;
  lastInc = 0.0;

  frame = XFRAME(XWINDOW (win)->frame);
  if (FRAME_LIVE_P(frame))
    {
      int i;
      EmacsView *view = FRAME_NS_VIEW(frame);
      NSView *sview = [[view window] contentView];
      NSArray *subs = [sview subviews];

      // disable optimization stopping redraw of other scrollbars
      view->scrollbarsNeedingUpdate = 0;
      for (i=[subs count]-1; i>=0; i--)
        if ([[subs objectAtIndex: i] isKindOfClass: [EmacsScroller class]])
          view->scrollbarsNeedingUpdate++;
      [sview addSubview: self];
    }

  [self setFrame: r];

  return self;
}

- (void)setFrame:(NSRect)newRect
{
  NSRect oldRect = [self frame];
  EmacsView *view;
  newRect.size.width = [EmacsScroller scrollerWidth];

  TRACE(EmacsScroller_setFrame);
//fprintf(stderr, "scrollbar set (%.0f, %.0f): %.0f x %.0f, live=%d\n", newRect.origin.x, newRect.origin.y, newRect.size.width, newRect.size.height,FRAME_LIVE_P(f));
//fprintf(stderr, "old rect = (%.0f, %.0f): %.0f x %.0f\n", oldRect.origin.x, oldRect.origin.y, oldRect.size.width, oldRect.size.height);

  if (/*NSEqualRects(oldRect , newRect) ||*/ !FRAME_LIVE_P(frame))
    return;

  view = FRAME_NS_VIEW(frame);

  if (!NSEqualRects(oldRect, newRect))
    {
      [super setFrame: newRect];
      [view dropSBRect: oldRect];
      [view addSBRect: newRect];
    }

  [self display];
}

- (void)dealloc
{
  TRACE(EmacsScroller_dealloc);
  if (!NILP(win))
    XWINDOW (win)->vertical_scroll_bar = Qnil;
  [super dealloc];
  return;
}

- (void)resetCursorRects
{
  NSRect visible = [self visibleRect];
  TRACE(resetCursorRects);

  if (!NSIsEmptyRect(visible))
    [self addCursorRect:visible cursor:[NSCursor arrowCursor]];
  [[NSCursor arrowCursor] setOnMouseEntered: YES];
}

- scrollerMoved:sender
{
  NSEvent *e=[[self window] currentEvent];
  TRACE(EmacsScroller_scrollerMoved);

  if (eventsleft <= 0)
    return nil;
  events->kind = scroll_bar_click;
  events->part = scroll_bar_handle;

  if (sender != nil)
    {
      if (repeat == NO)
        {
          repeat = YES;
        }
      else
        {
          double f = [self floatValue];
          double p = [self knobProportion];
          [self setFloatValue: f+lastInc*p knobProportion: p];
        }
    }

  if ([self floatValue] > 0.999)
    {
      XSET (events->x, Lisp_Int, SHRT_MAX);
      XSET (events->y, Lisp_Int, SHRT_MAX);
    }
  else
    {
      XSET (events->x, Lisp_Int, SHRT_MAX * [self floatValue]);
      XSET (events->y, Lisp_Int, SHRT_MAX);
    }
  events->code=0;
  events->modifiers=EV_MODIFIERS(e) | up_modifier;
  events->frame_or_window=win;
  events->timestamp=EV_TIMESTAMP (e);
  events++;
  eventsleft--;
  ns_send_appdefined (-1);

  return self;
}

- (int) checkSamePosition:(int) position portion:(int) portion whole:(int) whole
{
  return em_position==position && em_portion==portion && em_whole==whole
    && portion!=whole; //PENDING: needed for resize empty buf; another way?
}

- setPosition: (int)position portion: (int)portion whole: (int)whole
{
  TRACE(setPosition);

  em_position = position;
  em_portion = portion;
  em_whole = whole;

  if (portion>=whole)
    [self setFloatValue:0.0 knobProportion:1.0];
  else
    {
      float pos = (float)position/whole;
      float por = (float)portion/whole;
      [self setFloatValue: pos knobProportion: por];
    }
  return self;
}

- condemn
{
  TRACE(condemn);
  condemned=YES;
  return self;
}

- reprieve
{
  TRACE(reprieve);
  condemned=NO;
  return self;
}

- judge
{
  TRACE(judge);
  if (condemned)
    {
      // ensure other scrollbar updates after deletion
        EmacsView *view = (EmacsView *)FRAME_NS_VIEW(frame);
        if (view != nil)
          {
            view->scrollbarsNeedingUpdate++;
            [view dropSBRect: [self frame]];
          }
        [self removeFromSuperview];
        [self release];
    }
  return self;
}

/* Asynchronous mouse tracking for scroller.  This allow us to dispatch
   mouseDragged events without going into a modal loop.  */

- (void)mouseDown:(NSEvent *)e
{ 
  NSRect fr, r;
  // hitPart is only updated AFTER event is passed on
  NSScrollerPart part = [self testPart: [e locationInWindow]];
  double inc = 0.0, loc;
  int edge = 0;

  TRACE(EmacsScroller_mouseDown);
  repeat = NO;

  if (fake_event_p || part == NSScrollerKnobSlot)
    {
      // this is a result of mouseDragged sending a fake event, or
      // on GNUstep a click in the slot;
      // make knob centered at mouse, and send a fake up event
      fr = [self frame];
      r = [self rectForPart:NSScrollerKnobSlot];
      [self convertRect:r toView:nil];
      loc = (NSHeight(fr) - ([e locationInWindow].y - NSMinY(fr)) - NSMinY(r))
        / NSHeight(r);
      if (loc <= 0.0)
        {
          loc = 0.0;
          edge = -1;
        }
      else if (loc >= 1.0)
        {
          loc = 1.0;
          edge = 1;
        }
      [self setFloatValue: loc knobProportion: [self knobProportion]];
      [self scrollerMoved: nil];
    }
  else
    {
      switch (part)
        {
        case NSScrollerDecrementPage:
          inc = -1.0; break;
        case NSScrollerIncrementPage:
          inc = 1.0; break;
        case NSScrollerDecrementLine:
          inc = -0.1; break;
        case NSScrollerIncrementLine:
          inc = 0.1; break;
        default:  // NSScrollerNoPart, NSScrollerKnob, NSScrollerKnobSlot
          break; // handled below
        }

      if (inc != 0.0)
        {
          double f = [self floatValue];
          double p = [self knobProportion];
          lastInc = inc;
          [self setFloatValue: f+inc*p knobProportion: p];
          [super mouseDown: e];
          return;
        }
      else
        lastInc = 0.0;
    }

  // if we get here, user clicked the knob or we got a fake event;
  // either way we want to create a complementary mouseUp

  /* Fake a NX_LMOUSEUP event in the next call to
       `getNextEvent:waitFor:threshold:'.  This is made from our super's
       trackKnob: method which will then exit it's modal loop after calling
       our `scrollerMoved:' method.  */

  fake_event_p = NO;
  fake_event = [NSEvent mouseEventWithType: NSLeftMouseUp
                                  location: [e locationInWindow]
                             modifierFlags: [e modifierFlags]
                                 timestamp: [e timestamp]
                              windowNumber: [e windowNumber]
                                   context: [e context]
                               eventNumber: [e eventNumber]
                                clickCount: [e clickCount]
                                  pressure: [e pressure]];

  /* Get the rect of the Scroller's knob in the screens coordinate system  */
  fr = [self frame];
  r = [self rectForPart:NSScrollerKnob];
  [self convertRect:r toView:nil];
  loc = NSHeight(fr) - ([e locationInWindow].y - NSMinY(fr));

  if (edge || (NSMinY(r) <= loc && loc <= NSMaxY(r)))
    {
      /* Compute the relative offset in the knob of the mousedown.  */
      last_mouse_offset = edge ? 0.5 * edge
        : (loc - NSMinY(r)) / NSHeight(r) - 0.5;
    }
  else
    {
      /* XXX The mouse down is outside the knob, so pretend it's in the
         middle of the knob.  Since the knob will move there.  */
      last_mouse_offset = 0.0;
    }

  [super mouseUp: fake_event];
}

/*
 * This is called while we manually track scroller drags, rather than the
 * superclass doing it.  We generate a fake mouseDown which causes scroller
 * movement followed by (in self-mouseDown) a fake mouseUp.
 */
- (void)mouseDragged:(NSEvent *)e
{
    NSRect r;
    NSPoint location;

    TRACE(EmacsScroller_mouseDragged);

    /* Get the rect of the Scroller's knob in screens coordinates.  */
    r = [self rectForPart:NSScrollerKnob];
    r = [self convertRect:r toView:nil];

    /* Generate a NX_LMOUSEDOWN event on the position the mouse last was. */

    /* Make the mousedown happen at the same relative offset from the bottom
       of the knob as the initial mousedown.  */

    fake_event_p = YES;
    /* Move the x coordinate of the mouse event into the frame of
       the scrollbar handle.  */
    location=NSMakePoint(r.origin.x + r.size.width / 2,
                         [e locationInWindow].y-last_mouse_offset*NSHeight(r));

    fake_event = [NSEvent mouseEventWithType:NSLeftMouseDown
                          location:location
                          modifierFlags:[e modifierFlags]
                          timestamp:[e timestamp]
                          windowNumber:[e windowNumber]
                          context:[e context]
                          eventNumber:[e eventNumber]
                          clickCount:0
                          pressure:[e pressure]];

    [self mouseDown:fake_event];
}

@end

#if defined(MACOSX) && (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_3)
/* This undocumented Quartz function controls how fonts are anti-aliased.
   (Found from code in Mac wxWindows impl, discovered by running `nm' on
   the "QD" framework.)
   Mode 0 is normal anti-aliasing, mode 1 is no anti-aliasing, and mode 2 is
   4-bit pixel-aligned anti-aliasing (the old QuickDraw standard). */
extern void CGContextSetFontRenderingMode(CGContextRef cg, int v);
#endif

#if defined(MACOSX) || defined(GNUSTEP)
// ============================================================================
//
//   Low-level drawing routines (generated in nswraps.m for OpenStep)
//
// ============================================================================

void nswrap_moveshow(float x, float y, const int txt[], int size,
                     struct ns_font *font, NSColor *col, NSColor *bgCol,
                     int ul, float voffset, float w)
// ----------------------------------------------------------------------------
//  Move to target point and draw characters; voffset compensates for font var
// ----------------------------------------------------------------------------
{
    static unsigned char buf[1024];
    static float lastW = 0.0;
    float px, py;
    int i;
    NSString *str;

    TRACE(nswrap_moveshow);
    px = x;
    py = y + voffset;

    // convert mule ints to UTF-8 chars in buf
    muleToUTF8s(txt, size, buf);

#ifdef GNUSTEP
    // use DPS
  {
    static float xoffsets[512];
    NSGraphicsContext *ctxt = GSCurrentContext();

    // set up font and color
    DPSgsave(ctxt);
    [font->nsfont set];
    [col set];

    // do underline
    if (ul)
      {
        DPSmoveto(ctxt, px, py+2);
        DPSlineto(ctxt, px+size*w, py+2);
      }

    // draw with DPSxshow()
    if (size > 512)
      {
        fprintf(stderr, "nswrap_moveshow: run too long\n");
        return;
      }

    if (w != lastW)
      {
        for (i=0; i<512; i++)
          xoffsets[i] = w;
        lastW = w;
      }

    if (bgCol != nil)
      {
        [bgCol set];
        DPSmoveto(ctxt, px,py);
//[[NSGraphicsContext currentContext] GSSetTextDrawingMode: GSTextFillStroke]; /// not implemented yet
        DPSxshow(ctxt, buf, xoffsets, size); // size is an upper bound..
        DPSstroke(ctxt);
        [col set];
//[[NSGraphicsContext currentContext] GSSetTextDrawingMode: GSTextFill]; /// not implemented yet
      }
    DPSmoveto(ctxt, px,py);
    DPSxshow(ctxt, buf, xoffsets, size); // size is an upper bound..
    DPSstroke(ctxt);
    DPSgrestore(ctxt);
    return;
  }
#endif

  // remaining methods need this
  str = [NSString stringWithUTF8String: buf];

#if defined(MACOSX) && (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_3)
  // Use Quartz and an NSGlyphGenerator
  {
    static EmacsGlyphStorage *glyphStorage;
    static NSGlyphGenerator *glyphGenerator;
    static CGAffineTransform fliptf;
    static CGSize *advances;
    static BOOL firstTime = YES;
    CGContextRef gcontext =
      [[NSGraphicsContext currentContext] graphicsPort];
    unsigned int gInd=0, cInd=0;

    if (firstTime == YES)
      {
        firstTime = NO;
        glyphStorage = [[EmacsGlyphStorage alloc] init];
        glyphGenerator = [NSGlyphGenerator sharedGlyphGenerator];
        fliptf = CGAffineTransformMakeScale(1.0,-1.0);
        advances = (CGSize *) xmalloc(512*sizeof(CGSize));
        bzero(advances, 512*sizeof(CGSize));
      }

    [glyphStorage setString: str font: font->nsfont];
    [glyphGenerator generateGlyphsForGlyphStorage: glyphStorage
                        desiredNumberOfCharacters: glyphStorage->maxChar
                                       glyphIndex: &gInd
                                   characterIndex: &cInd];

    CGContextSaveGState(gcontext); // doesn't hurt, but not needed
    [col set];
    if (ul)
      {
        CGContextBeginPath(gcontext);
        CGContextMoveToPoint(gcontext, px, py+2);
        CGContextAddLineToPoint(gcontext, px+size*w, py+2);
        CGContextStrokePath(gcontext);
      }

    fliptf.c =  font->synthItal ? Fix2X(kATSItalicQDSkew) : 0.0;

    CGContextSetFont(gcontext, font->cgfont);
    CGContextSetFontSize(gcontext,font->size);

    if (ns_antialias_text == NO || font->size <= ns_antialias_threshold)
      CGContextSetShouldAntialias(gcontext, 0);
    else
      CGContextSetShouldAntialias(gcontext, 1);
    if (ns_use_qd_smoothing)
      CGContextSetFontRenderingMode(gcontext, 2); // 0 is Cocoa, 2 is QD

    CGContextSetTextMatrix(gcontext, fliptf);

    if (w != lastW)
      {
        for (i=0; i<512; i++)
          advances[i].width = w;
        lastW = w;
      }

    if (bgCol != nil)
      {
        // foreground drawing; erase first to avoid overstrike
        [bgCol set];
        CGContextSetTextDrawingMode(gcontext, kCGTextFillStroke);
        CGContextSetTextPosition(gcontext, px, py);
        CGContextShowGlyphsWithAdvances(gcontext,glyphStorage->cglyphs,
                                        advances, glyphStorage->maxGlyph);
        CGContextSetTextDrawingMode(gcontext, kCGTextFill);
      }
    [col set];
    CGContextSetTextPosition(gcontext, px, py);
    CGContextShowGlyphsWithAdvances(gcontext,glyphStorage->cglyphs,
                                    advances, glyphStorage->maxGlyph);

    CGContextRestoreGState(gcontext);
    return;
  }
#endif  // MACOSX 10.3+

  // Fallback approach that works on OS X and GNUstep, but suffers from
  // problem that NSLayoutManager tunes the spacing so that for certain
  // fonts and text characters, text goes off the grid that Emacs expects.
  {
    NSDictionary *attribs;
    NSNumber *num = nil;
    id objects[4], keys[4];
    int n = 0;

    if (font->nsfont)
      {
        keys[n] = (id) NSFontAttributeName;
        objects[n++] = font->nsfont;
      }
    if (col)
      {
        keys[n] = (id) NSForegroundColorAttributeName;
        objects[n++] = col;
      }
    if (ul)
      {
        num = [NSNumber numberWithInt: ul];
        keys[n] = (id) NSUnderlineStyleAttributeName;
        objects[n++] = num;
      }

    attribs = [NSDictionary dictionaryWithObjects: objects
                                          forKeys: keys
                                            count: n];

    [str drawAtPoint: NSMakePoint(px, py) withAttributes: attribs];
  }
}


#if defined(MACOSX) && (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_3)
// helper for Quartz-based drawing..
@implementation EmacsGlyphStorage

- init
{
  maxChar = 0;
  maxGlyph = 0;
  dict = [NSMutableDictionary new];
  return self;
}

- (void) setString: (NSString *)str font: (NSFont *)font
{
  [dict setObject: font forKey: NSFontAttributeName];
  attrStr = [[[NSAttributedString alloc] initWithString: str attributes: dict]
                autorelease];
  maxChar = [str length];
  maxGlyph = 0;
}

// NSGlyphStorage protocol
- (unsigned int)layoutOptions
{
  return 0;
}

- (NSAttributedString *)attributedString
{
  return attrStr;
}

- (void)insertGlyphs:(const NSGlyph *)glyphs length:(unsigned int)length
        forStartingGlyphAtIndex:(unsigned int)glyphIndex
        characterIndex:(unsigned int)charIndex
{
  len = glyphIndex+length;
  for (i=glyphIndex; i<len; i++)
    cglyphs[i] = glyphs[i-glyphIndex];
  if (len > maxGlyph)
    maxGlyph = len;
}

- (void)setIntAttribute:(int)attributeTag value:(int)val
        forGlyphAtIndex:(unsigned)glyphIndex
{
  return;
}

@end
#endif

// Preferences window
@implementation EmacsPrefsController

- init
{
  [NSBundle loadNibNamed: @"preferences" owner: self];
  prevShrinkSpace = QXFLOAT(ns_shrink_space);
  prevBlinkRate = QXFLOAT(ns_cursor_blink_rate);
#ifdef MACOSX
  prevUseHighlightColor = ns_use_system_highlight_color;
#endif
  return self;
}

- (void) showForFrame: (struct frame *)f
{
  frame = f;
  [self setPanelFromValues];
  [prefsWindow makeKeyAndOrderFront:self];
  [prefsWindow display];
}

- (void) setPanelFromValues
{
  int cursorType = ns_lisp_to_cursor_type(get_frame_param(frame,Qcursor_type));

  [shrinkSpaceSlider setFloatValue: QXFLOAT(ns_shrink_space)];
  [cursorBlinkSlider setFloatValue: (NILP(ns_cursor_blink_rate) ?
                                     0.0 : QXFLOAT(ns_cursor_blink_rate))];
  [cursorTypeMatrix selectCellWithTag: (cursorType == filled_box ? 1 :
                                        (cursorType == bar ? 2 :
                                         (cursorType == line ? 3 : 4)))];
  [altIsMetaMatrix selectCellWithTag: (!NILP(ns_alternate_is_meta) ? 1 : 2)];
#ifdef MACOSX
  [smoothFontsCheck setState: ns_antialias_text ? YES : NO];
  [useQuickdrawCheck setState: ns_use_qd_smoothing ? YES : NO];
  [useSysHiliteCheck setState: ns_use_system_highlight_color ? YES : NO];
#endif
}

- (void) setValuesFromPanel
{
  int cursorTag = [[cursorTypeMatrix selectedCell] tag];
  int altTag = [[altIsMetaMatrix selectedCell] tag];
  float blinkRate = [cursorBlinkSlider floatValue];
  float shrinkSpace = [shrinkSpaceSlider floatValue];

  if (shrinkSpace != prevShrinkSpace)
    {
      prevShrinkSpace = shrinkSpace;
      ns_shrink_space = make_float(shrinkSpace);
      ///free_float(ns_shrink_space);
      ns_frame_update_shrink_space(frame);
      ns_frame_update_line_height(frame);
      x_set_window_size(frame, 0, frame->width, frame->height);
    }
  ///free_float(ns_cursor_blink_rate);
  if (blinkRate != prevBlinkRate)
    {
      ns_cursor_blink_rate = (blinkRate==0.0 ? Qnil : make_float(blinkRate));
      if (blinkRate != 0.0 && prevBlinkRate != 0.0)
        {  // if changed rates, remove blink handler so change picked up
          [cursor_blink_entry invalidate];
          [cursor_blink_entry release];
          cursor_blink_entry = 0;
          if (ns_highlight_frame)
            {
              Lisp_Object tem=get_frame_param(ns_highlight_frame,Qcursor_type);
              ns_highlight_frame->output_data.ns->desired_cursor =
                ns_lisp_to_cursor_type(tem);
            }
        }
      prevBlinkRate = blinkRate;
    }
  FRAME_NEW_CURSOR(frame) =
                         (cursorTag == 1 ? filled_box :
                          (cursorTag == 2 ? bar :
                           (cursorTag == 3 ? line : hollow_box)));
  store_frame_param(frame, Qcursor_type,
                    ns_cursor_type_to_lisp(FRAME_NEW_CURSOR(frame)));
  ns_alternate_is_meta = (altTag == 1 ? Qt : Qnil);
#ifdef MACOSX
  ns_antialias_text = [smoothFontsCheck state];
  ns_use_qd_smoothing = [useQuickdrawCheck state];
  ns_use_system_highlight_color = [useSysHiliteCheck state];
  if (ns_use_system_highlight_color != prevUseHighlightColor)
    {
      prevUseHighlightColor = ns_use_system_highlight_color;
      if (ns_use_system_highlight_color == YES)
        {
          ns_selection_color = [[NSUserDefaults standardUserDefaults]
                                 stringForKey: @"AppleHighlightColor"];
          if (ns_selection_color == nil)
            ns_selection_color = @"grey";
        }
      else
        ns_selection_color = @"grey";
    }
#endif
  Fcall_interactively(intern("ns-save-preferences"), Qnil, Qnil);
}

// buttons
- (IBAction)cancel:(id)sender
{
#if defined(__linux__) && defined(close)
#undef close
  [prefsWindow close];
#define close sys_close
#else
  [prefsWindow close];
#endif  // linux && close
}

- (IBAction)ok:(id)sender
{
  [self setValuesFromPanel];
#if defined(__linux__) && defined(close)
#undef close
  [prefsWindow close];
#define close sys_close
#else
  [prefsWindow close];
#endif  // linux && close
}

- (IBAction)resetToDefaults:(id)sender
{
  ns_set_default_prefs();
  [self setPanelFromValues];
}

- (IBAction)runHelp:(id)sender
{
  //PENDING: requires user to mouse in window before help shows
  Fcall_interactively(intern("info-ns-emacs"), Qnil, Qnil);
}

- (IBAction)setColors:(id)sender
{
  Lisp_Object lispFrame;
  XSETFRAME(lispFrame, frame);
  Fns_popup_color_panel(lispFrame);
}

- (IBAction)setDefaultFont:(id)sender
{
  Lisp_Object lispFrame;
  XSETFRAME(lispFrame, frame);
  Fns_popup_font_panel(lispFrame);
}

@end

#endif // MACOSX || GNUSTEP
